//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      config.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2006-10-23 V0.1 started
//            2007-01-27 V0.2 shift with DMX Macro 
//            2007-10-16 V0.3 TURNOUT_FEEDBACK_ACTIVATED
//            2008-01-08 V0.4 serial_id and short_turnoff_time added to eeprom
//                            dcc_default_format added.
//                            num_of_*** added (for each command group) 
//            2008-02-02 V0.5 feedback_size added 
//            2008-06-18 V0.6 external Stop         
//
//-----------------------------------------------------------------
//
// purpose:   central station for dcc
// content:   EEPROM configuration
//            
//-----------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>                 // using sscanf and sprintf increases prog. by 4k!
#include <inttypes.h>
#include <avr/pgmspace.h>          // put var to program memory
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include <string.h>

#include "config.h"                // general structures and definitions
#include "rs232.h"                 // tx ready
#if (PARSER == LENZ)
  #include "lenz_parser.h"         // tell pc about status changes       
#endif
#if (PARSER == INTELLIBOX)
  #include "ibox_parser.h"         
#endif

//======================================================================

// Die folgende (assemblerhnliche) Defintion der EEPROM-Variablen ist
// erforderlich, um diese compilerunabhngig auf feste Adressen zu legen.
// Auf diese Adressen kann von auen mit den Befehl zum Lesen und Schreiben
// von Sonderoptionen zugegriffen werden.

// The following definitions (quite similar to assembler) are neccesary
// to assign fixed addresses to variables based in EEPROM.
// This is required to allow external access to eeprom with so called
// "special option" commands.

//======================================================================

const unsigned char opendcc_version PROGMEM = OPENDCC_VERSION;

// this should be linked to beginning of EEPROM!
// Offsets are defined in config.h

uint8_t ee_mem[] EEMEM =
{
    [eadr_OpenDCC_Version]          = OPENDCC_VERSION,
    [eadr_baudrate]                 = DEFAULT_BAUD,
    [eadr_OpenDCC_Mode]             = 0,                        // reserved
    [eadr_virtual_decoder_l]        = (unsigned char) (DMX_MACRO_ADDR),
    [eadr_virtual_decoder_h]        = (unsigned char) (DMX_MACRO_ADDR >> 8),
    [eadr_VersionMirror]            = OPENDCC_VERSION,      // mirror
    [eadr_CTS_usage]                = 255,                  // corresponds to IB SO 006 (255 never use CTS)
    #if (TURNOUT_FEEDBACK_ACTIVATED == 1)
    [eadr_s88_mode]                 = 0b0010,               // s88 mode = feedback only; see s88.c, S88_FEEDBACK
    #else
    [eadr_s88_mode]                 = 0b0001,               // s88 mode = hw read only; see s88.c, S88_HW_READ
    #endif
                                                            // Bit 0: 0 disabled, 1 hw read enabled
                                                            // Bit 1: 0 feedback disabled, 1 hw enabled 
    [eadr_s88_autoread]             = 0x06,                 // no of s88 bytes to be read automatically (total)
    [eadr_s88_size1]                = 0x02,
    [eadr_s88_size2]                = 0x02,
    [eadr_s88_size3]                = 0x02,
    [eadr_invert_accessory]         = 0x01,                 // set this Option for Lenz Parser
    [eadr_dcc_acc_repeat]           = NUM_DCC_ACC_REPEAT,   // Accessory Command repeat counter
    [eadr_dcc_acc_time]             = 100,                  // turn on time of acc (used by IB, default 100)
    [eadr_startmode_ibox]           = 0x00,                 // 0=Normal Mode,
                                                            // 1=fixed to P50X
    [eadr_feedback_s88_offset]      = 0,                    // offset for feedback numbers in s88 array, given in bytes
    [eadr_feedback_s88_type]        = 2,                    // status, okay, or error feedback
                                                            // 0: positive ack feedback (set s88, if feedback was received)
                                                            // 1: error feedback (set s88, if no feedback was received)
                                                            // 2: coil position
    [eadr_extend_prog_resets]       = 3,                    // add this number the number of resets command during programming
    [eadr_extend_prog_command]      = 3,                    // add this number the number of prog command to releave timing
    [eadr_dcc_pom_repeat]           = NUM_DCC_POM_REPEAT,   // CV20:
    [eadr_dcc_speed_repeat]         = NUM_DCC_SPEED_REPEAT, // CV21
    [eadr_dcc_func_repeat]          = NUM_DCC_FUNC_REPEAT,
    [eadr_reserved023]              = 0,
    [eadr_dcc_default_format]       = DCC_DEFAULT_FORMAT,
    [eadr_reserved025]              = 0,
    [eadr_reserved026]              = 0,
    [eadr_reserved027]              = 0,
    [eadr_reserved028]              = 0,
    [eadr_xpressnet_feedback]       = 0,                    // CV29: Xpressnet Feedback mode: 0=256trnt,512feedb. 1; only feedback,]              = 0,
    [eadr_s88_clk_timing]           = 2,                    // CV30: S88 Timing (unit 10us) clk high und clk low; rang 1..16
    #if (TURNOUT_FEEDBACK_ACTIVATED == 1)
    [eadr_feedback_s88_size]        = 64,                   // CV31: default 512 turnouts
    #else
    [eadr_feedback_s88_size]        = 0,                    // CV31: default no feedback
    #endif
    [eadr_s88_total_from_pc]        = 0,
    [eadr_I2C_present]              = 0,                    //  I2C present - return 0  
    [eadr_short_turnoff_time]       = SHORT_TURNOFF_TIME / TICK_PERIOD,   // 34: Time until shutdown[eadr_reserved034]              = 0,          
    [eadr_prog_short_toff_time]     = PROG_SHORT_TURNOFF_TIME / TICK_PERIOD,          
    [eadr_ext_stop_enabled]         = 0,                    // 36: 0=default, 1=enable external Stop Input
    [eadr_reserved037]              = 0,          
    [eadr_reserved038]              = 0,          
    [eadr_serial_id]                = 1,                    // 39: serial id, must be > 0         
};


//-----------------------------------------------------------------------
// Generelle Variablen und Defines
//
// OpenDCC_Version      Byte, ist hex zu interpretieren z.B. 0x04 = Version 0.04

// baudrate             Byte, Wertebereich siehe RS232

// OpenDCC_Mode         Byte, 0; z.Z. reserved

// virtual_decoder      Word, An dieser Stelle beginnt das remapping von Schaltbefehlen
//                            auf DMX
//                            0xffff: abgeschaltet
//                            DMX_MACRO_ADDR ist in config.h definiert

// s88_mode             Bit 0: 0: no hw reading
//                             1: Standard s88 reading
//                      Bit 1: 0: no turnout feedback
//                             1: turnout feedback enabled

// s88_size1            Byte  Anzahl der Bytes auf dem S88-Strang 1
//                            Size is given as bytes;  16 means 8 modules 
// s88_size2            Byte  Anzahl der Bytes auf dem S88-Strang 2
// s88_size3            Byte  Anzahl der Bytes auf dem S88-Strang 3

// invert_accessory     Bit   if 1: Invertiere bei Lenz-Protokoll rot-grn
//                                  bei accessory commands

// startmode_ibox       Byte  0: = Standard Start Mode (=p50-Mode)
//                            1: = P50X Mode

// dcc_acc_repeat       Byte  0...16: so oft wird ein DCC Accessory command wiederholt

// feedback_s88_offset  Byte  Offset in bytes, where feedback messages go to s88

// feedback_s88_type    EAN   0: Ack-Feedback
//                            1: Position Feedback  

