//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006-2007 Wolfgang Kufer
//               2003 LowLevel Routines by Mathias Dzionsko (madz@gmx.de)
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      dmxout.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2006-06-02 V0.1 started
//            2006-06-14 V0.2 Raumlichtsteuerung mit macros!
//            2006-09-27 V0.3 Documentation added
//            2007-01-15 V0.4 added LED Control, Bugfix in EEPROM Handling
//            2007-01-18 V0.5 Bugfix in do_dmx_operation
//
// tests:     2006-06-14 Test des Dmerungsbergang via Macros -> okay
//
//-----------------------------------------------------------------
//
// purpose:   lowcost central station for dcc
// content:   genereates dmx output and calculates dimm processes
//
// howto:     DMX is generated with inline assembler, tuned to 16MHz
//            Specs of DMX:
//            see http://www.soundlight.de/techtips/dmx512/dmx2000a.htm
//
//
// interface upstream:
//            init_dmxout(void)        // init
//            run_dmxout(void)         // multitask replacement
//            run_dmxkey(void)         // tracers for DMX mode
//            set_dmxcv(entry, local_index, data)
//            do_dmx_operation(ctrl_i) // activates a virtual decoder
//            do_dmx_macro(ctrl_i)     // activates a virtual macro
//
// interface downstream:
//            direct HW access via PORTB and Pin DMX_OUX
//            disable interrupt for 44us
//
//-----------------------------------------------------------------


#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/pgmspace.h>         // put var to program memory
#include <avr/eeprom.h>
#include <avr/interrupt.h>
#include <util/delay.h>
#include <string.h>

#include "config.h"                // general structures and definitions
// #include "dccout.h"                // next message
// #include "parser.h"                // pc_send
#include "status.h"                // timeout engine, set_state, keyboard
// #include "organizer.h"
// #include "rs232.h"
// #include "s88.h"





#if (DMX_ENABLED == 1)

// #define SIMULAT                     // if defined: extra test code

#ifndef DMX_PORT
 #warning: please define a target hardware
 #warning: I need:  DMX_OUT on DMX_PORT (which may be any IO-PORT) 
 #warning: i.e. #define DMX_OUT 4, #define DMX_PORT PORTB
#endif

#define _IN_RAM         0
#define _IN_EEPROM      1                  // define value
#define _IN_EEPROM_FIX  2


#define DMX_MEM_LOC _IN_EEPROM         // if _IN_RAM: allocate DXM-Ctrl and DMX-macro to Data (SRAM)
                                       // if _INEEPROM: allocate DXM-Ctrl and DMX-macro to EEPROM
                                       // if _INEEPROM_FIX: allocate DXM-Ctrl and DMX-macro to EEDMX


#if (DMX_MEM_LOC == _IN_EEPROM)
  #define EEMEM_DMX EEMEM
#endif
#if (DMX_MEM_LOC == _IN_EEPROM_FIX)
  #define EEMEM_DMX __attribute__((section(".ee_dmx")))
#endif


//------------------------------------------------------------------------------
// Timing-Definitionen zur DMX Programmierung

// DMX_UPDATE_PERIOD: damit wird festgelegt, wie oft der DMX Datenstrom erzeugt wird.
// DMX_RESOLUTION:    das ist die Zeiteinheit der einzelnen virtuellen Decoder fr die
//                    Helligkeitsrampe: diese haben 16 bit, so da bei 0,1s Resolution
//                    sich bis zu 100min lange Rampen programmieren lassen.
//                    (DMX_RESOLUTION sollte ein ganzzahliges Vielfaches der
//                    DMX_UPDATE_PERIOD sein. 
// DMX_MACRO_PERIOD:  das ist die Zeiteinheit bei der Macroprogrammierung, in der
//                    die einzelnen virtuellen Decoder nacheinander aufgerufen werden.
//                    1s bedeutet max. Verzgerung von 4,2 min.
//
// Erfahrungsgem dauern Modellbahndmmerungen etwa 1-3min, so da folgende Definitionen
// passen mssten:

#define  DMX_UPDATE_PERIOD    20000L    // 20ms -> 50Hz
#define  DMX_RESOLUTION      100000L
#define  DMX_MACRO_PERIOD   1000000L    // 1s -> 1Hz


// time is given in 0.1s resolution; we do DMX_UPDATE_PERIOD, therefore we have:
#define DMX_OVERSAMPL  (DMX_RESOLUTION / DMX_UPDATE_PERIOD)


// some security checks
#if (DMX_OVERSAMPL == 0)
#warning DMX - wrong timings
#endif
//------------------------------------------------------------------------------
// internal, but static:
enum dmxout_states
  {                                 // actual state
     IDLE,
     WF_TIMESLOT,                   // wait for next active slot
	 WF_PREAMBLE,                   // send preamble
	 WF_DMX_BYTES                   // send bytes
  } dmxout_state;


//-------------------------------------------------------------------------------
//
//   Data fields 
//
//   dmxvar:    'live' array for calculating data
//   dmxctrl:   array with control values (virtual decoders) -> may reside in eeprom
//   dmxmacro:  list of virtual decoders to call             -> may reside in eeprom
//
//--------------------------------------------------------------------------------
//
#ifndef SIZE_DMX                    // should be defined in config.h
  #warning local defines of SIZE_DMX
  #define SIZE_DMX         8         // number of dmx channels (max 250)
                                     // 9 bytes each entry
  #define ESIZE_DMXCTRL   64         // number of control fields (virtual decoders)
  #define ESIZE_DMXMACRO   4         // number of macro lists
  #define ESIZE_MACROLIST 16         // number of entries in one macro
#endif

struct s_dmxvar
  {
    int32_t actual;                 // actual dimm value
    int32_t delta;                  // actual delta to dimm value
    unsigned char dimm;             // final dimm value
  };

struct s_dmxvar dmxvar[SIZE_DMX] =       // array for calculating data
  {{ (100L << 22), 0, 0   },
   { (100L << 22), 0, 0   }
  };

// Scaling: actual and delta are 32 bit value; these are signed and scaled
// 2^22; dmx-values are 8 bit; the scaling is neccesary to achieve a good
// resolution when dimming; signed is used because of the easy up-down calculation.
// one spare bit is reserved, because of overrun checks when the target is not
// met.


//--------------------------------------------------------------------------------------
// dmxmacro
// This is a list of virtual decoders to call during the macro
// To activate a list: set dmxmacro_active_time[list] = 0;
// Note: only chars are allowed (see eeprom_read_... calls)
//
struct s_dmxmacro
  {
    struct                           // each macro consists of 16 call
      {
         unsigned char decoder;      // virtual decoder to call; 
                                     // offset of 1; void entries are 0
	     unsigned char time;         // time, when to call decoder, unit: 1s; 0..254
      }  entry[ESIZE_MACROLIST];
  };


#if (DMX_MEM_LOC == _IN_RAM)
struct s_dmxmacro dmxmacro[ESIZE_DMXMACRO] =     // array with control macros,
#else
struct s_dmxmacro dmxmacro[ESIZE_DMXMACRO] EEMEM_DMX =     // array with control macros,
#endif

                                                        // goes to EEPROM
  {
   {                                            // macro 0  -> enter night
     {{ 0+1, 0},      // decoder  0 nach  0s (dmx0 - day east off) 
      { 2+1, 5},      // decoder  2 nach  5s (dmx1 - day west off)
      { 8+1, 0},      // decoder  8 nach  0s (dmx4 - day east off) 
      {10+1, 5},      // decoder 10 nach  5s (dmx5 - day west off)
      { 7+1,30},      // decoder  7 nach 30s (dmx3 - night on)
      {15+1,30},      // decoder 15 nach 30s (dmx7 - night on)
      {13+1,20},      // decoder 13 nach 20s (dmx6 - red evening on)
      {12+1,42},      // decoder 12 nach 42s (dmx6 - red evening off)
      { 4+1, 0},      // decoder  3 nach  0s (dmx2 - red morning off)
      {16+1,50},      // decoder 16 nach 50s (dmx3 - dimm down night)
      {17+1,50},      // decoder 17 nach 50s (dmx7 - dimm down night)
     },
   },
   {                                            // macro 1  -> enter day
     {{ 1+1,17},      // decoder  1 nach 17s (dmx0 - day east on) 
      { 3+1,12},      // decoder  3 nach 12s (dmx1 - day west on)
      { 9+1,17},      // decoder  9 nach 17s (dmx4 - day east on) 
      {11+1,12},      // decoder 11 nach 12s (dmx5 - day west on)
      { 6+1,22},      // decoder  6 nach 22s (dmx3 - night off)
      {14+1,22},      // decoder 14 nach 22s (dmx7 - night off)
      { 5+1, 5},      // decoder  5 nach  5s (dmx2 - red morning on)
      { 4+1,27},      // decoder  4 nach 27s (dmx2 - red morning off)
      {12+1, 0},      // decoder 12 nach  0s (dmx6 - red evening off)
      { 7+1, 0},      // decoder  7 nach  0s (dmx3 - night on)
      {15+1, 0},       //decoder 15 nach  0s (dmx7 - night on)
     },    
   },
   {                                            // macro 2  -> enter night slowly
     {{20+1,  0},     // decoder 20 nach   0s (dmx0 - day east off) 
      {22+1, 15},     // decoder 22 nach  15s (dmx1 - day west off)
      {28+1,  0},     // decoder 28 nach   0s (dmx4 - day east off) 
      {30+1, 15},     // decoder 30 nach  15s (dmx5 - day west off)
      {27+1, 90},     // decoder 27 nach  90s (dmx3 - night on)
      {35+1, 90},     // decoder 35 nach  90s (dmx7 - night on)
      {33+1, 60},     // decoder 33 nach  60s (dmx6 - red evening on)
      {32+1,126},     // decoder 32 nach 126s (dmx6 - red evening off)
      {24+1,  0},     // decoder 23 nach   0s (dmx2 - red morning off)
      {36+1,150},     // decoder 36 nach 150s (dmx3 - dimm down night)
      {37+1,150},     // decoder 37 nach 150s (dmx7 - dimm down night)
     },
   },
   {                                            // macro 1  -> enter day slowly
     {{21+1, 51},      // decoder 21 nach 51s (dmx0 - day east on) 
      {23+1, 36},      // decoder 23 nach 36s (dmx1 - day west on)
      {29+1, 51},      // decoder 29 nach 51s (dmx4 - day east on) 
      {31+1, 36},      // decoder 31 nach 36s (dmx5 - day west on)
      {26+1, 66},      // decoder 26 nach 66s (dmx3 - night off)
      {34+1, 66},      // decoder 14 nach 66s (dmx7 - night off)
      {25+1, 15},      // decoder 25 nach 15s (dmx2 - red morning on)
      {24+1, 81},      // decoder 24 nach 81s (dmx2 - red morning off)
      {32+1,  0},      // decoder 32 nach  0s (dmx6 - red evening off)
      {27+1,  0},      // decoder 27 nach  0s (dmx3 - night on)
      {35+1,  0},       //decoder 35 nach  0s (dmx7 - night on)
     },    
   }
  } ;    // array with control macros 

// 
// dmxmacro_active_time is the elapsed time (in s) of a macro
// if set to 0: restart macro
// if 255, macro has terminated and is inactiv
unsigned char dmxmacro_active_time[ESIZE_DMXMACRO]; 
                                                            
    
//--------------------------------------------------------------------------------------

struct s_dmxctrl                    // goes to EEPROM
  {
    unsigned char target;           // dmx channel this control acts on
	unsigned char dimm;             // final dimm value
	unsigned char time_l;           // time to reach this value, unit: 0.1s; lower 8 bits
	unsigned char time_h;           // time to reach this value, unit: 0.1s; upper 8 bits
  };


// --------------- the folling initializing is used for following lights:
// dmx in room:
// 
//                 day east: 0        day west: 1
//
//  red morning:2                                       read evening: 6
//
//                 day east: 4        day west: 5
//
// night: 3 and 7
//
// There are 4 macros:
// macro 0  -> enter night in approx. 60s. uses virtual decoders 0..17
// macro 1  -> enter day in approx. 60s. uses virtual decoders 0..17
// macro 2  -> enter night in approx. 180s.  uses virtual decoders 20..37
// macro 3  -> enter day in approx. 180s. uses virtual decoders 20..37


#if (DMX_MEM_LOC == _IN_RAM)
struct s_dmxctrl dmxctrl[ESIZE_DMXCTRL]  =      
#else
struct s_dmxctrl dmxctrl[ESIZE_DMXCTRL] EEMEM_DMX =      // array with control value (virtual decoder)
#endif
  { {   0,   0,  400 & 0xFF,  400 / 256 },    //  0: dmx 0 off    in 40 sec
    {   0, 255,  400 & 0xFF,  400 / 256 },    //  1: dmx 0 on     in 40 sec
    {   1,   0,  400 & 0xFF,  400 / 256 },    //  2: dmx 1 off    in 40 sec
    {   1, 255,  400 & 0xFF,  400 / 256 },    //  3: dmx 1 on     in 40 sec
    {   2,   0,  100 & 0xFF,  100 / 256 },    //  4: dmx 2 off    in 10 sec
    {   2, 255,  100 & 0xFF,  100 / 256 },    //  5: dmx 2 on     in 10 sec
    {   3,   0,   50 & 0xFF,   50 / 256 },    //  6: dmx 3 off    in  5 sec
    {   3, 255,   50 & 0xFF,   50 / 256 },    //  7: dmx 3 on     in  5 sec
    {   4,  30,  400 & 0xFF,  400 / 256 },    //  8: dmx 4 off    in 40 sec
    {   4, 255,  400 & 0xFF,  400 / 256 },    //  9: dmx 4 on     in 40 sec
    {   5,  30,  400 & 0xFF,  400 / 256 },    // 10: dmx 5 off    in 40 sec
    {   5, 255,  400 & 0xFF,  400 / 256 },    // 11: dmx 5 on     in 40 sec
    {   6,   0,  100 & 0xFF,  100 / 256 },    // 12: dmx 6 off    in 10 sec
    {   6, 255,  100 & 0xFF,  100 / 256 },    // 13: dmx 6 on     in 10 sec
    {   7,   0,   50 & 0xFF,   50 / 256 },    // 14: dmx 7 off    in  5 sec
    {   7, 255,   50 & 0xFF,   50 / 256 },    // 15: dmx 7 on     in  5 sec
    {   3, 130,  100 & 0xFF,  100 / 256 },    // 16: dmx 3 middle in 10 sec
    {   7, 130,  100 & 0xFF,  100 / 256 },    // 17: dmx 7 middle in 10 sec
    {   0,   0,    0 & 0xFF,    0 / 256 },    // 18: void
    {   0,   0,    0 & 0xFF,    0 / 256 },    // 19: void
    
    {   0,   0, 1200 & 0xFF, 1200 / 256 },    // 20: dmx 0 off    in 120 sec
    {   0, 255, 1200 & 0xFF, 1200 / 256 },    // 21: dmx 0 on     in 120 sec
    {   1,   0, 1200 & 0xFF, 1200 / 256 },    // 22: dmx 1 off    in 120 sec
    {   1, 255, 1200 & 0xFF, 1200 / 256 },    // 23: dmx 1 on     in 120 sec
    {   2,   0,  300 & 0xFF,  300 / 256 },    // 24: dmx 2 off    in  30 sec
    {   2, 255,  300 & 0xFF,  300 / 256 },    // 25: dmx 2 on     in  30 sec
    {   3,   0,  150 & 0xFF,  150 / 256 },    // 26: dmx 3 off    in  15 sec
    {   3, 255,  150 & 0xFF,  150 / 256 },    // 27: dmx 3 on     in  15 sec
    {   4,  30, 1200 & 0xFF, 1200 / 256 },    // 28: dmx 4 off    in 120 sec
    {   4, 255, 1200 & 0xFF, 1200 / 256 },    // 29: dmx 4 on     in 120 sec
    {   5,  30, 1200 & 0xFF, 1200 / 256 },    // 30: dmx 5 off    in 120 sec
    {   5, 255, 1200 & 0xFF, 1200 / 256 },    // 31: dmx 5 on     in 120 sec
    {   6,   0,  300 & 0xFF,  300 / 256 },    // 32: dmx 6 off    in  30 sec
    {   6, 255,  300 & 0xFF,  300 / 256 },    // 33: dmx 6 on     in  30 sec
    {   7,   0,  150 & 0xFF,  150 / 256 },    // 34: dmx 7 off    in  15 sec
    {   7, 255,  150 & 0xFF,  150 / 256 },    // 35: dmx 7 on     in  15 sec
    {   3, 130,  300 & 0xFF,  300 / 256 },    // 36: dmx 3 middle in  30 sec
    {   7, 130,  300 & 0xFF,  300 / 256 },    // 37: dmx 7 middle in  30 sec
  };


//---------------------------------------------------------------------------------
// calc_dmx_byte((unsigned char dmx_index)
// dmx_index: dieser Kanal wird berechnet
//
// is called every timeslot (DMX_UPDATE_PERIOD), scans the dmxvar and builds dmxbyte

unsigned char calc_dmx_byte(unsigned char dmx_index)
  {
    register unsigned char dmx;

    dmx = dmxvar[dmx_index].actual >> 22;
	if (dmx != dmxvar[dmx_index].dimm)
       {
	     // perform one step
		 dmxvar[dmx_index].actual += dmxvar[dmx_index].delta;
		 
		 // limit to max values
         if (dmxvar[dmx_index].actual < 0) dmxvar[dmx_index].actual = 0;
         if (dmxvar[dmx_index].actual > 0x3fff0000L) dmxvar[dmx_index].actual = 0x3fff0000L;

         // this is not just actual += delta, due to rounding issues it may
		 // happen that the target is never met -> we check target crossing

         dmx = dmxvar[dmx_index].actual >> 22;
		 if (dmxvar[dmx_index].delta < 0)
		   {
		   	 if(dmx < dmxvar[dmx_index].dimm)
			    {
				  // under limit -> set actual at limit
                  dmxvar[dmx_index].actual = (int32_t)dmxvar[dmx_index].dimm << 22;
				  dmx = dmxvar[dmx_index].dimm;
				}
		   }
         else
		   {
		      if(dmx > dmxvar[dmx_index].dimm)
			    {
				  // above limit -> set actual at limit
                  dmxvar[dmx_index].actual = (int32_t)dmxvar[dmx_index].dimm << 22;
				  dmx = dmxvar[dmx_index].dimm;
				}
		   }
	     
       }
    return(dmx);
  }

void dmx_all_on(void)
  {
    unsigned char i;
    
    for (i=0; i<SIZE_DMX; i++)
	  {
	    // set all outputs
	    dmxvar[i].dimm = 255;
		dmxvar[i].actual = 255L << 22;
        dmxvar[i].delta = 0;
	  }

  }


//---------------------------------------------------------------------------------
// if a new dimm value should be activated, call this routine to update
// the runtime vars. Values are read from given ctrl field and put to the 
// dmx-field
//
// delta = difference (new-old) / steps 
//
// time is given in DMX_RESOLUTION; we do DMX_UPDATE_PERIOD, therefore we have
// oversammpling; this is an addtional divider when calculating delta.


#if (DMX_MEM_LOC == _IN_RAM)
 #define read_dmxctrl(index, mytype)  dmxctrl[index].mytype
#else
 #define read_dmxctrl(index, mytype)  eeprom_read_byte(&dmxctrl[index].mytype)
#endif

//----------------------------------------------------------------------------------
// do_dmx_operation(unsigned char ctrl_i)
//  activates a single dmx decoder
//  parameter: ctrl_i; this decoder is activated;

void do_dmx_operation(unsigned char ctrl_i)
  {
    unsigned char dmx_i;
    int32_t newex;
    int32_t diff;
    unsigned int dimmtime;

    dmx_i = read_dmxctrl(ctrl_i, target);                   // get target
	if (dmx_i < SIZE_DMX)                                   // ignore if DMX too large
	  {
    	dmxvar[dmx_i].dimm = read_dmxctrl(ctrl_i, dimm);    // copy dimm to target

        dimmtime = read_dmxctrl(ctrl_i, time_h)*256 + read_dmxctrl(ctrl_i, time_l);
        if (dimmtime == 0) dimmtime = 1;

        newex = ((int32_t) dmxvar[dmx_i].dimm ) << 22;
        diff = newex - dmxvar[dmx_i].actual;
        dmxvar[dmx_i].delta =
            diff / 
            ( DMX_OVERSAMPL * dimmtime );
      }
  }



//----------------------------------------------------------------------------------
// do_dmx_macro(unsigned char ctrl_i)
//  activates a macro, with contains virtual decoders

void do_dmx_macro(unsigned char ctrl_i)
  {
     dmxmacro_active_time[ctrl_i] = 0;
  }


//----------------------------------------------------------------------------------
// do_dmx(unsigned char ctrl_i)
//  ctrl_i:  macro or decoder to call
//     
//  activates a virtual decoder -> my be a macro or a single dmx control
//
//           |  Range:                           |     Type
//      -----|-----------------------------------|------------------------
//           |0 ...............ESIZE_DMXMACRO:   | macro
//           |ESIZE_DMXMACRO ..(ESIZE_DMXCTRL+   | single decoder
//           |                 ESIZE_DMXMACRO )  |


void do_dmx(unsigned char ctrl_i)
  {
    if (ctrl_i < ESIZE_DMXMACRO)
      {
        do_dmx_macro(ctrl_i);
      }
    else
      {
        ctrl_i -= ESIZE_DMXMACRO;
        if (ctrl_i < ESIZE_DMXCTRL)
          {
            do_dmx_operation(ctrl_i);
          }
        // else: out of range -> not handled, just ignored
      }
  }

//----------------------------------------------------------------------------------
// set_dmxcv(entry, local_index, data)
//  sets up the CV of a specific virtual decoder.
//
//  entry:        index of dmx_control to set.
//  local_index:  0: DMX Kanal
//                1: DIMM Wert, der ausgegeben wird
//                2: low(Zeit bis zum Erreichen des Wertes)
//                3: high(Zeit bis zum Erreichen des Wertes)
//  data:         Wert
//
// Beispiel:
//      Virteller Decoder 2 soll DMX Kanal 5 auf 50 in 30 sec steuern:
//      set_dmxcv(2, CV_DMXi, 5);
//      set_dmxcv(2, CV_DIMMi, 50);
//      set_dmxcv(2, CV_TIME_Li, 300 & 0xff);
//      set_dmxcv(2, CV_TIME_Hi, 300 / 256);
  


void set_dmxcv(unsigned char ctrl_i, unsigned char sub_i, unsigned char data)
  {
    unsigned char *myctrl;

    if ((ctrl_i < ESIZE_DMXCTRL) && (sub_i < sizeof dmxctrl[0]))
	  {
	    myctrl = (unsigned char *)dmxctrl;
		// code for RAM: myctrl[ctrl_i*sizeof dmxctrl[0] + sub_i] = data;
        eeprom_write_byte(&myctrl[ctrl_i*sizeof dmxctrl[0] + sub_i], data);
	  }
  }


#if (DMX_MEM_LOC == _IN_RAM)
 #define read_dmxmacro(mymac, myentry, mytype)  dmxmacro[mymac].entry[myentry].mytype
#else
 #define read_dmxmacro(mymac, myentry, mytype)  eeprom_read_byte(&dmxmacro[mymac].entry[myentry].mytype)
#endif

//===========================================================================
//
// Macro Processor
//
// Is called every second from run_dmxout; scans the dmxmacro field for decoders
// to run at the given time point. If a valid decoder is found, the corresponing
// operation is executed. (update dmxvar)
//
// howto activate a macro? -> just set dmxmacro_active_time[macro] = 0;

void proceed_dmxmacro(void)
  {
    unsigned char mymac;
    unsigned char m;
    unsigned char decoder;

    for (mymac = 0; mymac < ESIZE_DMXMACRO; mymac++)
      {
        // search through all macro lists
        if (dmxmacro_active_time[mymac] < 255)
          {
            // found an active one
            for (m=0; m<ESIZE_MACROLIST; m++)
              {
                decoder = read_dmxmacro(mymac, m, decoder);
                if ((decoder != 0) &&
                    (read_dmxmacro(mymac, m, time) == dmxmacro_active_time[mymac]))
                  {
                    // perform this call
                    do_dmx_operation(decoder-1);
    	          }
              }
            dmxmacro_active_time[mymac]++;
          }
      }
  }


//===========================================================================
//
// Lowlevel interface  (Copyright (C) 2003 Mathias Dzionsko (madz@gmx.de))
//
// Die Lowlevel Routinen sind minidmx entnommen und wurden mit Genehmigung
// von Mathias hier eingebaut. Danke!
//
#define DMX_PORT        PORTB

//---------------------------------------------------------------------------
// DMXSendByte
//
// Sendet ein Datenbyte ber die DMX-Leitung.
// Die bentigten Takte fr ein DMX-Bit sind 64 bei 16 MHz. (DMX hat 250kbit, 2%)
// Interrupts werden fr 44 s nicht zugelassen.
//
#define DMX_CYCLES_PER_BIT  (F_CPU / 250000L)
#define DMX_ASM_CYCLES  (1 +  3 +  6 +  0 *  3 +  2 +  4  )
//                      ror+call+ bit+ CNT*cycl+ jmp+ ret = 16
#define DMX_LOOP   ((DMX_CYCLES_PER_BIT - DMX_ASM_CYCLES) / 3)


// unsigned char i = DMX_LOOP;
// if F_CPU is changed, please control Bitduration: DMX_LOOP / F_CPU != 4us
// 14.06.2006: checked in Simulator
// fine adjust with one or more rjmp DMXSendBit4


void DMXSendByte(uint8_t value) __attribute__((noinline));

void DMXSendByte(uint8_t value)
{
  asm volatile (
"                cli                    \n" // keine Ints zulassen
"                clc                    \n" // 1      - Start Bit
"                rcall DMXSendBit       \n" // +3+33=37 Takte
"                                       \n"
"                ror %2                 \n" // 1      - Bit 0
"                rcall DMXSendBit       \n" // +3+33=37 Takte
"                ror %2                 \n" //        - Bit 1
"                rcall DMXSendBit       \n"
"                ror %2                 \n" //        - Bit 2
"                rcall DMXSendBit       \n"
"                ror %2                 \n" //        - Bit 3
"                rcall DMXSendBit       \n"
"                ror %2                 \n" //        - Bit 4
"                rcall DMXSendBit       \n"
"                ror %2                 \n" //        - Bit 5
"                rcall DMXSendBit       \n"
"                ror %2                 \n" //        - Bit 6
"                rcall DMXSendBit       \n"
"                ror %2                 \n" //        - Bit 7
"                rcall DMXSendBit       \n"
"                                       \n"
"                sec                    \n"  //        - Stop Bit 1
"                rcall DMXSendBit       \n"
"                sec                    \n"  //        - Stop Bit 2
"                rcall DMXSendBit       \n"
"                sei                    \n"  // Ints wieder zulassen
"                rjmp DMXSendByteExit   \n"
"                                       \n"
"DMXSendBit:     brcs DMXSendBit1       \n"
"                nop                    \n" // 1+1
"                cbi %0, %1             \n" // +2
"                rjmp DMXSendBit2       \n" // +2=6 Takte
"DMXSendBit1:    sbi %0, %1             \n" // 2+2
"                rjmp DMXSendBit2       \n" // +2=6 Takte
"                                       \n"
"DMXSendBit2:    ldi r31,16             \n" // 1  : Kufer: war 6, jetzt 16 (DMX_LOOP)
"DMXSendBit3:    dec r31                \n" // +1
"                brne DMXSendBit3       \n" // +1/+2
"                rjmp DMXSendBit4       \n" // +2
"DMXSendBit4:    ret                    \n" // +4=9+6*3=27 Takte
"DMXSendByteExit:                       \n" // neu: 54 Takte
//                          %0               %1            %2
    : : "I" (_SFR_IO_ADDR(DMX_PORT)), "I" (DMX_OUT), "r" (value) : "r31"
  );
}

//---------------------------------------------------------------------------
// DMXSendReset

// Sendet ein Reset-Signal mit anschlieendem Start-Byte ber die
// DMX-Leitung.

void DMXSendReset(void)
  {
    DMX_PORT &= ~(1<<DMX_OUT);          // RESET-Signal senden
    _mydelay_us(100);                   // 100s (min. 88s) warten

    DMX_PORT |= (1<<DMX_OUT);           // MARK-Signal senden
    _mydelay_us(10);                    // 10s (min. 8s) warten

    DMXSendByte(0);                     // Startbyte (0) senden
  }


//=================================================================================
//
// run_dmxout: multitask replacement, must be called in loop
//
//=================================================================================

unsigned char cur_dmx_chan;

signed int last_dmx_run;   // timer variable to create a update grid;
signed int last_macro_run;

void run_dmxout(void)
  {
    
    switch (dmxout_state)
      {
        case IDLE:
            last_dmx_run = timerval;                    // remember time 
            last_macro_run = 0;
		    dmxout_state = WF_TIMESLOT;
            break;

        case WF_TIMESLOT:
            // note: cast the difference down to char, otherwise the wrap around fails
            if ((timerval - last_dmx_run) < (DMX_UPDATE_PERIOD / TICK_PERIOD))  return;

            last_dmx_run = timerval;                    // remember time 
		
            last_macro_run++;
            if (last_macro_run == (DMX_MACRO_PERIOD / DMX_UPDATE_PERIOD)) 
              {
                 proceed_dmxmacro();
                 last_macro_run = 0;
              }                   

            // 20ms passed, now calc new dmx vector
			
			dmxout_state = WF_PREAMBLE;
            break;

        case WF_PREAMBLE:
            DMXSendReset();
			dmxout_state = WF_DMX_BYTES;
			cur_dmx_chan = 0;
            break;

        case WF_DMX_BYTES:
		    if (cur_dmx_chan == SIZE_DMX)
			  {
			    // all dmx done
				dmxout_state = WF_TIMESLOT;
				return;
			  }
			
			DMXSendByte(calc_dmx_byte(cur_dmx_chan));
            cur_dmx_chan ++;
            break;
     }
  }


signed int last_go_stroke;
unsigned char last_go_stroke_valid;


void run_dmxkey(void)
  {
    register unsigned char temp;
   
    if (last_go_stroke_valid)
      {
        if ((timerval - last_go_stroke) > (500000L / TICK_PERIOD))
          #if (500000L / TICK_PERIOD) > 127
            #error TICK_PERIOD too small
          #endif
          {
            last_go_stroke_valid = 0;
          }
      }

    temp=run_key();


    if (temp == STOP_STROKE) 
      {  
        dmxmacro_active_time[0] = 0;                    // call night macro
        // slow blinking red led, green off
        LED_GO_OFF;
        led_pwm.go_rest = 0;
            
        LED_STOP_ON;
        led_pwm.stop_rest = 100000L / TICK_PERIOD; 
        led_pwm.stop_ontime = 700000L / TICK_PERIOD;
        led_pwm.stop_offtime = 200000L / TICK_PERIOD;
      }
    if (temp == GO_STROKE) 
      {  
        if (last_go_stroke_valid)
          {
            // doppel 'klick'
            dmx_all_on();
            LED_GO_ON;                                      // fast blinking green
            led_pwm.go_rest = 100000L / TICK_PERIOD; 
            led_pwm.go_ontime = 350000L / TICK_PERIOD;
            led_pwm.go_offtime = 100000L / TICK_PERIOD;
            
          }
        else
          {
            
            dmxmacro_active_time[1] = 0;                    // call day macro
            last_go_stroke = timerval;
            last_go_stroke_valid = 1;
            
            LED_GO_ON;
            led_pwm.go_rest = 100000L / TICK_PERIOD; 
            led_pwm.go_ontime = 700000L / TICK_PERIOD;
            led_pwm.go_offtime = 200000L / TICK_PERIOD;
            LED_STOP_OFF;
            led_pwm.stop_rest = 0;
          }
      }
  }



void init_dmxout(void)
  {
    unsigned char i;
    dmxout_state = IDLE;
    for (i=0; i<SIZE_DMX; i++)
	  {
	    // clear all outputs
	    dmxvar[i].dimm = 0;
		dmxvar[i].actual = 0;
        dmxvar[i].delta = 0;
	  }

     for (i=0; i<ESIZE_DMXMACRO; i++)
	  {
	    // clear all outputs
	    dmxmacro_active_time[i] = 255;     // time over, macro inactiv
      }
  }




//=================================================================================
//
// Test und Simulation
//
//=================================================================================

#ifdef SIMULAT
#warning extra test code -> turn off SIMULAT

// This simulates two ramps - first up, then down;
// note: assignment to PORTx, otherwise the compiler would eat the useless code ;-)

void simu_dmxrampe(void)
  {
     int i;

     do_dmx_operation(1);
	 do_dmx_operation(3);

     for (i=0; i< 70; i++)
	   {	  
         PORTA = calc_dmx_byte(0);
         PORTB = calc_dmx_byte(1);
		 PORTC = i;
       }
     do_dmx_operation(0);
	 do_dmx_operation(2);

     for (i=0; i< 120; i++)
	   {	  
         PORTA = calc_dmx_byte(0);
         PORTB = calc_dmx_byte(1);
       }
     while(1);
  }

#define CV_DMXi     0
#define CV_DIMMi    1
#define CV_TIME_Li  2
#define CV_TIME_Hi  3


void delay20ms(void)
  {
    int i;
    for (i=0; i< 2; i++)  
	  { // each 10ms 
      _mydelay_us(10000);
      }
  }

void simu_dmxout(void)
  {
    int i, j;


 
    init_dmxout();
    // Decoder 0: DMX:0, DIMM: 50, Rampe 30s 
    set_dmxcv(0, CV_DMXi, 0);
    set_dmxcv(0, CV_DIMMi, 50);
	set_dmxcv(0, CV_TIME_Li, 300 & 0xff);
 	set_dmxcv(0, CV_TIME_Hi, 300 / 256);
    // Decoder 1: DMX:0, DIMM: 250, Rampe 2s 
    set_dmxcv(1, CV_DMXi, 0);
	set_dmxcv(1, CV_DIMMi, 250);
	set_dmxcv(1, CV_TIME_Li, 20 & 0xff);
 	set_dmxcv(1, CV_TIME_Hi, 20 / 256);


    while(1)
      {
    	do_dmx_operation(0);
    	for (j=0; j < 35*50; j++)
          {
            dmxout_state = IDLE;
            for (i=0; i< 11; i++) 
              {
                last_dmx_run = timerval - (DMX_UPDATE_PERIOD-TICK_PERIOD);
                run_dmxout(); 
              }
            delay20ms();    
          }

     	do_dmx_operation(1);
    	for (j=0; j < 10*50; j++)
          {
    		dmxout_state = IDLE;
            for (i=0; i< 11; i++) 
              {
                last_dmx_run = timerval - (DMX_UPDATE_PERIOD-TICK_PERIOD);
                run_dmxout(); 
              }
            delay20ms(); 
          }
      }
  }

#endif   // SIMULAT

#endif   // DMX_ENABLED
