//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      hardware.h
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2007-03-27 V0.01 copied from config.h,
//            all removed but hardware accesses.
//
//-----------------------------------------------------------------
//
// purpose:   lowcost central station for dcc
// content:   central definitions used in the project
//            all hardware project settings are done here!!
//            
//            1.   Prozessor, Timing
//                 (change here, if OpenDCC is run on a different uP)
//            2.   IOs
//                 (change here, if OpenDCC is run on a different board)
//
//-----------------------------------------------------------------
#ifndef __HARDWARE_H__
#define __HARDWARE_H__


//========================================================================
// 1. Processor Definitions
//========================================================================
//
#define  ATMEGA32      0
#define  ATMEGA644     1
#define  ATMEGA644P    2        // This type is required for Xpressnet-Extension

#define  PROCESSOR     ATMEGA32           // or ATMEGA644
// if changed: check timings, timer settings, baudrate

#if (PROCESSOR == ATMEGA32)
  // atmega32:   2kByte SRAM, 1kByte EEPROM
  #define SRAM_SIZE    2048
  #define EEPROM_SIZE  1024
  #define EEPROM_BASE  0x810000L
#endif

#if (PROCESSOR == ATMEGA644)
  // atmega644:   4kByte SRAM, 2kByte EEPROM
  #define SRAM_SIZE    4096
  #define EEPROM_SIZE  2048
  #define EEPROM_BASE  0x810000L
#endif

#if (PROCESSOR == ATMEGA644P)
  // atmega644:   4kByte SRAM, 2kByte EEPROM
  #define SRAM_SIZE    4096
  #define EEPROM_SIZE  2048
  #define EEPROM_BASE  0x810000L
#endif

//
#ifndef F_CPU
// prevent compiler error by supplying a default 
# warning "F_CPU not defined for <config.h>", set default to 16MHz 
# define F_CPU 16000000UL
// if changed: check every place where it is used (possible range overflow in preprocessor!)
#endif

//------------------------------------------------------------------------
// Fuses of ATmega32 (just for documentation)
//------------------------------------------------------------------------
// CKSEL3...1 = 1111 (not blown = 16 MHz external crystal)
// CKOPT = 0 (blown = high output swing)
// SUT1..0 = 01 (blown + not blown)
// BODLEVEL = 0 (blown = reset at 4V)



//========================================================================
// 2. Port Definitions
//========================================================================
// Port A
#define S88DIN1    0     // in
#define S88DIN2    1     // in
#define S88DIN3    2     // in
#define EXT_STOP   3     // in
#define LED_GO     4     // out
#define LED_STOP   5     // out
#define KEY_STOP   6     // in
#define KEY_GO     7     // in
// Port B
#define S88_CLK    0     // out
#define S88_RESET  1     // out
#define ACK_DETECT 2     // in    this is only a short pulse -> use int!
#define S88_LOAD   3     // out
#define DMX_OUT    4     // out
#define MY_CTS     6     // out
#define MY_RTS     7     // in
// Port C
#define SHORT_MAIN 0     // in
#define SHORT_PROG 1     // in
#define LED_CTRL   6     // out
#define LED_RS232  7     // out
// Port D
#define MY_RXD     0     // in
#define MY_TXD     1     // out
#define JUMPER1    2     // in
#define JUMPER2    3     // in
#define NDCC       4     // out
#define DCC        5     // out
#define ENA_PROG   6     // out  high = enable programming
#define ENA_MAIN   7     // out  high = enable main

//------------------------------------------------------------------------
// LED-Control and IO-Macros

#define LED_CTRL_ON      PORTC &= ~(1<<LED_CTRL)
#define LED_CTRL_OFF     PORTC |= (1<<LED_CTRL)
#define LED_RS232_ON     PORTC &= ~(1<<LED_RS232)
#define LED_RS232_OFF    PORTC |= (1<<LED_RS232)
#define LED_GO_ON        PORTA &= ~(1<<LED_GO)
#define LED_GO_OFF       PORTA |= (1<<LED_GO)
#define LED_STOP_ON      PORTA &= ~(1<<LED_STOP)
#define LED_STOP_OFF     PORTA |= (1<<LED_STOP)

// Note LEDs are active low -> state on == pin low!
#define LED_CTRL_STATE   (!(PINC & (1<<LED_CTRL)))
#define LED_RS232_STATE  (!(PINC & (1<<LED_RS232)))
#define LED_GO_STATE     (!(PINA & (1<<LED_GO)))
#define LED_STOP_STATE   (!(PINA & (1<<LED_STOP)))

#define MY_CTS_HIGH      PORTB |= (1<<MY_CTS)
#define MY_CTS_LOW       PORTB &= ~(1<<MY_CTS)
#define MY_CTS_STATE     (PORTB & (1<<MY_CTS))

#define DMX_OUT_HIGH     PORTB |= (1<<DMX_OUT)
#define DMX_OUT_LOW      PORTB &= ~(1<<DMX_OUT)

// this results after compile: sbi(PORTB,MY_CTS)

#define MAIN_TRACK_ON    PORTD |= (1<<ENA_MAIN)
#define MAIN_TRACK_OFF   PORTD &= ~(1<<ENA_MAIN)
#define MAIN_TRACK_STATE (PIND & (1<<ENA_MAIN)))
#define PROG_TRACK_ON    PORTD |= (1<<ENA_PROG)
#define PROG_TRACK_OFF   PORTD &= ~(1<<ENA_PROG)
#define PROG_TRACK_STATE (PIND & (1<<ENA_PROG))

// Abfragen:
#define MAIN_IS_SHORT    (!(PINC & (1<<SHORT_MAIN)))
#define PROG_IS_SHORT    (!(PINC & (1<<SHORT_PROG)))
#define ACK_IS_DETECTED  ( (PINB & (1<<ACK_DETECT)))
#define KEY_GO_PRESSED   (!(PINA & (1<<KEY_GO)))
#define KEY_STOP_PRESSED (!(PINA & (1<<KEY_STOP)))
#define EXT_STOP_PRESSED (!(PINA & (1<<EXT_STOP)))

// RTS is aktive low
#define MY_RTS_DETECT    (!(PINB & (1<<MY_RTS)))

#define CHECK_JUMPER1    (!(PIND & (1<<JUMPER1)))
#define CHECK_JUMPER2    (!(PIND & (1<<JUMPER2)))

// S88:
#define S88_LOAD_HIGH    PORTB |= (1<<S88_LOAD)
#define S88_LOAD_LOW     PORTB &= ~(1<<S88_LOAD)
#define S88_RESET_HIGH   PORTB |= (1<<S88_RESET)
#define S88_RESET_LOW    PORTB &= ~(1<<S88_RESET)
#define S88_CLK_HIGH     PORTB |= (1<<S88_CLK)
#define S88_CLK_LOW      PORTB &= ~(1<<S88_CLK)

#define READ_S88_D1      (PINA & (1<<S88DIN1))
#define READ_S88_D2      (PINA & (1<<S88DIN2))
#define READ_S88_D3      (PINA & (1<<S88DIN3))



#endif   // hardware.h
