//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006, 2007 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      ibox_parser.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2006-09-30 V0.01 started
//            2006-10-13 V0.02 Parser runs first tests
//            2007-01-12 V0.03 Startmode is read from eeprom
//            2007-01-16 V0.04 implemented break detection
//            2007-01-17 V0.05 bugfix in P50X-Parser
//            2007-01-22 V0.06 BAUD_4800 dazu.
//            2007-01-26 V0.07 Progmode dazu
//            2007-02-01 V0.08 Registermode dazu
//            2007-03-15 V0.09 POM dazu
//            2007-03-25 V0.10 Stop und Go der 6050 Syntax dazu
//            2007-06-08 V0.11 Bugfix in change baud (TXC was read on wrong reg)
//            2007-07-08 V0.12 Bugfix in enter_feedback to s88 (L952):
//                             (turnout addr was given instead coil addr)
//            2007-07-14 V0.13 Bugfix: no string was answered with zza1
//            2007-12-27 V0.15 Bugfix in enter_feedback to s88 (L952):
//                             (wrong turnout addr (shifted by -4, broadcast-offset)
//                             now completely removed
//            2008-02-02 V0.16 feedback_size added
//            2008-07-15 V0.17 bugfix in char2hex (Meldung von Lukas Schrittwieser)
//            2008-08-04 V0.18 Changed return values from organizer
//                             Bugfix in XPT_TERM
//                        
// tested:    Parser with loco and turnout commands
//          
// Limitations compared to IBOX:
//            - Only commands used for PC operation are implemented,
//              OpenDCC does not support:
//               * S88 timer nor counter (are these functions of any use?)
//               * virtual loco addresses
//               * multi control
//            - Turnout Status is not stored inside OpenDCC after
//              power down - turnouts are reported as 0 initially
//            - Turnoutaddr: IBOX is valid from 1 .. to 2047
//              we internally use 0 .. 4095
//            - P50 commands not implemented (except 0xc4 ;-)
//            - S88 limited to 2040 events
//
// Known issues and 2do:
//            - BiDi reserved commands are commented
//            - PROGMODE only partially tested
//            - Broadcasts not implemented
//            - Prog ASCII Modes for Register and Page are missing
//            - POM Ascii Modes to test - XA_PoM_Acc, XA_PoM_Lok
//            - XPTRL      damit liest die IB die CV 17 und 18, wandelt die darin
//              enthaltenen Werte in die sogenannte 'Lange Adresse" um und gibt sie dezimal aus
//                Beispiel: 4711
//
// XPTWL  
//              
//-----------------------------------------------------------------
//
// #define PARSER_TEST        // if defined, add some routines to test parser

// purpose:   lowcost central station for dcc
// content:   reads pc-commands from rs232 and generates calls
//            to organizer.c
// sections:  1. Tools (converter)
//            2. Parser
//
//
// interface upstream:
//            init_parser(void)         // set up the queue structures
//            run_parser(void)          // multitask replacement, must be called
//                                      // every 20ms (approx)
//
// interface downstream:
//            rx_fifo_read()            // to rs232
//            rx_fifo_ready()
//            tx_fifo_write()
//            tx_fifo_full()
//
//            pc_send_single_rm(addr);  // to s88
//
//                                      // to organizer
//
//-----------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <ctype.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <util/delay.h>
#include <string.h>

#include "config.h"                // general structures and definitions
#include "dccout.h"
#include "status.h"                // timeout engine
#include "organizer.h"
#include "rs232.h"
#include "s88.h"



#if (PARSER == INTELLIBOX)

#include "ibox_programmer.h"            // service mode
#include "ibox_parser.h"

//-----------------------------------------------------------------------------------
// Interface fr events, die in der Zentrale passieren und an den PC gemeldet werden.
// Dies besteht aus 2 Typen:
// a) Zustandsnderungen (z.B. Tastendruck oder Kurzschlu)
// b) Rckmeldungen
//
// a) Zustandsnderungen

unsigned char power_event = 0;      // is set, whenever a power off had occured
                                    // is cleared if read by host
// unsigned char prog_event = 0;       // is set, whenever a programming is terminated
                                    // is cleared if read by host

void event_send(t_BC_message mess)
  {
    switch(mess)
          {
            case ALL_OFF:
                    // Abgeschaltet
                power_event = 1;
                break;
            case ALL_ON:
                // eingeschaltet
                break;
            case ALL_LOCO_OFF:
                // Halt
                break;
            case PROGMODE:
                // Programmiermode
                break;
          }
  }


// b) Rckmeldungen  (ausgelagert nach s88.c)

// void pc_send_rueckmeldung(void)             // broadcast: rckmeldung siehe S88.c
//
// Note: IB macht keine Broadcasts, hier soll der Host immer XStatus und Xevent pollen

// Alle Broadcast messages aus status werden hier abgesendet.



//----------------------------------------------------------------------------
//
// Variablen und Tools fr den IBOX-Parser
//
// Im Gegensatz zu sonstigen C Gepflogenheiten halten wir den Pointer auf pcc
// global; damit kann der String mit einem Pointer durchsucht werden.
// 


unsigned char pcc[64];          // pc_message speicher

unsigned char pcc_size;         // if known: size of current command (for p50x_binary)
unsigned char pcc_index;        // current write or read index

unsigned char *ibcmd;           // globaler Pointer auf commandstring

// Variablen des Parsers:

unsigned char P50X_mode=0;      // if 1: leading X not needed for p50X-Commands
                                // this is loaded from eeprom at startup

//------------------------------------------------------------------------------
// Hilfsrountinen:

signed char char2hex (unsigned char *arg)
  {
    if ((*arg >= '0') && (*arg <= '9')) return (*arg - '0');
    if ((*arg >= 'A') && (*arg <= 'F')) return (*arg - 'A' + 10);
    if ((*arg >= 'a') && (*arg <= 'f')) return (*arg - 'a' + 10);
    return (-1);  // no hex
  }

signed char char2bin (unsigned char *arg)
  {
    if (*arg == '0') return(0);
    if (*arg == '1') return(1);
    return (-1);  // no bin
  }

/*
// Standard C Konventions (not used in OPENDCC)
// axtoi bricht beim ersten non-hex-Wert ab

unsigned int axtoi(char *hexStg)
  {
    unsigned char n = 0;            // position in string
    unsigned int intValue = 0;      // integer value of hex string
    signed char temp;               // convertresult
    

    while (n < 16)                 // security limit against wrong string    
      {
        if ((temp=char2hex(hexStg[n++])) < 0) return(intValue);
        intValue = (intValue << 4) + temp;
      }
    return (intValue);
  }

unsigned int abtoi(char *binStg)
  {
    unsigned char n = 0;            // position in string
    unsigned int intValue = 0;      // integer value of bin string
    signed char temp;               // convertresult
    

    while (n < 32)                 // security limit against wrong string    
      {
        if ((temp=char2bin(binStg[n++])) < 0) return(intValue);
        intValue = (intValue << 1) + temp;
      }
    return (intValue);
  }
*/

//-------------------------------------------------------------
// Diese Funktionen nochmal, jetzt als Spezialversion mit
// globalen Variablen; Der pcc_index bleibt auf dem ersten falschen
// Zeichen stehen.

// axtoi bricht beim ersten non-hex-Wert ab

unsigned int get_cmdvalx(void)
  {
    unsigned int intValue = 0;      // integer value of hex string
    signed char temp;               // convertresult
    

    while (*ibcmd != 0)                 // security limit against wrong string    
      {
        if ((temp=char2hex(ibcmd)) < 0) return(intValue);
        intValue = (intValue << 4) + temp;
        ibcmd++;
      }
    return (intValue);
  }

unsigned int get_cmdvalb(void)
  {
    unsigned int intValue = 0;      // integer value of bin string
    signed char temp;               // convertresult
    

    while (*ibcmd != 0)             
      {
        if ((temp=char2bin(ibcmd)) < 0) return(intValue);
        intValue = (intValue << 1) + temp;
        ibcmd++;
      }
    return (intValue);
  }
            

// the following routine reads one arguement from commandline
// parameter:   pointer to int, where to put the result
//              parsing rules:
//              leading %: the number is parsed as binary
//              leading $: the number is parsed as hex
//              char r and g are parsed as 0 and 1
// return:      -1: EOL reached
//              -2: no valid arg, syntax error
//              0:  no parameter found, *result will be 0xffff;
//              1:  there is an arg, value is returned in *result
//
// side effect: ibcmd steht auf dem ersten Zeichen nach Delimiter
//
int get_px50arg(unsigned int *result)
  {
    unsigned int argValue = 0;      // integer value of hex string
    
    while (*ibcmd == 0x20) ibcmd++;  // skip leading blanks
    
    if (*ibcmd == 0)
      {
        return(-1);  // empty arg, end of line!
      }
    if (*ibcmd == ',')
      {
        ibcmd++;
        *result = 0xffff;
        return(0);  // empty arg 
      }
    if (*ibcmd == '$')
      {
        ibcmd++;
        argValue = get_cmdvalx();
        if (*ibcmd == ',') ibcmd++;   // skip trailing ',' 
      }
    else if (*ibcmd == '%')
      {
        ibcmd++;
        argValue = get_cmdvalb();
        if (*ibcmd == ',') ibcmd++;   // skip trailing ','  
      }
    else if (*ibcmd == 'R')
      {
        ibcmd++;
        argValue = 0;
        if (*ibcmd == ',') ibcmd++;   // skip trailing ','  
      }
    else if (*ibcmd == 'G')
      {
        ibcmd++;
        argValue = 1;
        if (*ibcmd == ',') ibcmd++;   // skip trailing ','  
      }
    else if ((*ibcmd >= '0') && (*ibcmd <= '9'))
      {
        argValue = atoi((char *)ibcmd);
        while ((*ibcmd >= '0') && 
               (*ibcmd <= '9'))  ibcmd++;  // skip numbers

        if (*ibcmd == ',') ibcmd++;   // skip ',' 
      }
    else
          {
            return(-2);                    // void arg
          }

    *result = argValue;
    return(1);
  }



//-------------------------------------------------------------------------------


unsigned char lok_addr_out_of_range(unsigned int *addr)       // Hilfsroutine: range der Lok Addr prfen
  {
    if ((*addr >= 1) && (*addr <= 10239)) return(0);
    else return(1);
  }

unsigned char scan_locobuffer(unsigned int *addr)       // Hilfsroutine: locobuffer durchsuchen
  {
    register unsigned char i;
    
    // im locobuffer nach dieser Lok suchen
    
    for (i=0; i<SIZE_LOCOBUFFER; i++)
      {
        if (locobuffer[i].address == *addr)
          {
            return(i);
          }
      }
    return (SIZE_LOCOBUFFER);   // max if not found
  }


//---------------------------------------------------------------------------------
// Senderoutinen zum PC

void pc_send(unsigned char out)                     // send single char
  {
    while (!tx_fifo_ready()) ;                      // busy waiting!!!
                                                    // vorlufig kein Timeout
    tx_fifo_write(out);                             // send character
  }


void pc_send_string_P(const char *progmem_string)   // send string in PROGMEM
  {
    char c;

    /* we need to read each byte from program memory */
    while ( (c = pgm_read_byte(progmem_string++)) )
      {
        pc_send(c);
      }
  }



void pc_send_answer_P(const char *progmem_string)   // send string in PROGMEM + cr ]
  {
    char c;

    /* we need to read each byte from program memory */
    while ( (c = pgm_read_byte(progmem_string++)) )
      {
        pc_send(c);
      }
    pc_send(0x0D);
    pc_send(']');
  }


void pc_send_num(unsigned int num)                  // send an int + leading blank
  {
    char buffer[6];         // max. 5 digits
    unsigned char i=0;

    

    itoa(num, buffer, 10);         // 10 byte long

    pc_send(' ');
    while (buffer[i])
      {
        pc_send(buffer[i++]);
      }
  }

void pc_send_int(unsigned int num)                  // send an int
  {
    char buffer[6];         // max. 5 digits
    unsigned char i=0;

    itoa(num, buffer, 10);         // 10 byte long

    while (buffer[i])
      {
        pc_send(buffer[i++]);
      }
  }


const char hexdig[] PROGMEM = "0123456789ABCDEF";

void pc_send_hex(unsigned char value)                  // send a char as hex, leading 0
  {
    unsigned char i=0;

    i = value >> 4;
    i &= 0x0f;
    pc_send(pgm_read_byte(hexdig[i]));

    i = value & 0x0f;
    pc_send(pgm_read_byte(hexdig[i]));
  }

void pc_send_bin(unsigned char value)                  // send a char as bin, leading 0
  {
    unsigned char i=0;

    for (i=0; i<8; i++)
      {
        if (value & 0x80)    pc_send('1');
        else                 pc_send('0');

        value = value << 1;
      }
  }


//===============================================================================
//
//  2. Parser
//     a) binary parser (p50xb)
//     b) ascii parser (p50xa)
//     c) 6050 parser (p50)
// 
//-------------------------------------------------------------------------------
//
//  There is a function for each p50x command, these functions are called by means
//  of a parser table. There are two table, one for binary command and one for ascii
//  commands
//
//=====================================================================================
//
//  2. a) BINARY - PART
//    
//------------------------------------------------------------------------------------------
// parse_p50x_binary
//  
#define  XMsg_OK      0x00      // - kein Fehler, Befehl korrekt ausgefhrt
#define  XMsg_SYERR   0x01      // not implemented
#define  XMsg_BADPRM  0x02      // Parameterfehler (Parameterbyte auerhalb der zulssigen Werte)
#define  XMsg_PWOFF   0x06      // Befehl nicht ausgefhrt, da OpenDCC im Modus STOP
#define  XMsg_NOTSPC  0x07      // Weichen-Queue ist voll, es knnen momentan keine weiterern
                                // Weichenbefehle angenommen werden.
#define  XMsg_NOLSPC  0x08      // Kein Platz im Kommandobuffer, Kommando spter wiederholen
#define  XMsg_NODATA  0x0A      // Lok ist nicht im Refresh-Buffer, daher kein Status verfgbar
#define  XMsg_NOSLOT  0x0B      // Kein PLatz in der Refresh-Queue (Slot)
#define  XMsg_LKBUSY  0x0D      // Lok wird bereits von einer anderen Unit kontrolliert
#define  XMsg_BADTNP  0x0E      // Die Weichenadresse ist fr das eingestellte Protokoll ungltig
#define  XMsg_BADSOV  0x0F      // bad SO value
#define  XMsg_LKHALT  0x41      // Lok-Kommando akzeptiert, aber nicht ausgefhrt, da Steuerung
                                // im HALT-Modus ist. Die Geschwindigkeit wird auf 0 gesetzt.
#define  XMsg_LKPOFF  0x42      // Lok-Kommando akzeptiert, aber nicht ausgefhrt, da Steuerung
                                // im STOP-Modus ist. Die Geschwindigkeit wird auf 0 gesetzt.



void XError(void)
  {
    pc_send(XMsg_BADPRM);
  }

//-------------------------------------------------------------------------------
// 0x80; XLOK (5)
// Parameters   1           2           3       4
// byte:        addr low    addr high   speed   func
//
// comment:     speed:  (0..127: 0 = Stop, 1 = Stop/Em.Stop) Speed is given as 0..127 
//                      must be converted internally to local speed.
//              func:
//              bit#       7     6     5     4     3     2     1     0
//                      +-----+-----+-----+-----+-----+-----+-----+-----+
//                      |ChgF |Force| Dir | FL  | F4  | F3  | F2  | F1  |
//                      +-----+-----+-----+-----+-----+-----+-----+-----+
// 
//              where:
//                  ChgF    set if F1..F4 to be used for setting F1..F4 of
//                          Lok (otherwise F1..F4 are ignored)
//                  Force   if set (1), then the XLok command is 'forced'
//                          even in case of a Lok already controlled by a
//                          non-PC device
//                  Dir     Lok direction: 1 = forward, 0 = reverse
//                  FL      Lok light status: 1 = on, 0 = off
//                  F4..F1  Lok F4..F1 status (if ChgF is set)
// 
// Reply:       00h (cmd Ok)
//              XBADPRM (02h)   illegal parameter value
//              XNOLSPC (08h)   there is no space in the Lok cmd buffer, please try later!
//              XNOSLOT (0Bh)   there is no slot available
//              XBADLNP (0Ch)   Lok# is illegal for this protocol
//              XLKBUSY (0Dh)   Lok already controlled by another device
//              XLKHALT (41h)   Command accepted (Lok status updated), but IB in 'Halt' mode!
//              XLkPOFF (42h)   Command accepted (Lok status updated), but IB in Power Off!


void XLok(void)
  {
    unsigned int *addr_ptr;
    unsigned char speed;
    unsigned char retval = 0;
    
    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    if (lok_addr_out_of_range(addr_ptr))
      {
        pc_send(XMsg_BADPRM); return;
      }

    speed = (pcc[4] & 0x7F);
    if (pcc[5] & 0x20) speed |= 0x80;
    
    if (organizer_ready())
      {
        // enter speed for this loco
        retval = do_loco_speed(*addr_ptr, speed);
        retval |= do_loco_func_grp0(*addr_ptr, pcc[5]>>4);  // Licht eintragen
        if (pcc[5] & 0x80)
          {
            // update functions 1..4 
            retval |= do_loco_func_grp1(*addr_ptr, pcc[5]);
          }
      }

    if (retval & (1 << ORGZ_FULL)) 
      {
        pc_send(XMsg_NOLSPC); //  XNOLSPC (08h) there is no space in the Lok cmd buffer
      }
    else
      {
        pc_send(XMsg_OK);  // !!! Missing: RUN-MODE abfragen und mit zurckliefern!
                           // Power off???
      }
  }
  
//-------------------------------------------------------------------------------
// 0x83; XLkDisp (3)
// Parameters   1           2          
// byte:        addr low    addr high   
//
// comment:     addr high:  if 0xff: ask for dispatch status
//                          must be converted internally to local speed.
// Reply        (if Lok address high byte not equal 0FFh, i.e. 'dispatch put'):
//              1st     either 00h (cmd Ok) or error code (e.g. XBADLNP, XNODATA)
//              or (if Lok address high byte equal 0FFh, i.e. 'dispatch get'):
//              1st     either 00h (there is no dispatched slot to 'get') or slot #
//
// --->>>>> not supported -> always XMsg_OK


void XLkDisp(void)
  {
    pc_send(XMsg_OK);
  }

//-------------------------------------------------------------------------------
// 0x84; XLokSts (3)
// Parameters   1           2          
// byte:            addr low    addr high   
//
// Reply        1st     either 00h (cmd Ok, 3 bytes shall follow) or error code
//              2nd     Speed (0..127: 0 = Stop, 1 = Stop/Em.Stop)
//              3rd     same format as 4th parameter of XLok cmd (bit #6 & 7 = 0)
//              4th     'real' Lok speed (in terms of the Lok type/configuration)
//              see IB Doc files for 'real speed'      
//
// Error codes:  XBADPRM (02h)  illegal parameter value
//               XNODATA (0Ah)  no Lok status available (Lok is not in a slot)
//               XBADLNP (0Ch)  Lok# is illegal for this protocol
//
// not supported -> always 0x00


void XLokSts(void)
  {
    unsigned char myfunc, i;
    unsigned int *addr_ptr;
    
    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    if (lok_addr_out_of_range(addr_ptr))
      {
        pc_send(XMsg_BADPRM); return;
      }
    
    i = scan_locobuffer(addr_ptr);

    if (i==SIZE_LOCOBUFFER)            // not found - was not used yet
      {
        pc_send(0x0a);
      }
    else
      {
        pc_send(XMsg_OK);
        pc_send(locobuffer[i].speed & 0x7F);
        myfunc = locobuffer[i].speed & 0x80;
        myfunc = myfunc >> 2;
        myfunc |= locobuffer[i].fl << 4;
        myfunc |= locobuffer[i].f4_f1;
        pc_send(myfunc);
        pc_send(convert_speed_to_rail(locobuffer[i].speed, locobuffer[i].format) );
      }
  }

//-------------------------------------------------------------------------------
// 0x85; XLokCfg (3)
// Parameters   1           2          
// byte:        addr low    addr high   
//
// Reply:       1st     either 00h (cmd Ok, 4 bytes shall follow) or error code
//              2nd     Lok protocol: 0..3 = Mrk, SX, DCC, FMZ
//              3rd     # of (non-Stop) speed steps supported by this Lok type.
//                      Currently this would be: 14, 15, 27, 28, 31 or 126
//                      In case of OpenDCC: 14, 28 or 126
//
//              The rest of the reply depends on whether this Lok is a virtual
//              one or a real one:
//              - Virtual loco: 4th: Real Lok# (low)
//                              5th: Real Lok# (high) corresponding to this virtual one
//                 ->>> not supported in OpenDCC
//              - Real loco: 4th and 5th:  0FFh

void XLokCfg(void)
  {
    unsigned char myformat, i = 28;
    unsigned int *addr_ptr;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    if (lok_addr_out_of_range(addr_ptr))
      {
        pc_send(XMsg_BADPRM); return;
      }

    myformat = get_loco_format(*addr_ptr);
    
    pc_send(XMsg_OK);
    pc_send(0x02);
    switch(myformat)
      {
        case DCC14:
                i = 14;
                    break;
        case DCC27:
                i = 27;
                break;
        case DCC28:
                i = 28;
                break;
        case DCC128:
                i = 126;
                break;
      }
    pc_send(i);
    pc_send(0xff);
    pc_send(0xff);
  }

//-------------------------------------------------------------------------------
// This command is not defined in IB!  --> Extension of OpenDCC
//
// 0x86; XLokCfgSet (5)
// Parameters   1           2                   3               4  
// byte:            addr low    addr high   Lok Proto   Speed Steps
//
//              Proto: like in XLokCfg (only DCC is supported)
//              Speed Steps: 14, 28, 126
//
// Reply:               1st     either 00h or error code BAD Param
//

void XLokCfgSet(void)
  {
    unsigned char myformat, valid = 0;
    unsigned int *addr_ptr;
    
    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    if (lok_addr_out_of_range(addr_ptr))
      {
        pc_send(XMsg_BADPRM); return;
      }
    
    if (pcc[4] == 2)
          {
            switch (pcc[5])
          {
            case 14:
                myformat = DCC14;
                valid = 1;
                break;
            case 28:
                myformat = DCC28;
                valid = 1;
                break;
            case 126:
                myformat = DCC128;
                valid = 1;
                break;
          }
      }
    if (valid)
      {
        if (store_loco_format(*addr_ptr, myformat))
          {
            pc_send(XMsg_OK);
          }
        else pc_send(XMsg_NOSLOT);  // failed - to many locos with extra format
      }
    else
      {
          pc_send(XMsg_BADPRM);  // Bad Parameter
      }
  }

//-------------------------------------------------------------------------------
// 0x88; XFunc (4)
// Parameters   1           2           3
// byte:        addr low    addr high   Status
//
//              Status: F1 (bit #0) .. F8 (bit #7)
//
// Reply:       1st     either 00h (cmd Ok) or error code
//
//              Error codes:
//              XBADPRM (02h)   illegal parameter value
//              XNOSLOT (0Bh)   there is no slot available
//              XBADLNP (0Ch)   Lok# is illegal for this protocol

void XFunc(void)
  {
    unsigned int *addr_ptr;
    unsigned char retval;
    
    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    if (lok_addr_out_of_range(addr_ptr))
      {
        pc_send(XMsg_BADPRM); return;
      }

    if (organizer_ready())
      {
        retval = do_loco_func_grp1(*addr_ptr, pcc[4]);
        retval |= do_loco_func_grp2(*addr_ptr, pcc[4]>>4);
        pc_send(XMsg_OK);
      }
    else
      {
        pc_send(XMsg_NOSLOT);
      }
  }

//-------------------------------------------------------------------------------
// 0x89; XFuncX (4)
// Parameters   1           2           3
// byte:        addr low    addr high   Status
//
//              Status: F9 (bit #0) .. F16 (bit #7)
//
// Reply:       1st     either 00h (cmd Ok) or error code
// Note:        F13-F16 not supported!!!
//
//              Error codes:
//                  XBADPRM (02h)   illegal parameter value
//                  XNOSLOT (0Bh)   there is no slot available
//                  XBADLNP (0Ch)   Lok# is illegal for this protocol

void XFuncX(void)
  {
    unsigned int *addr_ptr;
    unsigned char retval;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    if (lok_addr_out_of_range(addr_ptr))
      {
        pc_send(XMsg_BADPRM); return;
      }

    if (organizer_ready())
      {
        retval = do_loco_func_grp3(*addr_ptr, pcc[4] );
        pc_send(XMsg_OK);
      }
    else
        pc_send(XMsg_NOSLOT);
  }

//-------------------------------------------------------------------------------
// 0x8C; XFuncSts (3)
// Parameters   1           2           
// byte:        addr low    addr high   
//
//  
// Reply:       1st     either 00h (cmd Ok, 1byte shall follow) or error code
//              Status: F1 (bit #0) .. F8 (bit #7)
//              Error codes:
//              XBADPRM (02h)       illegal parameter value
//              XNODATA (0Ah)   no Lok status available (Lok is not in a slot)
//              XBADLNP (0Ch)   Lok# is illegal for this protocol

void XFuncSts(void)
  {
    unsigned char i, myfunc = 0;
    unsigned int *addr_ptr;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    if (lok_addr_out_of_range(addr_ptr))
      {
        pc_send(XMsg_BADPRM); return;
      }
    
    i = scan_locobuffer(addr_ptr);

    if (i==SIZE_LOCOBUFFER)            // not found - was not used yet
      {
        pc_send(XMsg_NODATA);
      }
    else
      {
        pc_send(XMsg_OK);
        myfunc = locobuffer[i].f4_f1;
        myfunc |= locobuffer[i].f8_f5<<4;
        pc_send(myfunc);
      }
  }

//-------------------------------------------------------------------------------
// 0x8D; XFuncXSts (3)
// Parameters   1           2           
// byte:        addr low    addr high   
//
//  
// Reply:       1st     either 00h (cmd Ok, 1byte shall follow) or error code
//              Status: F9 (bit #0) .. F16 (bit #7)
//
// Note:        F13..F16 not supported
//
//              Error codes:
//              XBADPRM (02h)   illegal parameter value
//              XNODATA (0Ah)   no Lok status available (Lok is not in a slot)
//              XBADLNP (0Ch)   Lok# is illegal for this protocol

void XFuncXSts(void)
  {
    unsigned char i, myfunc = 0;
    unsigned int *addr_ptr;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    if (lok_addr_out_of_range(addr_ptr))
      {
        pc_send(XMsg_BADPRM); return;
      }
    
    i = scan_locobuffer(addr_ptr);

    if (i==SIZE_LOCOBUFFER)            // not found - was not used yet
      {
        pc_send(0x0a);
      }
    else
      {
        pc_send(XMsg_OK);
        myfunc = locobuffer[i].f12_f9;
        pc_send(myfunc);
      }
  }

//-------------------------------------------------------------------------------
// 0x90; XTrnt (3)
// Parameters   1           2           
// byte:        addr low    addr high + 'color' and status bits  
//
// 
//              bit#   7     6     5     4     3     2     1     0
//                  +-----+-----+-----+-----+-----+-----+-----+-----+
//                  |Color| Sts | Res |NoCmd| n.u.| A10 |  A9 |  A8 |
//                  +-----+-----+-----+-----+-----+-----+-----+-----+
// 
//              where:
//                  A10..A0: Addr of Turnout -> max. 2048 This is a
//                           limitation compared to NMRA
//                           Range: 1..2047!!
//                  Color:   1 = closed (green), 0 = thrown (red)
//                  Sts:     turnout status (1 = on, 0 = off)
//                  Res:     set if this turnout is to be reserved
//                  NoCmd:   if set then no turnout cmd is actually sent
//                           to the tracks. 
//                  n.u.        not currently used (0) -> may be A11?
// 
//  
// Reply:       1st     either 00h (cmd Ok) or error code
//              Status: F9 (bit #0) .. F16 (bit #7)
//
// Note:        Res and NoCmd are ignored and not supported !
//
//              Error codes:
//              XBADPRM (02h)       illegal parameter value
//              XPWOFF (06h)        Error: the Power is Off!
//              XNOTSPC (09h)       Error: the fifo for turnout cmds is full, please try later
//              XBADTNP (0Eh)       Error: illegal Turnout address for this protocol
//              XNOISPC (10h)       Error: the I2C fifo is full, please try later
//              XLOWTSP (40h)       Warning: the fifo for turnout cmds is 75% full
//
// Mapping:     Trntout 1 maps to decoder 1, output 0 (to be compatible to the IB)
//              In my opinion this is a severe error of the IB. 

void XTrnt(void)
  {
    unsigned char output, activate;
    unsigned int *addr_ptr;
    unsigned int trnt_addr;
    
    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    trnt_addr = ((*addr_ptr) & 0x0FFF) - 1;    // range down to 0...4094
    output = (pcc[3]>>7) & 0x01;
    activate = (pcc[3]>>6) & 0x01;
                                                // feedback entry deleted, 27.12.07
    
    if (do_accessory(trnt_addr, output, activate) )  
      {
        if (power_event) pc_send(XMsg_PWOFF);
        else pc_send(XMsg_OK);
      }
    else
      {
        pc_send(0x40);
      }
  }

//-------------------------------------------------------------------------------
// 0x93; XTrntFree (1)
//
// Allows resetting the 'Res' (reserved) bit of all turntouts with only one command.
// This command is not supported by OpenDCC
// Parameters   none
//  
// Reply:               00h 

void XTrntFree(void)
  {
        pc_send(XMsg_OK);
  }


//-------------------------------------------------------------------------------
// 0x94; XTrntSts (3)
// Parameters   1           2           
// byte:            addr low    addr high 
//  
// Reply:               1st     either 00h (cmd Ok, 1 byte shall follow) or error code
//              2nd    7     6     5     4     3     2     1     0
//                  +-----+-----+-----+-----+-----+-----+-----+-----+
//                  |n.u. |n.u. |n.u. |n.u. |Conf1|color| Res |Conf0|
//                  +-----+-----+-----+-----+-----+-----+-----+-----+
//
//              Conf0/1:  00              Motorola
//                        10              DCC
//                        01              SX
//                        11              FMZ
// Note:        Command not supported - allways DCC, red, not reserved
//              A Storage in EEPROM would be required
//
//              Error codes:
//              XBADPRM (02h)       illegal parameter value
//              XBADTNP (0Eh)       Error: illegal Turnout address for this protocol

void XTrntSts (void)
  {
    unsigned int *addr_ptr;
    unsigned char position;

    addr_ptr = (unsigned int *) &pcc[2];    // use pcc[2] + pcc[3] as int
    pc_send(XMsg_OK);
    position = recall_turnout(*addr_ptr -1) << 2;
    pc_send(0x01 | position);               // format is always DCC
  }

//-------------------------------------------------------------------------------
// 0x95; XTrntGrp (2)
// Parameters   1       
// byte:        addr (of 'group' address = ((turnout address - 1) / 8) + 1 )
//  
// Reply:       1st  either 00h (cmd Ok, 2 bytes shall follow) or error code
//              2nd: Group color
//                   Richtung der Weiche (1 Bit pro Weiche, 1 = gerade/grn,
//                   0 = abbiegen/rot). Weiche #1..8 entspricht Bit #0..7.
//              3rd: group reserved not supported - always 0
//
// Note:        Command partially supported - depends on SIZE_TURNOUTBUFFER
//
//              Error codes:
//                  XBADPRM (02h)       illegal parameter value

void XTrntGrp (void)
  {
    unsigned char positions;
    pc_send(XMsg_OK);
    positions = recall_turnout_group(pcc[2]);
    pc_send(positions);
    pc_send(0x00);         // opendcc has no reserved turnouts - always 0
  }

//-------------------------------------------------------------------------------
// 0x98; XSensor (2)
// Parameters   1       
// byte:            S88 module number (1..128)
//  
// Reply:       1st     either 00h (cmd Ok, 2 bytes shall follow) or error code
//              2nd     contacts #1..8 (bits #7..0)
//              3rd     contacts #9..16 (bits #7..0)
//
// Note:        original IB reads only actual state, we read accumulated state
//              and clear change flags.
//
//              Error codes:
//                  XBADPRM (02h)       illegal parameter value

void XSensor (void)
  {
    if ((pcc[2]*2) > s88_total)
      {
        // Error, number too high
        pc_send(XMsg_BADPRM);
      }
    else
      {
        make_s88_ib_single_report(pcc[2]-1);   // local count starts with 0

        pc_send(0);                 
        pc_send(s88_ib_report[1]);
        pc_send(s88_ib_report[2]);
      }
  }

//-------------------------------------------------------------------------------
// 0x99; XSensOff (1)
//
// This cmd tells the IB to report as new 'sensor events' the status of
// all sensors which aren't 'off'.
// Clear all s88_data to zero and s88_change.
//
// Parameters   none      
//  
// Reply:               1st     always 00h 
//
// Note:        used by TC

void XSensOff (void)
  {
    sense_off_s88();
    pc_send(0);
  }

//-------------------------------------------------------------------------------
// 0x9C; X88PGet (2)
//
// This cmd gets S88 Config Parameters
//
// Parameters   1       
// byte:            S88 Parameter number
//               00h    # of automatically read s88 'half modules'
//                      (check also P50Xa 'SE' cmd)
//               01h    s88 module 'half number' which is source for the s88 timers
//               02h    s88 module 'half number' which is source for the s88 counters
//  
// Reply:       either 00h (cmd Ok, one byte shall follow) or error code.
//              Eventual 2nd byte holds the s88 parameter value 
//
// Note:        01 and 02 not supported -> we do no timers nor counters 
//                              these commands are replaced by:
//              01: get S88 half moduls on string 1
//              02: get S88 half moduls on string 2
//              03: get S88 half moduls on string 3
//
//              command with param 0 is used by TC
//              according to Docu: IB has a default of 16 - actual reported are 17.

void X88PGet (void)
  {
    pc_send(0);
    switch(pcc[2])
      {
        default:
        case 0:    pc_send(s88_total);   // should be 0 ..62
                                         // we report actual values
                   break;
        case 1:    pc_send(s88_size1);   // should be 0 ..62
                   break;
        case 2:    pc_send(s88_size2);   // should be 0 ..62
                   break;
        case 3:    pc_send(s88_size3);   // should be 0 ..62
                   break;
        case 4:    pc_send(eeprom_read_byte((void *)eadr_feedback_s88_size));   
                   break;
      }
  }

//-------------------------------------------------------------------------------
// 0x9D; X88PSet (3)
//
// This cmd sets S88 Config Parameters
//
// Parameters   1       
// byte:        S88 Parameter number
//               00h    # of automatically read s88 'half modules'
//                      (check also P50Xa 'SE' cmd)
//               01h    s88 module 'half number' which is source for the s88 timers
//               02h    s88 module 'half number' which is source for the s88 counters
// byte:        Value
//  
// Reply:               either 00h (cmd Ok) or error code.
//              Eventual 2nd byte holds the s88 parameter value 
//
// Note:        00: is ignored
//              01 and 02 not supported -> we do no timers nor counters 
//                              these commands are replaced by:
//              01: set S88 half moduls on string 1
//              02: set S88 half moduls on string 2
//              03: set S88 half moduls on string 3
//
//              This is only temporary setting, use eeprom routines to store
//              permanently

void X88PSet(void)
  {
    switch(pcc[2])
      {
        default:
        case 0: // do not use it for string 1
                // ingore this setting :-)   Sizes should be set by params 1,2 and 3
                // but keep the setting and report it, if asked.
                // eeprom_write_byte((void *)eadr_s88_size1,pcc[3]);   // should be 0 ..62
                eeprom_write_byte((void *)eadr_s88_total_from_pc,pcc[3]);
                break;   
        case 1: eeprom_write_byte((void *)eadr_s88_size1,pcc[3]);   // should be 0 ..62
                break;
        case 2: eeprom_write_byte((void *)eadr_s88_size2,pcc[3]);   // should be 0 ..62
                break;
        case 3: eeprom_write_byte((void *)eadr_s88_size3,pcc[3]);   // should be 0 ..62
                break;
        case 4: eeprom_write_byte((void *)eadr_feedback_s88_size,pcc[3]);   // should be 0 ..127
                break;
      }
    eeprom_busy_wait();
    if (init_s88(1))     pc_send(XMsg_OK);
    else                 pc_send(XMsg_BADPRM);
  }

//-------------------------------------------------------------------------------
// We dont support Timer nor counter !!
                            // *** Xs88Tim (09Eh) - length = 1+1 bytes
                            // Parameters (byte):
                            // 1st      Bit #0..3: s88 Timer number (1..8)
                            //  Bit #7 = set if Timer to be reset after reading (i.e. after this cmd)
                            //  (Bit #4..6 are reserved: must always be 0)
                            // Reply:
                            // 1st      00h (cmd Ok, 2 bytes shall follow) or error code
                            //  As usual, only in case of no error, more bytes are sent:
                            // 2nd      s88 Timer value (low byte)
                            // 3rd      s88 Timer value (high byte)
                            // The timer value is to be interpreted in 200 ms units.

                            // Provided the corresponding s88 input is 'on' (closed), each s88 Timer
                            // is incremented (i.e., Timer = Timer + 1) approximately every 200 ms.
                            // Upon eventually reaching the maximum value (0FFFFh), there is no
                            // 'wrap around': the timer stays at this maximum value.


                            // *** Xs88Cnt (09Fh) - length = 1+1 bytes
                            // Parameters (byte):
                            // 1st      Bit #0..3: s88 Counter number (1..8)
                            //  Bit #7 = set if Counter to be reset after reading (i.e. after this cmd)
                            //  (Bit #4..6 are reserved: must always be 0)
                            // Reply:
                            // 1st      00h (cmd Ok, 2 bytes shall follow) or error code
                            //  As usual, only in case of no error, more bytes are sent:
                            // 2nd      s88 Counter value (low byte)
                            // 3rd      s88 Counter value (high byte)

                            // The (time) resolution of the Counters depends on the number of s88
                            // modules which are being automatically read by the IB: the lower this
                            // number, the higher the resolution of s88 counters (i.e. they would be
                            // able to keep track also of very fast on/off cycles).
                            // A reasonable estimate, while automatically reading the first 8 s88
                            // modules, would be about 50 ms resolution for the Counters - or better.
                            // Upon eventually reaching the maximum value (0FFFFh), there is no
                            // 'wrap around': the counter stays at this maximum value.

//-------------------------------------------------------------------------------
// 0xa0; XVer (1)
// Parameters   none     
// Reply:       iterated format consisting of one 1st byte telling
//                  the amount of bytes which shall follow. If this 1st byte
//                  is zero, then this is the last byte of the XVer reply.
//                  For example, the IB replies with:
//                  02h, <SPU version low>, <SPU version high>,
//                  02h, <KPU version low>, <KPU version high>,
//                  01h, <PPU version>,
//                  01h, <LIPU version>,
//                  01h, <DNG version>,
//                  05h, <IB serial number: 5 bytes (digits 98, 76, 54, 32, 10)>,
//                  00h     
//
// Note:        with OpenDCC we report only two numbers: Version and serial 

void XVer (void)
  {
    pc_send(0x02);
    pc_send(eeprom_read_byte((void *)eadr_OpenDCC_Version));
    pc_send(eeprom_read_byte((void *)eadr_serial_id));
    pc_send(0x00);
  }

//-------------------------------------------------------------------------------
// 0xa1; XP50XCh (2)
// Parameters   1
//              new 'X' valid 0 or 0x50 to 0x5F
//     
// If 00h is specified (as parameter value), then the P50X protocol
// shall be disabled. Re-enable is possible only by PC break or Intellibox
// reset. The latter would only work provided that the Intellibox has been
// configured that way using its menus (i.e. configured for 'P50X enabled').
// 
// Reply:
// 1st  00h (cmd Ok) or error code
//
// Note:        not supported

void XP50XCh (void)
  {
    pc_send(0x00);
  }

//-------------------------------------------------------------------------------
// 0xa2; XStatus (1)
// Parameters   none
//     
// Reply:   bit#   7     6     5     4     3     2     1     0
//              +-----+-----+-----+-----+-----+-----+-----+-----+
//              |Sts2 |VReg |ExtCU|Halt | Pwr | Hot | Go  |Stop |
//              +-----+-----+-----+-----+-----+-----+-----+-----+
//
//      where:
//              Sts2    currently always reported as 0. If set, then
//                      one more byte is to be expected as part of
//                      the XStatus reply (this would be an eventual
//                      expansion for a future software release)
//                      This would be similar to what happens with
//                      the XEvent cmd.
//              VReg    set if voltage regulation (N scale) is enabled
//              ExtCU   set if an external I2C Central Unit is present
//              Halt    set if in Halt mode (Lok's stopped, Pwr On)
//              Pwr         set if we are in Power On
//              Hot         Overheating condition detected
//              Go          set if a [Go] key on an external I2C device
//                          is currently pressed
//              Stop    set if a [Stop] key on an external I2C device
//                          is currently pressed
//
// Note:        not supported

void XStatus (void)
  {
    unsigned char status_report = 0x40;  // OpenDCC always uses voltage regulation

    switch(opendcc_state)                   // this is the current running state
      {
        case RUN_OKAY:            
            status_report |= 0x08;
            break;
        case RUN_STOP:                      // DCC Running, all Engines Emergency Stop
            status_report |= 0x18;
            break;
        case RUN_OFF:                       // Output disabled (2*Taste, PC)
            break;
        case RUN_SHORT:                     // Kurzschluss, hektisch blinken im Wechsel
            break;
        case RUN_PAUSE:                     // DCC Running, all Engines Speed 0
            status_report |= 0x10;
            break;
        case PROG_OKAY:
            status_report |= 0x08;
            break;
        case PROG_SHORT:           
            break;
        case PROG_OFF:
            status_report |= 0x10;
            break;
        case PROG_ERROR:        
            status_report |= 0x08;
            break;
      }
    // check keys, no debouncing!
    if (KEY_STOP_PRESSED) status_report |= 0x01;
    if (KEY_GO_PRESSED) status_report |= 0x02;
    
    pc_send(status_report);
  }

//-------------------------------------------------------------------------------
// 0xa3; XSOSet (4)
// Parameters   1           2           3
// byte:            addr low    addr high   value
//
//              addr:   no of Spezial option
//              value:  unsigned char
//  
// Reply:               1st     either 00h (cmd Ok) or error code          
//
// Note:        This command is not supported by org. intellibox
//              OpenDCC uses it for configuring
//              Error codes:
//                  XBADPRM (02h)       illegal parameter value
//                  XSYERR  (01h)       Sytemerror
//              No rempapping of SO001 - Baudrate occurs

void XSOSet(void)
  {
    unsigned int *addr_ptr;
    
    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int

    if (*addr_ptr > 1024)  
      {
        pc_send(XMsg_BADPRM);
      }
    else
      {
        eeprom_write_byte((void *) *addr_ptr, pcc[4]);
        pc_send(XMsg_OK);
      }
  }

//-------------------------------------------------------------------------------
// 0xa4; XSOGet (3)
// Parameters   1           2           
// byte:        addr low    addr high   
//
//              addr:   no of Special option 0..999
//              value:  unsigned char
//  
// Reply:       1st     either 00h (cmd Ok) or error code          
//
// Note:        This command is not supported by org. intellibox
//              OpenDCC uses it for configuring
//              Error codes:
//                  XBADPRM (02h)       illegal parameter value
//                  XSYERR  (01h)       Sytemerror
//
// Notes:       Railware checks some of these and suggest optimizations :-)
//              TC reads SO-0001 (Baudrate) and sets then this baud rate
//              with P50Xa-Command 'b'
//
// Emulated Special Options
// 001      Intellibox default power-on Baudrate.
//              0 = 2400 bps
//              1 = 4800 bps
//              2 = 9600 bps
//              3 = 19200 bps
//              4 = 38400 bps
//

void XSOGet(void)
  {
    unsigned int *addr_ptr;
    unsigned char retval;
    
    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    
    if (*addr_ptr > 1024)  
      {
        pc_send(XMsg_BADPRM);
      }
    else
      {
        retval = eeprom_read_byte((void *) *addr_ptr);
        if (*addr_ptr == 1)
          {
            // remap SO from OpenDCC (=Lenz-Enum) to IBOX Enum
            switch(retval)
              {
                case BAUD_2400:   retval = 0; break;
                case BAUD_4800:   retval = 1; break;
                case BAUD_9600:   retval = 2; break;
                case BAUD_19200:  retval = 3; break;
                case BAUD_38400:  retval = 4; break;
                case BAUD_57600:  retval = 5; break; // wahrscheinlich??
                case BAUD_115200: retval = 6; break; // wahrscheinlich??                
              }
          }
        pc_send(XMsg_OK);
        pc_send(retval);  
      }
  }
  
//-------------------------------------------------------------------------------
// 0xa5; XHalt (1)
// Parameters   none
//     
// Reply:       00h (cmd Ok) or error code
//
// Note:        

void XHalt (void)
  {
    set_opendcc_state(RUN_STOP);
    pc_send(XMsg_OK); 
  }

//-------------------------------------------------------------------------------
// 0xa6; XPwrOff (1)
// Parameters   none
//     
// Reply:       00h (cmd Ok) or error code
//
//
// Note:        

void XPwrOff (void)
  {
    set_opendcc_state(RUN_OFF);
    pc_send(XMsg_OK); 
  }

//-------------------------------------------------------------------------------
// 0xa7; XPwrOn (1)
// Parameters   none
//     
// Reply:       00h (cmd Ok) or error code
//              error code XPWOFF (06h): the Power is Off!
//              XPWOFF is reported if an error condition (e.g. 'overheating')
//              prevents the Intellibox from turning on the power to the layout.
//
//
// Note:        This command will last a few ms, we have to check short
//              currently skipped !!!!

void XPwrOn (void)
  {
    power_event = 0;
    set_opendcc_state(RUN_OKAY);
    
    // !!! no_timeout.parser = (SHORT_TURNOFF_TIME / TICK_PERIOD) + 2;   // Timeout aufziehen

    // !!! while (no_timeout.parser) run_state();  // busy waiting, but run state engine
    
    if (power_event) pc_send(XMsg_PWOFF);    
    else pc_send(XMsg_OK); 
  }

//-------------------------------------------------------------------------------
// 0xC4; XNOP (1)
// Parameters   none
//     
// Reply:       00h (cmd Ok) 
//

void XNOP (void)
  {
    pc_send(XMsg_OK); 
  }

//-------------------------------------------------------------------------------
// 0xC6; XP50Len1 (2)
// Parameters   P50 Command
//     
// Reply:       none
//
// Note:        not implemented    

void XP50Len1 (void)
  {
  }

//-------------------------------------------------------------------------------
// 0xC7; XP50Len2 (3)
// Parameters   P50 Command (two bytes)
//     
// Reply:       none
//
// Note:        not implemented    

void XP50Len2 (void)
  {
  }

//-------------------------------------------------------------------------------
// 0xC8; XEvent (1)
//
// Reply:       1..n Bytes; if msb is set, a byte will follows
//
// 1st  event flags to be interpreted as:
//
//      bit#   7     6     5     4     3     2     1     0
//          +-----+-----+-----+-----+-----+-----+-----+-----+
//          | Ev2 |  x  |Trnt |TRes |PwOff| Sen | IR  | Lok |
//          +-----+-----+-----+-----+-----+-----+-----+-----+
//
//      where:
//          Ev2     set if also the 2nd byte of the XEvent
//                  reply shall be sent (i.e.: there is at least
//                  one event also in the 2nd or in the 3rd byte)
//          x       (reserved for future use)
//          Trnt    there has been at least one non-PC Turnout cmd
//          TRes    there has been at least one non-PC attempt at
//          changing the status of a 'reserved' Turnout
//          PwOff   there **has been** (not: is!) a Power Off
//          Sen      there has been at least one sensor event
//                        (s88 or LocoNet)
//          IR          there has been at least one infra-red event
//          Lok     there has been at least one non-PC Lok cmd
//
//
//2nd   event flags to be interpreted as:
//
//      bit#   7     6     5     4     3     2     1     0
//          +-----+-----+-----+-----+-----+-----+-----+-----+
//          | Ev3 | Sts | Hot |PTSh |RSSh |IntSh|LMSh |ExtSh|
//          +-----+-----+-----+-----+-----+-----+-----+-----+
//
//      where:
//              Ev3     set if also the 3rd byte of the XEvent reply
//                      shall be sent (i.e.: there is at least an
//                      event also in the 3rd byte)
//              Sts     an XStatus cmd should be issued
//              Hot     Overheating condition detected
//              PTSh    while the PT relay was on (e.g., PT in 'PT only'
//                      mode), a non-allowed electrical connection btw
//                      the Programming Track and the rest of the layout
//                      has been detected
//              RSSh    overload (short) on the DCC Booster C/D lines
//                      or on the LocoNet (B connector) Rail Sync +/-
//                      lines (or on the PT, if the PT relay was on)
//              IntSh   short reported by the internal Booster
//              LMSh    overload (short) on the Lokmaus bus
//              ExtSh   short reported by an external Booster
//
//
//3rd   event flags to be interpreted as:
//
//      bit#   7     6     5     4     3     2     1     0
//          +-----+-----+-----+-----+-----+-----+-----+-----+
//          | Ev4 |  x  |  x  |ExVlt|TkRel| Mem |RSOF |  PT |
//          +-----+-----+-----+-----+-----+-----+-----+-----+
//
//      where:
//              Ev4     not currently used, reported as 0
//              x       not currently used
//              ExVlt   an external voltage source is present
//                      (prior to turning on the layout). E.g.
//                      an external transformer is in contact
//                      with the rails.
//              TkRel   report Lok 'take' and 'release' events
//                      from non-PC Lok controllers
//                      (to be documented)
//              Mem     set if there has been at least one
//                      'memory' event (to be documented - related
//                      to the future IB 'memory' software expansion)
//              RSOF    set if an RS-232 rx overflow has been
//                      detected (the PC probably does not
//                      correctly handle the CTS line)
//              PT      a PT event is available.
//                      N.B. This bit must be 'cleared' by sending
//                           the XPT_Event cmd!

// Note: with OpenDCC there are only 3 events: S88, OFF, Programming

void XEvent (void)
  {
    unsigned char event1 = 0, event2 = 0, event3 = 0;
            
    if (s88_event & (1<<S88_HW_CHANGE)) event1 |= 0x04;
    if (power_event)
      {
        event1 |= 0x08;
        power_event = 0;
      }
    if (prog_event)
      {
        event3 |= 0x01;
      }

    // verkettung bauen
    if (event3 != 0) event2 |= 0x80;     // a event will follow
    if (event2 != 0) event1 |= 0x80;

    pc_send(event1);
    if ((event1 & 0x80) == 0) return;
    pc_send(event2);
    if ((event2 & 0x80) == 0) return;
    pc_send(event3);
  }

//-------------------------------------------------------------------------------
// 0xC9; XEvtLok (1)
// Parameters   none
//     
// Reply:       0x80
//
// Note:        not implemented  -> no Loks to report 

void XEvtLok (void)
  {
    pc_send(0x80);
  }


//-------------------------------------------------------------------------------
// 0xCA; XEvtTrn (1)
// Parameters   none
//     
// Reply:       0x00 or count of turnouts (max 64)
//              For each reported turnout event, two bytes are sent.
//                 1           2           
//              addr low    addr high + 'color' and status bits  
//
//              bit#   7     6     5     4     3     2     1     0
//                  +-----+-----+-----+-----+-----+-----+-----+-----+
//                  |Color| Sts | Res |NoCmd| n.u.| A10 |  A9 |  A8 |
//                  +-----+-----+-----+-----+-----+-----+-----+-----+
// 
//              where:
//                  A10..A0: Addr of Turnout -> max. 2048 This is a
//                           limitation compared to NMRA
//                           Range: 1..2047!!
//                  Color:   1 = closed (green), 0 = thrown (red)
//                  Sts:     turnout status (1 = on, 0 = off)
//                  Res:     0
//                  NoCmd:   0 
//                  n.u.        not currently used (0) -> may be A11?
//              The format is the same used with the XTrnt P50Xb
//
// Note:        not implemented  -> no Turnouts to report 
//              may be we use this breaking for blocking in case of
//              no switching turnouts

void XEvtTrn (void)
  {
    pc_send(0x00);
  }


//-------------------------------------------------------------------------------
//
// 0xCB; XEvtSen (1)
// Action:      Report eventual sensor events 
// Parameters   P50 Command (two bytes)
//     
// Reply:       iterated format consisting of one 1st which tells:
//              - 1st byte = 00h -> 'no further sensor event to report'
//              - 1st byte > 00h -> s88 module number whose status is reported
//                with the next 2 bytes:
//                  1st byte: status of s88 module, inputs #1..8 (bit #7..0).
//                  2nd byte: status of s88 module, inputs #9..16 (bit #7..0).
//              may be repeated up to 128 times, results in 2048 s88_total
//
// Note:        S88 events in general are reported by XEvent
//              A Answer to XEvtSen does NOT include the SAME s88 module twice      

void XEvtSen (void)
  {
    
    while (make_s88_ib_report())
      {
        pc_send(s88_ib_report[0]);
        pc_send(s88_ib_report[1]);
        pc_send(s88_ib_report[2]);
      }
    pc_send(0);                         // nothing (more) to report
  }


//=========================================================================
// Note on all Programming Commands:
// Upon issuing a (valid) P50Xb PT cmd which asks for a PT task,
// the IB shall immediately reply. The reply consists of a single byte:
// 
// 00h	cmd accepted, task in progress
// 80h	PT is busy (e.g. in use by a LocoNet device or previous task
//  	not yet completed)
// 02h	bad parameter value in cmd, cmd rejected
// 
// The 'busy' reply does NOT apply to commands which do not ask
// for a read or write task to be performed on the PT.

// Upon task completion, the Intellibox shall set the PT event flag
// (as read by the "are there events?" cmd (XEvent = 0C0h)).
// The PC can now issue the XEvtPT cmd (0CEh) and get the result
// of the PT task. The result is transmitted to the PC using
// the following format:


//-------------------------------------------------------------------------------
//
// 0xCE; XEvtPT (1)
// Action:      Report result of Programming Task
// Parameters   none
//     
// Reply:       1st byte = 0xF5 -> nothing to report / not yet completed
//              1st byte = [1,2,3,4,6] # of bytes to follow
//
//              2nd byte: status of Programming Task, values:
//                  0x00        Command completed, no errors
//                  0xFF        Timeout
//                  0xFE        No acknowledge from decoder (but a write maybe was successful)
//                  0xFD        Short! (on the PT)
//                  0xFC        No decoder detected
//                  0xFB        Generic Error
//                  0xFA        Error during DCC direct bit mode operation
//                  0xF9        No acknowledge to paged operation (paged r/w not supported?)
//                  0xF8        Error during Selectrix read
//                  0xF7        XPT_DCCQD: Ok (direct bit read mode is (probably) supported)
//                  0xF6        XPT_DCCQD: Not Ok (direct bit read mode is (probably) not supported)
//                  0xF4        Task terminated (see XPT_Term cmd)
//                  0xF3        No task to terminate (see XPT_Term cmd)
//                  0xF2        Cannot terminate task (see XPT_Term cmd)
//
//              3rd byte (sent only if 1st byte = 2, 3, 4, 6)
//                  1st byte resulting from a PT read task:
//                  - the content of a Register or CV (DCC Read cmds);
//                  - 1st decoder address ('search' cmds);
//                  - low byte of long addr read by DCC long addr read cmd;
//              4th byte (sent only if 1st byte = 3, 4, 6)
//                  2nd byte resulting from a PT read task:
//                  - eventual 2nd decoder address ('search' cmds);
//                  - high byte of long addr read by DCC long addr read cmd;

void XEvtPT (void)
  {
    prog_event = 0;
    if (programmer_busy())   
      {
        pc_send(0xF5);    // nothing to report / not yet completed
        return;
      }
    switch(prog_result_size)
      {
        case 0:
            pc_send(1);
            pc_send(prog_result);
            break;
            
        case 1:
            pc_send(2);
            pc_send(prog_result);
            pc_send(prog_data);
            break;

        case 2:
            pc_send(3);
            pc_send(prog_result);
            pc_send((unsigned char)prog_cv);
            pc_send((unsigned char)(prog_cv>>8));
            break;
      }
  }


//-----------------------  ######## Related P50Xb cmds (DCC)

//-------------------------------------------------------------------------------
//
// 0xDE;  XDCC_PD (0DEh) - length = 1+5 bytes
// Action:      DCC operations mode programming for locos (POM)
// Note         7-bit (1..127) or a 14-Bit (128..10239)
//              automatically selected
//
// Parameters   1st     low byte of long address
//              2nd     high byte of long address
//              3rd     CV# low byte
//              4th     CV# high byte
//              5th     write value (0..255)
//              
// Reply:       0x00    Ok, accepted 
//              0x80    Busy
void XDCC_PD (void)
  {
    unsigned int *addr_ptr;
    unsigned int *cv_ptr;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int

    cv_ptr = (unsigned int *) &pcc[4];     // use pcc[4] + pcc[5] as int

    if (lok_addr_out_of_range(addr_ptr))
      {
        pc_send(XMsg_BADPRM); return;
      }
    if ((*cv_ptr < 1) || (*cv_ptr > 1024))
      {
        pc_send(XMsg_BADPRM); return;
      }


    if (organizer_ready())
      {
        if (do_pom_loco(*addr_ptr, *cv_ptr, pcc[6]) )
          {
            pc_send(XMsg_OK);   return;
          }
      }
    pc_send(XMsg_NOSLOT);
  }


//-------------------------------------------------------------------------------
//
// 0xDF;  XDCC_PA (0DFh) - length = 1+5 bytes
// Action:      DCC operations mode programming for locos (POM - Accessory)
// Note         address must be in range 0..510 (= decoder address, not turnout!)
//              (Decoder address = (Turnout address - 1) / 4)
//
// Parameters   1st     low byte of long address
//              2nd     high byte of long address
//              3rd     CV# low byte
//              4th     CV# high byte
//              5th     write value (0..255)
//              
// Reply:       0x00    Ok, accepted 
//              0x80    Busy

void XDCC_PA (void)
  {
    unsigned int *addr_ptr;
    unsigned int *cv_ptr;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int

    cv_ptr = (unsigned int *) &pcc[4];     // use pcc[4] + pcc[5] as int

    if (lok_addr_out_of_range(addr_ptr))
      {
        pc_send(XMsg_BADPRM); return;
      }
    if ((*cv_ptr < 1) || (*cv_ptr > 1024))
      {
        pc_send(XMsg_BADPRM); return;
      }

    if (organizer_ready())
      {
        if (do_pom_accessory(*addr_ptr, *cv_ptr, pcc[6]) )
          {
            pc_send(XMsg_OK);   return;
          }
      }
    pc_send(XMsg_NOSLOT);
  }


//-------------------------------------------------------------------------------
//
// 0xE0; XPT_Sts (1)
// Action:      Report Programming Task Status
// Parameters   none
//
// Reply:      two bytes
//              1st     byte:
//              bit#   7     6     5     4     3     2     1     0
//                  +-----+-----+-----+-----+-----+-----+-----+-----+
//                  | Evt |  x  |  x  |  x  |  x  |  x  | Rel | Mode|
//                  +-----+-----+-----+-----+-----+-----+-----+-----+
//              Evt:  0 = no PT event is pending
//                    1 = a PT event is pending
//              Rel:  0 = PT relay is off
//                    1 = PT relay is on (always in OpenDCC Mode)
//              Mode: 0 = PT in 'PT only' mode
//                    1 = PT in 'auto' mode
//           
//              2nd     zero if no PT task is executing (PT is 'idle')
//              otherwise a non-zero value is returned
void XPT_Sts (void)
  {
    unsigned char retval;

    retval = 2;  // relay on, PT only mode

    if (prog_event) retval |= 0x80;
    pc_send(retval);    

    if (programmer_busy()) pc_send(0x01);
    else pc_send(0x00);    
    
  }


//-------------------------------------------------------------------------------
//
// 0xE1; XPT_On (1)
// Action:      Turn on the PT relay
// Parameters   none
//
// Reply:       ok.
void XPT_On (void)
  {
    pc_send(XMsg_OK);
    enter_progmode();
  }

//-------------------------------------------------------------------------------
//
// 0xE2; XPT_Off (1)
// Action:      Turn off the PT relay
// Parameters   none
//
// Reply:       0x00    Ok, PT relay has been turned off
//              0x80    PT is busy!
//              0xF1    Cannot turn off the PT relay: PT in 'PT only' mode!
void XPT_Off (void)
  {
    leave_progmode();
    pc_send(XMsg_OK);
    // !!!
    
  }



/*

######## DCC

*** XPT_DCCSr (0EAh) - length = 1 byte
Search for address of a DCC decoder
(decoders which are not 'readable', e.g. Roco/Lenz digital crane)
(via kurzes Einschalten Fahrstufe - not supported)

*** XPT_DCCQA (0EBh) - length = 1 byte
Read address using obsolete "Address Query" packet
(can only report 7-bit addresses in range 1..111)

*/

//-------------------------------------------------------------------------------
//
// 0xEC; XPT_DCCRR (3)
// Action:      Read using register mode
// Parameters   1st     Register# (1..8)
//              2nd     00
// 
// Reply:       0x00    Ok, accepted

void XPT_DCCRR (void)
  {
    unsigned int *addr_ptr;
    unsigned char retval = 0;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    
    retval = my_XPT_DCCRR(*addr_ptr);

    pc_send(retval);
    return;
  }


//-------------------------------------------------------------------------------
//
// 0xED; XPT_DCCWR (4)
// Action:      Write using register mode
// Parameters   1st     Register# (1..8)
//              2nd     00
//              3rd Write value
// 
// Reply:       0x00    Ok, accepted



void XPT_DCCWR (void)
  {
    unsigned int *addr_ptr;
    unsigned char retval = 0;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    
    retval = my_XPT_DCCWR(*addr_ptr, pcc[4]);

    pc_send(retval);
    return;
  }



//-------------------------------------------------------------------------------
//
// 0xEE; XPT_DCCRP (3)
// Action:      Read using direct (byte) mode
// Parameters   1st     CV# low byte
//              2nd     CV# high byte  ->  CV# must be in range 1..1024
// 
// Reply:       0x00    Ok, accepted

void XPT_DCCRP (void)
  {
    unsigned int *addr_ptr;
    unsigned char retval = 0;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    
    retval = my_XPT_DCCRP(*addr_ptr);

    pc_send(retval);
    return;
  }

//-------------------------------------------------------------------------------
//
// 0xEF; XPT_DCCWP (4)
// Action:      Write using paged mode
// Parameters   1st     CV# low byte
//              2nd     CV# high byte  ->  CV# must be in range 1..1024
//              3rd     write value (0..255)
// Reply:       0x00    Ok, accepted
//

void XPT_DCCWP (void)
  {
    unsigned int *addr_ptr;
    unsigned char retval = 0;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    
    retval = my_XPT_DCCWP(*addr_ptr, pcc[4]);

    pc_send(retval);
    return;
  }



//-------------------------------------------------------------------------------
//
// 0xF0; XPT_DCCRD (3)
// Action:      Read using direct (byte) mode
// Parameters   1st     CV# low byte
//              2nd     CV# high byte  ->  CV# must be in range 1..1024
// 
// Reply:       0x00    Ok, accepted

void XPT_DCCRD (void)
  {
    unsigned int *addr_ptr;
    unsigned char retval = 0;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    
    retval = my_XPT_DCCRD(*addr_ptr);

    pc_send(retval);
    return;
  }

//-------------------------------------------------------------------------------
//
// 0xF1; XPT_DCCWD (4)
// Action:      Write using direct (byte) mode
// Parameters   1st     CV# low byte
//              2nd     CV# high byte  ->  CV# must be in range 1..1024
//              3rd     write value (0..255)
//
// Reply:       0x00    Ok, accepted

void XPT_DCCWD (void)
  {
    unsigned int *addr_ptr;
    unsigned char retval = 0;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    
    retval = my_XPT_DCCWD(*addr_ptr, pcc[4]);

    pc_send(retval);
    return;
  }


//-------------------------------------------------------------------------------
//
// 0xF2; XPT_DCCRB (3)
// Action:      Read using direct (bit) mode
// Parameters   1st     CV# low byte
//              2nd     CV# high byte  ->  CV# must be in range 1..1024
//                      N.B. a whole CV (8 bits) are read, not only 1 bit!
//
//              
// Reply:       0x00    Ok, accepted !!! wo geht der Wert hin????

void XPT_DCCRB (void)
  {
    unsigned int *addr_ptr;
    unsigned char retval = 0;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    
    retval = my_XPT_DCCRB(*addr_ptr);

    pc_send(retval);
    return;
  }



//-------------------------------------------------------------------------------
//
// 0xF3;  XPT_DCCWB (0F3h) - length = 1+4 bytes
// Action:      Write a *single* bit using direct (bit) mode
// Parameters   1st     CV# low byte
//              2nd     CV# high byte  ->  CV# must be in range 1..1024
//              3rd     Bitposition (0..7)
//              4th     write value (0..1)
//              
// Reply:       0x00    Ok, accepted

void XPT_DCCWB (void)
  {
    unsigned int *addr_ptr;
    unsigned char retval = 0;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    
    retval = my_XPT_DCCWB(*addr_ptr, pcc[4], pcc[5]);

    pc_send(retval);
    return;
  }


//-------------------------------------------------------------------------------
//
// 0xF4;  XPT_DCCQD (0F4h) - length = 1 bytes
// Action:      Write a *single* bit using direct (bit) mode
// Parameters   none
//              
// Reply:       0x00    Ok, accepted

void XPT_DCCQD (void)
  {
    unsigned char retval = 0;

    retval = my_XPT_DCCQD();

    pc_send(retval);

  }

//-------------------------------------------------------------------------------
//
// 0xF5;  XPT_DCCRL (0F4h) - length = 1 bytes
// Action:      Read long (loco) address from CV17/18
//              and automatically set bit #5 of CV29 (is checked???)
// Note         Only compatible with decoders supporting direct bit
//              read and write mode!
// Parameters   none
//              
// Reply:       0x00    Ok, accepted 


void XPT_DCCRL (void)
  {
    unsigned char retval = 0;

    retval = my_XPT_DCCRL();

    pc_send(retval);
  }

//-------------------------------------------------------------------------------
//
// 0xF6;  XPT_DCCWL (0F6h) - length = 1+2 bytes
// Action:      Write long (loco) address into CV17/18
//              and automatically set bit #5 of CV29 
// Note         Only compatible with decoders supporting direct bit
//              read and write mode!
// Parameters   1st     low byte of long address
//              2nd     high byte of long address
//              
// Reply:       0x00    Ok, accepted 


void XPT_DCCWL (void)
  {
    unsigned int *addr_ptr;
    unsigned char retval = 0;

    addr_ptr = (unsigned int *) &pcc[2];     // use pcc[2] + pcc[3] as int
    
    retval = my_XPT_DCCWL(*addr_ptr);

    pc_send(retval);
  }



/*

*** XPT_DCCRA (0F7h) - length = 1 byte
Read long (accessory decoder) address from CV513/521.


*** XPT_DCCWA (0F8h) - length = 1+2 bytes
Write long (accessory decoder) address into CV513/521.
Parameters (byte):
1st     low byte of address
2nd     high byte of address
N.B. address must be in range 0..511


*/

//-------------------------------------------------------------------------------
//
// 0xFE; XPT_Term (1)
// Action:      Terminate (abort) current PT task
// Parameters   none
//
// Reply:       0x00    Ok, terminated (an XEvtPT cmd would now report: 01h, 0F4h)
//              0xF3    no task is active!
//              0xF2    could not terminate task! 
void XPT_Term (void)
  {
    pc_send(my_XPT_Term());
  }

/*

*** XPT_DCCEWr (0DCh) - length = 1+4 bytes
Generate and program the 'speed table' (CV's #67 through 94) of a
DCC 28 speed steps decoder. Three curve types can be specified:
1 = linear (constant speed increase: speed proportional to speed step)
2 = 'exponential' (finer speed control at low speed steps)
3 = 'logarithmic' (finer speed control at high speed steps)

Furthermore, the minimum and maximum speed must be specified (in range
1 through 101). The max speed can be adjusted so as to eventually
take into account a non 'full' (0 through 255) format of CV's 67
through 94 - or, of course, to reduce the maximum speed.
SMax = 101 gives a max speed of 254 (value programmed into CV #94).
SMax =  40 gives CV #94 = 100
SMax =  13 gives CV #94 = 32

SMin = 1 gives a minum speed of 2 (value programmed into CV #67).

Parameters (byte):
1st     (not currently used)
2nd     SMin (1..123) minimum speed
3rd     SMax (2..124) maximum speed
4th     SKL (1..3) Steuerkennlinie (speed curve)

N.B. SMin must be less than SMax.


*/




//===========================================================================================

// Liste aller IBOX Routinen - Zugleich Parsertebelle und Sprungverteiler

// In dieser Tabelle wird nach dem Opcode gesucht, dann die angegebene Length von Parametern
// vom Host nachgeladen und dann die angegebene Funktion gerufen; Parameterbergabe in globalen
// Variablen pcc[2, fortfolgende].
// Length = 0 bedeutet, diese Funktion ist void.
// Nicht gefundene Opcodes werden mit BadSyntax zurckgemeldet.

typedef struct
{
  unsigned char opcode;
  unsigned char length;             // Befehlslnge
  void (*fnAction)(void);
} T_ACTION;

T_ACTION ParseTab[] PROGMEM =
  {
    // Opcode / Length / Call   // Name  -           Length / implemented / simulated / tested / comment
    { 0x80,  5, XLok },         // XLok (080h)        = 1+4 / i / s / t /
//  { 0x81,  5  XlokX },        // XLokX (081h)       = 1+4 / n / - / - / not supported, extension by Tams EasyControl
    { 0x83,  3, XNOP },         // XLkDisp (083h)     = 1+2 / i /         not supported -> always 0x00
    { 0x84,  3, XLokSts },      // XLokSts (084h)     = 1+2 / i / - / -
    { 0x85,  3, XLokCfg },      // XLokCfg (085h)     = 1+2 / i / - / - ERROR
    { 0x86,  5, XLokCfgSet },   // XLokCfgSet (086h)  = 1+4 / i / - / - / -> OpenDCC extension
    { 0x88,  4, XFunc },        // XFunc (088h)       = 1+3 / i / -
    { 0x89,  4, XFuncX },       // XFuncX (089h)      = 1+3 / i / - / - / F13-F16 not supported
    { 0x8C,  3, XFuncSts },     // XFuncSts (08Ch)    = 1+2 / i /
    { 0x8D,  3, XFuncXSts },    // XFuncXSts (08Dh)   = 1+2 / i / - / - / F13-F16 not supported           
    { 0x90,  3, XTrnt },        // XTrnt (090h)       = 1+2 / i / s / t / 
    { 0x93,  1, XTrntFree },    // XTrntFree (093h)   = 1   / i / - / - / not supported by OpenDCC                        
    { 0x94,  3, XTrntSts },     // XTrntSts (094h)    = 1+2 / i / 
    { 0x95,  2, XTrntGrp },     // XTrntGrp (095h)    = 1+1 / i /
    { 0x98,  2, XSensor },      // XSensor (098h)     = 1+1 / i / - / - /  
    { 0x99,  1, XSensOff },     // XSensOff (099h)    = 1   / i /       / S88 Sensor rcksetzen
    { 0x9C,  2, X88PGet },      // X88PGet (09Ch)     = 1+1 / i /       / get s88 config - modified by OpenDCC (3 s88 streams)
    { 0x9D,  3, X88PSet },      // X88PSet (09Dh)     = 1+2 / i / - / - / set s88 config - modified by OpenDCC 
    { 0x9E,  2, XError },       // Xs88Tim (09Eh)     = 1+1 / n / - / - / not supported by OpenDCC 
    { 0x9F,  2, XError },       // Xs88Cnt (09Fh)     = 1+1 / n / - / - / not supported by OpenDCC 
    { 0xa0,  1, XVer },         // XVer (0A0h)        = 1   / i / - / - / query version - modified by OpenDCC 
//  { 0xa1,  2, XError },       // XP50XCh (0A1h)     = 1+1 / n / - / - / change X char - not supported by OpenDCC 
    { 0xa2,  1, XStatus },      // XStatus (0A2h)     = 1   / i / - / - / report status
    { 0xa3,  4, XSOSet },       // XSOSet (0A3h)      = 1+3 / i / - / - / write directly to eeprom (IBOX doesnt)
    { 0xa4,  3, XSOGet },       // XSOGet (0A4h)      = 1+2 / i / s / t / read directly from eeprom
    { 0xa5,  1, XHalt },        // XHalt (0A5h)       = 1   / i / - / - /  
    { 0xa6,  1, XPwrOff },      // XPwrOff (0A6h)     = 1   / i / s / t /  
    { 0xa7,  1, XPwrOn },       // XPwrOn (0A7h)      = 1   / i / s / t /  
    { 0xc4,  1, XNOP },         // XNOP (0C4h)        = 1   / i / s / t /
    { 0xc6,  2, XP50Len1 },     // XP50Len1 (0C6h)    = 1+1 / i / - / - / not supported by OpenDCC 
    { 0xc7,  3, XP50Len2 },     // XP50Len2 (0C7h)    = 1+2 / i / - / - / not supported by OpenDCC 
    { 0xc8,  1, XEvent },       // XEvent (0C8h)      = 1   / i / - / - /  
    { 0xc9,  1, XEvtLok },      // XEvtLok (0C9h)     = 1   / n / - / - / not supported by OpenDCC -> 0x80
    { 0xca,  1, XNOP },         // XEvtTrn (0CAh)     = 1   / i / - / - / not supported by OpenDCC -> 0x00 (same as NOP)
    { 0xcB,  1, XEvtSen },      // XEvtSen (0CBh)     = 1   / i / - / - / report sensor events
    { 0xcc,  1, XNOP },         // XEvtIR (0CCh)      = 1   / n / - / - / IR Event
    { 0xce,  1, XEvtPT },       // XEvtPT (0CEh)      = 1   / i / s / t / Programming Event
    { 0xcf,  1, XNOP },         // XEvtTkR (0CFh)     = 1   /
    { 0xd0,  1, XNOP },         // XEvtMem (0D0h)     = 1   /
//  { 0xda,  6, XDCC_PDR },     // neu                = 1+4 / *           DCC operations mode cv-read for Loks (nur BiDi)
//  { 0xdb,  6, XDCC_PAR },     // neu                = 1+4 / *           DCC operations mode cv-read for Acc
    { 0xdc,  5, XNOP },         // XPT_DCCEWr (0DCh)  = 1+4 / n / - / - / Speed Kennlinie (not supported by OpenDCC)
    { 0xdd,  5, XNOP },         // XPT_FMZEWr (0DDh)  = 1+4 / - / - / - / FMZ Write Extended
    { 0xde,  6, XDCC_PD },      // XDCC_PD (0DEh)     = 1+5 / i / s / t / DCC operations mode programming for Loks
    { 0xdf,  6, XDCC_PA },      // XDCC_PA (0DFh)     = 1+5 / i / - / - / DCC operations mode programming for Acc
    { 0xe0,  1, XPT_Sts },      // XPT_Sts (0E0h)     = 1   / - / - / - / Programming Status
    { 0xe1,  1, XPT_On },       // XPT_On (0E1h)      = 1   / i / - / - / Programming On
    { 0xe2,  1, XPT_Off },      // XPT_Off (0E2h)     = 1   / i / - / - / Programming Off
//  { 0xe4,  1, XNOP },         // XPT_SXRd (0E4h)    = 1   / n / - / - / Selectrix Read
//  { 0xe5,  6, XNOP },         // XPT_SXWr (0E5h)    = 1+5 / n / - / - / Selectrix Write
//  { 0xe6,  1, XNOP },         // XPT_SXSr (0E6h)    = 1   / n / - / - / Selectrix Search
//  { 0xe7,  1, XNOP },         // XPT_FMZSr (0E7h)   = 1   / n / - / - / FMZ Search
//  { 0xe8,  2, XNOP },         // XPT_FMZWr (0E8h)   = 1+1 / n / - / - / FMZ Write
//  { 0xe9,  2, XNOP },         // XPT_MrkSr (0E9h)   = 1+1 / n / - / - / Mrklin Search
    { 0xea,  1, XNOP },         // XPT_DCCSr (0EAh)   = 1 
    { 0xeb,  1, XNOP },         // XPT_DCCQA (0EBh)   = 1   / n / - / - / Adress Query (obsolete command)
    { 0xec,  3, XPT_DCCRR },    // XPT_DCCRR (0ECh)   = 1+2 / i / - / - / Read register
    { 0xed,  4, XPT_DCCWR },    // XPT_DCCWR (0EDh)   = 1+3 / i / - / - / write register
    { 0xee,  3, XPT_DCCRP },    // XPT_DCCRP (0EEh)   = 1+2 / i / - / - / Read using paged mode + scan
    { 0xef,  4, XPT_DCCWP },    // XPT_DCCWP (0EFh)   = 1+3 / i / - / - / Write using paged mode
    { 0xf0,  3, XPT_DCCRD },    // XPT_DCCRD (0F0h)   = 1+2 / i / s / t / Read using direct mode + scan
    { 0xf1,  4, XPT_DCCWD },    // XPT_DCCWD (0F1h)   = 1+3 / i / s / t / Write using direct (byte) mode
    { 0xf2,  3, XPT_DCCRB },    // XPT_DCCRB (0F2h)   = 1+2 / i / - / - / Read using direct mode + bit operations
    { 0xf3,  5, XPT_DCCWB },    // XPT_DCCWB (0F3h)   = 1+4 / i / - / - / Write Single bit
    { 0xf4,  1, XPT_DCCQD },    // XPT_DCCQD (0F4h)   = 1   / i / - / - / Test decoder for bit operation
    { 0xf5,  1, XPT_DCCRL },    // XPT_DCCRL (0F5h)   = 1   /             read long addr
    { 0xf6,  3, XPT_DCCWL },    // XPT_DCCWL (0F6h)   = 1+2               write long addr
    { 0xf7,  1, XNOP },         // XPT_DCCRA (0F7h)   = 1 
    { 0xf8,  3, XNOP },         // XPT_DCCWA (0F8h)   = 1+2 
//  { 0xf9,  9, XNOP },         // XPT_U750 (0F9h)    = 1+8 / n / - / - / Uhlenbrock Special
//  { 0xfa, 12, XNOP },         // XPT_U755 (0FAh)    = 1+11/ n / - / - / Uhlenbrock Special
//  { 0xfb, 11, XNOP },         // XPT_U760 (0FBh)    = 1+10/ n / - / - / Uhlenbrock Special
    { 0xfc,  1, XNOP },         // XPT_Src (0FCh)     = 1 
    { 0xfd,  1, XNOP },         // XPT_Ctrl (0FDh)    = 1   / n / - / - / (to be defined)
    { 0xfe,  1, XPT_Term },     // XPT_Term (0FEh)    = 1   / i / - / - / PTerminate (abort) current PT task
    { 0x00,  0, XError },};


//--------------------------------------------------------------------------------
// This routine checks for p50x-opcode and returns the parameter count
//
/* Code fr ParseTab in RAM:
unsigned char get_p50xb_length(unsigned char opcode)
  {
    T_ACTION *p;
    p=ParseTab;
    while(p->opcode != 0)
      {
        if (p->opcode == opcode)
          {
             return(p->length);
          }
        p++;
      }
    return(0);
  }
*/

// Code fr ParseTab in FLASH:
unsigned char get_p50xb_length(unsigned char opcode)
  {
    const T_ACTION  *p;
    p=ParseTab;
    unsigned char oc;
    unsigned char length;

    while((oc=pgm_read_byte(&p->opcode)) != 0)
      {
        if (oc == opcode)
          {
             length = pgm_read_byte(&p->length);   // read from  flash
             return(length);
          }
        p++;
      }
    return(0);
  }

//--------------------------------------------------------------------------------
// This routine checks for p50x-opcode and calls the corresponding function

/*
// RAM:
void parse_p50x_binary (unsigned char opcode)
  {
    const T_ACTION* p;
    p=ParseTab;
    while(p->opcode != 0)
      {
        if (p->opcode == opcode)
          {
             p->fnAction();
             return;
          }
        p++;
      }
    return;
  }
*/

void parse_p50x_binary (unsigned char opcode)
  {
    const T_ACTION* p;
    p=ParseTab;
    unsigned char oc;
    void (*fp)(void);   

    while((oc=pgm_read_byte(&p->opcode)) != 0)
      {
        if (oc == opcode)
          {
             fp = (void *) pgm_read_word(&p->fnAction);   // assign function addr read from flash
              (*fp)();                  // now call this function
             return;
          }
        p++;
      }
    return;
  }



//===============================================================================================
//
// ASCII-Part
//    
//------------------------------------------------------------------------------------------
// parse_p50x_ascii
//  
// ASCII Mode: - ignore leading blanks, char must be upper case
//             - if parameter starts with $, it is read as hex, with % read as bin, msb first
//             - when reading numbers: ignore leading zeros
//             - if parameters are missing, they are interrogated
//             - backspace will kill last character - we DON'T support this.
//             - answer is always terminated by <cr> and ]


    
// -- Help messages
const char Help_Msg[]           PROGMEM = "Help cmds: HL, HF, HT";
const char Help_on_Lok[]        PROGMEM = "L {Lok#, [Speed], [FL], [Dir], [F1], [F2], [F3], [F4]}";
const char Help_on_Turnout[]    PROGMEM = "T {Trnt#, [Color], [Status]}";
const char Help_on_Function[]   PROGMEM = "F {Lok#, [F1], [F2], [F3], [F4], [F5], [F6], [F7], [F8]}";


// -- Parser Messages
const char Msg_Empty[]      PROGMEM = "";
const char Msg_PwrOn[]      PROGMEM = "Pwr on";
const char Msg_PwrOff[]     PROGMEM = "Pwr off";
const char Msg_Halted[]     PROGMEM = "Halted!";
const char Msg_DCCProg[]    PROGMEM = "DCC program";
const char Msg_Version[]    PROGMEM =  {'O','p','e','n','D','C','C',' ','V','0','.',
                                         (((OPENDCC_VERSION & 0xF0) >> 4) + 0x30),
                                       #if ((OPENDCC_VERSION & 0x0F) < 10)
                                         '0',((OPENDCC_VERSION & 0x0F) + 0x30),
                                       #else
                                         '1',((OPENDCC_VERSION & 0x0F) + 0x26),
                                       #endif  
                                         0}; 
const char Msg_unknown[]    PROGMEM = "unknown command"; 
const char Msg_DCC[]        PROGMEM = "DCC "; 
const char Msg_SO[]         PROGMEM = "SO"; 
const char Msg_ZzA0[]       PROGMEM = "Mixed P50/P50X mode"; 
const char Msg_ZzA1[]       PROGMEM = "P50X only mode (P50 is disabled)"; 
const char Msg_BadPara[]    PROGMEM = "Bad Parameter"; 
const char Msg_failed[]     PROGMEM = "failed"; 
const char Msg_unused[]     PROGMEM = "unused"; 
const char Msg_NoLok[]      PROGMEM = "No Lok";
const char Msg_NotImpl[]    PROGMEM = "Not Implemented";
const char Msg_BaudChgd[]   PROGMEM = "Baud rate changed!";

const char Msg_Okay[]       PROGMEM = "Ok";

const char Msg_Prog[]       PROGMEM = "Prog ";


unsigned int parameter[10];

void parse_parameters(void)
  {
    unsigned char i;
    
    for (i=0; i<(sizeof(parameter)/sizeof(unsigned int)); i++)
      {
        parameter[i] = 0xfffe;          // mark as void
      }

    for (i=0; i<(sizeof(parameter)/sizeof(unsigned int)); i++)
      {
        if (get_px50arg(&parameter[i]) < 0)
          {
            // eol or void parameter reached
            break;
          }
      }
  }


void XA_not_implemented(void)
  {
    pc_send_answer_P(Msg_NotImpl);
  }

void XA_Empty(void)
  {
    pc_send_answer_P(Msg_Empty);
  }

void XA_unknown(void)
  {
    pc_send_answer_P(Msg_unknown);
    return;

    #ifdef myoldcode1
    pc_send_string_P(Msg_unknown);
    pc_send(' ');
    unsigned char i=0;
    while (pcc[i])
      {
        pc_send(pcc[i]);
        i++;
      }
    pc_send(' ');
    pc_send_num(parameter[0]);
    
    XA_Empty();
    #endif
  }


 
// "xZzA1"      "P50X only" mode On
// "ZzA0"       "P50X only" mode Off (turns on the P50) 

void XA_ZZA(void)
  {
    if (parameter[0] == 1)
      {
        P50X_mode = 1;
        pc_send_answer_P(Msg_ZzA1);
      }
    else if (parameter[0] == 0)
      {
        P50X_mode = 0;
        pc_send_answer_P(Msg_ZzA0);
      }
    else pc_send_answer_P(Msg_BadPara);
  }    

// Help
void XA_Help(void)
  {
    pc_send_answer_P(Help_Msg);
  }

// Help on Lok
void XA_Help_on_Lok(void)
  {
    pc_send_answer_P(Help_on_Lok);
  }

// Help on Function
void XA_Help_on_Function(void)
  {
    pc_send_answer_P(Help_on_Function);
  }

// Help on Turnout
void XA_Help_on_Turnout(void)
  {
    pc_send_answer_P(Help_on_Turnout);
  }

// Halt
// Die Funktion hlt alle Loks an (per Emergency-Stop), schaltet die Gleis-
// signale jedoch nicht ab. Solange die Anlage in diesem Zustand ist, kann keine
// Lok angefahren werden, der Rest funktioniert jedoch normal.
//Antwort: "Halted!"
           
void XA_Halt(void)
  {
     set_opendcc_state(RUN_STOP);
     pc_send_answer_P(Msg_Halted);
  }

void XA_Stop(void)
  {
     // power off !
     set_opendcc_state(RUN_OFF);
     pc_send_answer_P(Msg_PwrOff);
  }

void XA_Go(void)
  {
     set_opendcc_state(RUN_OKAY);
        // hier mte man warten, runstate abfragen und fallweise den Errorcode senden
        pc_send_answer_P(Msg_PwrOn);
  }

void XA_ShowStatus(void)
  {
    // Shows the current status of OpenDCC
   
    switch(opendcc_state)                   // this is the current running state
      {
        case RUN_OKAY:            
            pc_send_answer_P(Msg_PwrOn);
            break;
        case RUN_STOP:                      // DCC Running, all Engines Emergency Stop
            pc_send_answer_P(Msg_Halted);
            break;
        case RUN_OFF:                       // Output disabled (2*Taste, PC)
            pc_send_answer_P(Msg_PwrOff);
            break;
        case RUN_SHORT:                     // Kurzschluss, hektisch blinken im Wechsel
            pc_send_answer_P(Msg_PwrOff);
            break;
        case RUN_PAUSE:                     // DCC Running, all Engines Speed 0
            pc_send_answer_P(Msg_Halted);
            break;
        case PROG_OKAY:
            pc_send_answer_P(Msg_DCCProg);
            break;
        case PROG_SHORT:           // Kurzschlu, hektisch blinken
            pc_send_string_P(Msg_DCCProg);
            pc_send_answer_P(Msg_PwrOff);
            break;
        case PROG_OFF:
            pc_send_string_P(Msg_DCCProg);
            pc_send_answer_P(Msg_PwrOff);
            break;
        case PROG_ERROR:         // wie go, CTRL blinkt
            pc_send_string_P(Msg_DCCProg);
            pc_send_answer_P(Msg_failed);          
            break;
      }
  }


// V: Version
// Gibt die Versions- und Seriennummer von OpenDCC aus:void XA_ShowVersion(void)
void XA_ShowVersion(void)
  {
     pc_send_answer_P(Msg_Version);
  }


//    {
//      // MT: not implemented
//      // *** MT {[TMin], [TMax]}
//      // TMin minimum turnout activation time (0..255)
//      // TMax maximum turnout activation time (0..255)
//    }


// SR: not implemented
            // Flag     0 or 1
            // This allows reading or setting the s88 'auto-reset' feature.
            // The command 'SR0' corresponds to the P50 command 080h.
            // The command 'SR1' corresponds to the P50 command 0C0h.

void XA_S88_read(void)
  {
    // not implemented
    // *** SS {s88}
    // s88      s88 module number to be read (1..31)
    if ((parameter[0] >= 1) && (parameter[0] <= 31) )
      {
        pc_send_num(parameter[0]);
        XA_Empty();
      }
    else pc_send_answer_P(Msg_BadPara);
  }

void XA_S88_set_modules(void)
  {
    // not implemented
    // *** SE {s88h}
    // s88h     sets the highest s88 module 'half number' to be
    //  automatically read by the IB
    //  *OR*
    //  disable automatic reading (s88h = 0)
    // Allowed range is: 0..62
    //
    // For example, in order to have the IB automatically read only
    // the very first s88 module, the following command should be
    // used: 'SE 2'.
    // The IB default (per Special Option #8) is: s88h = 16, which implies
    // reading automatically the first 8 s88 modules (16 'half modules').
    // If a P50 cmd for reading the s88 modules requires more than what
    // is currently automatically read by the IB, then the IB automatically
    // raises this number. This does NOT apply to event-style reading!
    // You can also specify 0 as a value for s88h. In this case, the IB
    // shall cease reading s88 modules automatically and only read them
    // when an explicit P50 or P50Xa cmd shall be sent.

    // This command allows to override, up to the next IB reset, the value
    // specified by SO #8.

    // N.B. No sensor event can be generated for s88 modules which are not
    // being read automatically by the IB. Therefore, it is quite important
    // to let the IB know, either manually through the IB menus or through
    // this 'SE' cmd (or X88PSet), how many s88 modules are actually connected.
    // LocoNet sensors, on the contrary, would be active even if s88 automatic
    // reading has been disabled.// *** SE {s88}

    if ((parameter[0] <= 62) )
      {
        pc_send_num(parameter[0]);
        XA_Empty();
      }
    else pc_send_answer_P(Msg_BadPara);
  }

// Hilfsprogramm 
void pc_send_lok_status(unsigned char locobuffer_i)
  {
    pc_send('L');
    pc_send_num(parameter[0]);
    pc_send_num(locobuffer[locobuffer_i].speed & 0x7F);
    pc_send_num((locobuffer[locobuffer_i].speed & 0x80) >> 7);
    pc_send_num(locobuffer[locobuffer_i].fl);
    pc_send_num(locobuffer[locobuffer_i].f4_f1);
    pc_send_num(convert_speed_to_rail(locobuffer[locobuffer_i].speed, locobuffer[locobuffer_i].format) );
    // Note: sending all f1..f4 is not correct, but useful!
    XA_Empty();
  }

void XA_Lok(void)
  {
    unsigned char i;
        unsigned char speed, dir, func;
        // *** L {Lok#, [Speed], [FL], [Dir], [F1], [F2], [F3], [F4]}
    // Lok#     0..9999 (depending on protocol, not every address is legal!)
    // Speed    0..127 (0 = Stop, 1 = Stop/Em.Stop (protocol dependent))
    // FL       Light status (1 = on, 0 = off)
    // Dir      Direction (1 = forward, 0 = reverse). Direction may also be
    //  specified as 'f' = forward or 'r' = reverse
    // F1       F1 status (1 = on, 0 = off)
    // etc...

    // depending on the Lok protocol, not all parameters are internally used
    // or have the same range.
    // For example, the (real) address range of each protocol is:
    // Mrk      1..80 (1..255 for Uhlenbrock decoders)
    // SX       0..111
    // DCC      1..9999 (not every DCC decoder supports addr. > 99 or 127)
    // FMZ      1..119
    // however, virtual addresses can go from 0 to 9999.
    // 
    // Depending on Lok protocol, Speed = 1 means (check also initial
    // introductory note):
    // Mrk Old          same as Stop
    // Mrk New          Emergency Stop
    // Selectrix        same as Stop
    // DCC              Emergency Stop
    // FMZ              Emergency Stop
    // 
    // Another variable is the number of additional functions (apart from FL,
    // i.e. usually the Lok lights - also known as F0):
    // Mrk              4
    // Selectrix        1
    // DCC              4+ (F5.. to be specified using the P50Xa 'F' cmd)
    // FMZ              1
    // 
    // If only the first (mandatory) cmd parameter is specified (i.e. Lok#),
    // then the current status of this Lok shall be displayed, provided, of
    // course, that this Lok has ever been commanded before! (Or else the
    // message 'No Lok data!' would be displayed.)
    // 
    // The Speed parameter is *internally* scaled down in compliance with
    // the Lok protocol. Min speed is always set by Speed = 2.
    // Max Speed is always set by Speed = 127. The actual number of non-Stop
    // speed steps varies depending on the Lok protocol:
    // Mrk              14
    // Selectrix        31
    // DCC              14/27/28/126
    // FMZ              15

    // All parameters excluding the Lok# are optional. If at least one is
    // specified, then Lok status is updated only as far as the specified
    // parameter is concerned. Should no previous status be available, then
    // the default is:
    // Speed = 0
    // Dir = 1 = forward
    // FL (light) = Off
    // F1..F4 = Off
    // This only applies to the case in which at least one status parameter
    // has been specified!
    // In fact, as already written above, if no previous status is known and
    // if the P50Xa 'L' cmd only specifies the Lok# parameter, then an error
    // message is sent ('No Lok data').

    // Address conflict with other devices:
    // if the Lok specified by the 'L' cmd is already controlled by a non-PC
    // device, then the error message 'Lok busy!' shall be displayed.

    // Halt mode:
    // If the Intellibox is in Halt mode, then Lok status (as specified by the
    // 'L' cmd) is updated, but any Speed changes involving a non-Stop speed
    // shall NOT be effective.
    // In this case, the reply shall be: "Halted!".

    // Power Off:
    // If the Intellibox is in Power Off, then Lok status (as specified by the
    // 'L' cmd) is updated. In this case, the reply shall be: "Pwr Off".

    if (lok_addr_out_of_range(&parameter[0]))
      {
        pc_send_answer_P(Msg_BadPara);
        return;
      }
    
    if (parameter[1] == 0xfffe)  // no more parameters -> read LOK 
      {
        // read setting: [Speed], [FL], [Dir], [F1], [F2], [F3], [F4]
        i = scan_locobuffer(&parameter[0]);
        if (i==SIZE_LOCOBUFFER)            // not found - was not used yet
          {
            pc_send_answer_P(Msg_NoLok); return;
          }
        else
          {
            pc_send_lok_status(i);
            return;
          }
      }
        else
          {    // set lok parameters
            // search locobuffer
            i = scan_locobuffer(&parameter[0]);
        if (i==SIZE_LOCOBUFFER)            // not found - was not used yet
          {
            // new lok -> set defaults
                        speed = 0;
                        dir = 0x80;
                        func = 0;
          }
                else
                  { // load old values
                    speed = locobuffer[i].speed & 0x7F;
            dir = locobuffer[i].speed & 0x80;
            func = locobuffer[i].fl;
          }
                // now check for changes
        if ( (parameter[1] >= 0) && (parameter[1] <= 126) ) 
                  {
             speed = parameter[1];
                  }
        if ( (parameter[2] == 0) || (parameter[2] == 1) ) 
                  {
             func = parameter[2];
                  }
        if (parameter[3] == 0)  dir = 0;
                if (parameter[3] == 1)  dir = 0x80;
                
                do_loco_speed(parameter[0], dir | speed);
                do_loco_func_grp0(parameter[0], func);
        pc_send_lok_status(i);
          }
  }

void XA_Lok_Cfg(void)
  {
    // *** LC {Lok#}
    // Lok#     0..9999 (depending on protocol, not every address is legal!)
    // Reports Lok protocol configuration (e.g. 'Mot Old', etc...) and
    // number of non-Stop speed steps.
    // Motorola I: "Motorola Old"  (not with OpenDCC)
    // Motorola II: "Motorola New" (not with OpenDCC)
    // DCC: "DCC"
    // Ist die Lok nicht im Refresh-Buffer (bei der IB heit das wohl 'Slot')
    // vermerkt, so wird "unused" ausgegeben. Bei illegaler (oder fehlender) Lok-
    // Adresse wird "Loco adress out of range" ausgegeben.

    unsigned char result;    

    result = scan_locobuffer(&parameter[0]);

    if (result==SIZE_LOCOBUFFER)            // not found - was not used yet
      {
        pc_send_answer_P(Msg_unused);
      }
    else
      {
        pc_send_string_P(Msg_DCC);
        switch(locobuffer[result].format)
          {
            case DCC14:     pc_send_num(14);
                            break;
            case DCC27:
            case DCC28:     pc_send_num(28);
                            break;
            case DCC128:    pc_send_num(126);
                            break;
          }
        XA_Empty();
      }
  }

void XA_Lok_Cfg_Set(void)
  {
    // LS command -> new in OpenDCC!
    // *** LS {Lok#, [Format], [Steps]}
    // Format: may be either MM1, MM2 or DCC -> this is coded as 0,1,2
    // Steps: may be 14, 28, 126
    unsigned char valid = 0, myformat = 0;
 
    switch (parameter[2])
      {
        case 14:
            myformat = DCC14;
            valid = 1;
            break;
        case 28:
            myformat = DCC28;
            valid = 1;
            break;
        case 126:
            myformat = DCC128;
            valid = 1;
            break;
      }
    if (lok_addr_out_of_range(&parameter[0]))  valid = 0;
    if (valid)
      {
        if (store_loco_format(parameter[0], myformat))
          {
            XA_Lok_Cfg();                   // read back format an report it
          }
        else pc_send_answer_P(Msg_failed);  // failed - to many locos with extra format
      }
    else
      {
        pc_send_answer_P(Msg_BadPara);  // Bad Parameter
      }
  }

void XA_Turnout(void)
  {
    // *** T {Trnt#, [Color], [Status]}
    // Trnt#    1..2040 (depending on protocol, not every address is legal!)
    // Please note: turnout address, NOT turnout *decoder* address!
    // Legal Motorola format turnout address range: 1..320
    // Legal DCC format turnout address range: 1..2040
    // Color    'r' (red = thrown) or 'g' (green = closed)
    // ('r' may also be spec'd as '0', 'g' also as '1')
    // Status   1 = on, 0 = off (if not specified -> off)

    // If only the first (mandatory) cmd parameter is specified (i.e. Turnout#),
    // then the current status of this Turnout shall be displayed.

    // In addition to the usual 'Syntax Error' reply, this cmd can also
    // reply with:
    // 'Turnout fifo is full!'
    // 'Power is Off!'
    // 'Illegal Turnout address for this protocol!'
    //
    // Color: 0 oder 'r' = Rot / Abzweig, 1 oder 'g' = Grn / gerade aus
    // Status: 1 = ein, 0 oder nicht angegeben = aus

    // 
    pc_send('T');
    pc_send_num(parameter[0]);
    if ((parameter[0] >= 1) && (parameter[0] <= 2040) )
      {
        if ( ((char)parameter[1] == 0) || ((char)parameter[1] == 1) )
          {
            pc_send_num(parameter[1]);
            pc_send_num(parameter[2]);
            if ( ((char)parameter[2] == 0) || ((char)parameter[2] == 1) )
              {
                if ( do_accessory(parameter[0]-1, parameter[1], parameter[2]) )  
                  {
                    XA_Empty();
                  }
                else
                    pc_send_answer_P(Msg_failed);
                return;
              }
          }
        else
          {  // color falsch (keine Fehlermeldung, immer abfragen)
            unsigned char position;
            position = recall_turnout(parameter[0]-1);
            pc_send_num(position);
          }
      }
    pc_send_answer_P(Msg_BadPara);
  }
    
void XA_Baud(void)
  {
    t_baud baud;
    long myrate;
    // *** B [baud_rate]
    // The officially supported IB baud rates are: 2400, 4800, 9600 and 19200 bps.
    // 38400 bps should also work - even if we DO NOT officially support this
    // RS-232 speed. (+57600)
    // The XNOP P50Xb cmd can be used in order to automatically detect both
    // the current baud rate and the currently active protocol(s): either P50 only,
    // P50X only, or "mixed" (P50 and P50X).
    // Answer: "Baud rate changed!"
    if (parameter[0] == 0xFFFE)
      {
        pc_send('B');
        myrate = pgm_read_dword(&baudrate[actual_baudrate]);
        if (myrate >= 65536)
        pc_send_num((unsigned int) myrate/65536L);
        pc_send_num((unsigned int) myrate);
        XA_Empty();
        return;
      }
    if (parameter[0] == 2400) baud = BAUD_2400;
    else if (parameter[0] == 4800) baud = BAUD_4800;
    else if (parameter[0] == 9600) baud = BAUD_9600;
    else if (parameter[0] == 19200) baud = BAUD_19200;
    else if (parameter[0] == 38400) baud = BAUD_38400;
    else if (parameter[0] == 57600) baud = BAUD_57600;
    else
      {
        baud = DEFAULT_BAUD;
        parameter[0] = pgm_read_dword(&baudrate[baud]);
      }
    pc_send_answer_P(Msg_BaudChgd); _mydelay_us(2000);
    
    // This was the old answer in V0.13
    // pc_send('B'); pc_send_num(parameter[0]); XA_Empty();
    
    while(!tx_all_sent());                  // busy waiting until all sent
                   
    init_rs232(baud);                 // now switch and flush fifos.
  }


// Hilfsprogramm 
void pc_send_lok_function_1_8(unsigned char locobuffer_i)
  {
    pc_send('F');
    pc_send_num(parameter[0]);
    pc_send_num(locobuffer[locobuffer_i].f4_f1);   // nicht ganz intellibox, tuts aber 
    pc_send_num(locobuffer[locobuffer_i].f8_f5);
    XA_Empty();
  }

void XA_Function(void)
  {
    // *** F {Lok#, [F1], [F2], [F3], [F4], [F5], [F6], [F7], [F8]}
    // Lok#     0..9999 (depending on protocol, not every address is legal!)
    // F1       F1 status (1 = on, 0 = off)
    // etc...

    // If only the first (mandatory) cmd parameter is specified (i.e. Lok#),
    // then the current status of this Lok shall be displayed - as far as
    // functions are concerned.

    unsigned char i, j, func;

    if (!lok_addr_out_of_range(&parameter[0]) )
      {
        if (parameter[1] == 0xfffe)  // no more parameters -> read LOK 
          {
                // read setting: [F1], [F2], [F3], [F4], ....
            i = scan_locobuffer(&parameter[0]);
            if (i==SIZE_LOCOBUFFER)            // not found - was not used yet
              {
                pc_send_answer_P(Msg_NoLok); return;
              }
            else
              {
                pc_send_lok_function_1_8(i);
                return;
              }
          }
        else
          {    // set lok parameters
               // search locobuffer
            i = scan_locobuffer(&parameter[0]);
            if (i==SIZE_LOCOBUFFER)            // not found - was not used yet
              {
                // new lok -> set defaults
                // wer macht das?
                func = 0;
              }
            else
              { // load old values
                func = locobuffer[i].f8_f5 << 4;
                func |= locobuffer[i].f4_f1;
              }

            for (j=0; j<8; j++)
              {
                if (parameter[j] == 0)
                  {
                    unsigned char mymask = 1 << j;
                    func = func & ~mymask;
                  }
                if (parameter[j] == 1)
                  {
                    unsigned char mymask = 1 << j;
                    func = func | mymask;
                  }
              }

            // now store the new functions
            do_loco_func_grp1(parameter[0], func);
            do_loco_func_grp2(parameter[0], func >> 4);
            pc_send_lok_function_1_8(i);
          }
      }
    else
      {
        pc_send_answer_P(Msg_BadPara);
      }
  }


// Hilfsprogramm 
void pc_send_lok_function_9_12(unsigned char locobuffer_i)
  {
    pc_send('F');
    pc_send('X');
    pc_send_num(parameter[0]);
    pc_send_num(locobuffer[locobuffer_i].f12_f9);   // nicht ganz intellibox, tuts aber 
    XA_Empty();
  }

void XA_FunctionX(void)
  {
    // *** FX {Lok#, [F9], [F10], [F11], [F12], ...?}
    // Lok#     0..9999 (depending on protocol, not every address is legal!)
    // F9       F9 status (1 = on, 0 = off)
    // etc...
    
    unsigned char i, j, func;

    if (!lok_addr_out_of_range(&parameter[0]) )
      {
        if (parameter[1] == 0xfffe)  // no more parameters -> read LOK 
          {
                // read setting: [F1], [F2], [F3], [F4], ....
            i = scan_locobuffer(&parameter[0]);
            if (i==SIZE_LOCOBUFFER)            // not found - was not used yet
              {
                pc_send_answer_P(Msg_NoLok); return;
              }
            else
              {
                pc_send_lok_function_9_12(i);
                return;
              }
          }
        else
          {    // set lok parameters
               // search locobuffer
            i = scan_locobuffer(&parameter[0]);
            if (i==SIZE_LOCOBUFFER)            // not found - was not used yet
              {
                // new lok -> set defaults
                // wer macht das?
                func = 0;
              }
            else
              { // load old values
                func = locobuffer[i].f12_f9;
              }

            for (j=0; j<4; j++)
              {
                if (parameter[j] == 0)
                  {
                    unsigned char mymask = 1 << j;
                    func = func & ~mymask;
                  }
                if (parameter[j] == 1)
                  {
                    unsigned char mymask = 1 << j;
                    func = func | mymask;
                  }
              }

            // now store the new functions
            do_loco_func_grp3(parameter[0], func);
            pc_send_lok_function_9_12(i);
          }
      }
    else
      {
        pc_send_answer_P(Msg_BadPara);
      }
  }
  

void XA_SpecialOption(void)
  {
    // *** SO {number, [value]}
    // Allows reading the value of a Special Option.
    // if a number is given: set this option (in OpenDCC, this is a eeprom address)
    // Legal range for 'number' is 0..999.
    // SO:      Meaning
    //  1       Baudrate
    //  2       Virtual Decoder enabled
    // 2-99             Reserved
    // 100-227  locobuffer, different types (addr low, addr high + format)
    //          DMX virtual decoder

    if (parameter[0] > 1024)  
          {
            pc_send_answer_P(Msg_BadPara); return;
          }
    if (parameter[1] == 0xfffe)  // no more parameters -> read SO
      {
        pc_send_string_P(Msg_SO);
        pc_send_num(parameter[0]);
        pc_send_num(eeprom_read_byte((void *) parameter[0]) );
        XA_Empty(); return;
      }
    else
      {
        if (parameter[1] > 255)  
          {
            pc_send_answer_P(Msg_BadPara); return;
          }
        else
          {
            eeprom_write_byte((void *)parameter[0], parameter[1]);
          }
      }     
  }





//--------------------------------------------------------------------------------
//
// PoM-Befehle

void XA_PoM_Lok(void)
  {
    // *** PD {addr (1..9999), CV# (1..1024), value (0..255)}
    // Program (write) a Lok decoder using operations mode (main line) programming.
    // (The Intellibox uses the specified address value in order to
    // tell whether to use a 7-bit (1..127) or a 14-Bit (128..9999) packet).
    // N.B. Long addresses 0..127 are legal 14-bit addresses,
    // however the Intellibox uses 7-bit packets for address range 1..127 - and
    // address 0 (a legal 14-bit address) is not supported by the Intellibox
    // for DCC decoders.
    // (Long addresses in range 10000..10239 are legal and supported.)

    if (!lok_addr_out_of_range(&parameter[0]) )
      {
        if (parameter[1] < 1024)
          {
            if (do_pom_loco(parameter[0], parameter[1], parameter[2]) )
              {
                XA_Empty(); return;
              }
          }
      }
    pc_send_answer_P(Msg_BadPara); return;
  }


void XA_PoM_Acc(void)
  {
    // *** PA {addr (0..510), CV# (1..1024), value (0..255)}
    // Program (write) an Accessory decoder using operations mode (main line)
    // programming.
    // N.B. The ***decoder*** address is to be specified, NOT the turnout
    //      address! (Decoder address = (Turnout address - 1) / 4 + 1)
    // If the IB buffer for operations mode commands is not free, this cmd
    // replies with 'Busy!'.

    if (parameter[0] < 511)
      {
        if (parameter[1] < 1024)
          {
            if (do_pom_accessory(parameter[0], parameter[1], parameter[2]) )
              {
                XA_Empty(); return;
              }
          }
      }
    pc_send_answer_P(Msg_BadPara); return;
  }



//--------------------------------------------------------------------------------
//
// PT-Befehle



void XA_ProgTrack(void)
  {
    // *** PT
    // show PT & PT relay status
    // *** PT1/0
    // PT relay On/Off

    if (parameter[0] == 0)                  // parameter = 0 -> leave progmode
      {
        leave_progmode();
        pc_send_string_P(Msg_Prog);
        pc_send_num(0);
        XA_Empty(); return;
      }
    if (parameter[0] == 1)  
      {
        enter_progmode();
        pc_send_string_P(Msg_Prog);
        pc_send_num(1);
        XA_Empty(); return;
      }

    // otherwise make a query

    switch(opendcc_state)                   // this is the current running state
      {
        case RUN_OKAY:            
        case RUN_STOP:                      // DCC Running, all Engines Emergency Stop
        case RUN_OFF:                       // Output disabled (2*Taste, PC)
        case RUN_SHORT:                     // Kurzschluss, hektisch blinken im Wechsel
        case RUN_PAUSE:                     // DCC Running, all Engines Speed 0
            pc_send_string_P(Msg_Prog);
            pc_send_num(0);
            XA_Empty(); return;
            break;
        case PROG_OKAY:
        case PROG_SHORT:           
        case PROG_OFF:
        case PROG_ERROR:        
            pc_send_string_P(Msg_Prog);
            pc_send_num(1);
            XA_Empty(); return;
            break;
      }
  }


/*

######## DCC

*** PTDS
DCC Search: address search (read)
(to be used with 'non-readable' decoders)


*** PTQA
DCC Query Address: address read for very old decoders
(e.g. c82, old Lenz models)


*** PTRR {Register# (1..8)}
DCC Read using register mode


*** PTWR  {Register# (1..8), Value (0..255)}
DCC Write using register mode


*** PTRP {CV# (1..1024)}
DCC Read using paged mode


*** PTWP  {CV# (1..1024), Value (0..255)}
DCC Write using paged mode

*/

void XA_ProgReadByte(void)
  {
    // *** PTRD {CV# (1..1024)}
    // DCC Read using direct (byte) mode 
    // Answer to XPTRD8: => "99 = $63 = %01100011"	
    // is used by Lokprogrammer from Henning Voosen


    if (parameter[0] > 1024)  
      {
        pc_send_answer_P(Msg_BadPara); return;
      }

    my_XPT_DCCRD(parameter[0]);

    // process until done - busy waiting!

    while (programmer_busy())
      {
        run_programmer();                   // dirty - we are busy waiting
      }
    
    if (prog_result == 0)
      {
        pc_send_int(prog_data);
        pc_send(' '); pc_send('='); pc_send(' ');
        pc_send('$');
        pc_send_hex(prog_data);
        pc_send(' '); pc_send('='); pc_send(' ');
        pc_send('%');
        pc_send_bin(prog_data);
        // pc_send_string_P(Msg_Okay);      // this was 0.13  
        // pc_send_num(prog_data);
        XA_Empty();
      }
    else
      {
        pc_send_string_P(Msg_failed);
        pc_send_num(prog_result);
        XA_Empty();
      }
  }


void XA_ProgWriteByte(void)
  {
    // *** PTWD  {CV# (1..1024), Value (0..255)}
    // DCC Write using direct (byte) mode    (Lokprogrammer)
    // is used by Lokprogrammer from Henning Voosen

    if (parameter[0] > 1024)  
      {
        pc_send_answer_P(Msg_BadPara); return;
      }

    if (parameter[1] > 255)  
      {
        pc_send_answer_P(Msg_BadPara); return;
      }
    
    my_XPT_DCCWD(parameter[0], parameter[1]);

    // process until done - busy waiting!

    while (programmer_busy())
      {
        run_programmer();                   // dirty - we are busy waiting
      }
          
    if (prog_result == 0)
      {
        pc_send_string_P(Msg_Okay);
        // pc_send_num(prog_data);
        XA_Empty();
      }
    else
      {
        pc_send_string_P(Msg_failed);
        pc_send_num(prog_result);
        XA_Empty();
      }
  }



/*


*** PTRB {CV# (1..1024)}
DCC Read using direct (bit) mode
(N.B. a **whole** CV (8 bits) is read, not just one bit!)


*** PTWB  {CV#, Bit# (0..7), Value (0..1)}
DCC Write using direct (bit) mode
(N.B. a **single** bit of a CV is written, not the whole CV!)


*** PTQD
DCC Query Direct
(Does the decoder support direct (bit) read commands?)


*** PTRL
DCC Read long address from CV17/18
(N.B. this cmd also **automatically** sets Bit #5 of CV29)
Only compatible with decoders supporting direct bit read and write mode!


*** PTWL {long address (128..9999)}
DCC Write long address into CV17/18
(N.B. this cmd also **automatically** sets Bit #5 of CV29)
Only compatible with decoders supporting direct bit read and write mode!
N.B. Long addresses in range 0..127 are legal 14-bit addresses,
however, since the Intellibox uses 7-bit packets in this
address range, we do not allow these values with the PTWL cmd.
(Long addresses in range 10000..10239 are legal but not supported.)



######## Possible future cmds

*** PTRA
DCC Read Address (auto 7/14-Bit identification and test)


*** PTWA {address (1..9999)}
DCC Write Address (auto 7/14-Bit identification and test)


*** PTID
Automatic decoder identification and address search



######## Related P50Xa cmds (DCC)

*** CV {CV17 (192..231), CV18 (0..255))
Convert CV17, CV18 values into corresponding long address
(N.B. the Intellibox only supports long addresses in range 128..9999)


*** CA {long address (0..10239)}
Convert long address into corresponding CV17, CV18 values
(N.B. the Intellibox only supports long addresses in range 128..9999)



*/




//================================================================================


// Liste aller IBOX, ASCII - Routinen - Zugleich Sprungverteiler
//
// Note: the scan stops after the first match, therefore a single char command
//       must be the last command with this character.

// Typedef fr den Sprungverteiler
typedef struct
{
  unsigned char command[5];
  void (*fnAction)(void);
} T_AACTION;

T_AACTION AsciiParseTab[] PROGMEM =
  {
    // Command / Call                   // implemented / simulated / tested / comment 
    {{"ZZA"},  XA_ZZA },                // i / s / t / we ignore case
    {{"Y"},    XA_ShowStatus },         // i /
    {{"STOP"}, XA_Stop },               // i / -         
    {{"."},    XA_Stop },               // i / -         
    {{"GO"},   XA_Go },                 // i / -         
    {{"!"},    XA_Go },                 // i / - 
    {{"LS"},   XA_Lok_Cfg_Set},         // i / -
    {{"LC"},   XA_Lok_Cfg},             // i / -
//  {{"LM"},   XA_Lok_Consist},         // n / - / - / multi-tractions not supported by IB and OpenDCC
    {{"L"},    XA_Lok},                 // i / - / - / 10-2006: only speed and light, no functions
//  {{"TR"},   XA_TurnoutReserve},      // n / - / - / Turnout Reserv not supported by OpenDCC
    {{"T"},    XA_Turnout},             // i / -
    {{"FX"},   XA_FunctionX},           // - / -
    {{"F"},    XA_Function},            // - / -
    {{"SS"},   XA_S88_read},            // - / -
    {{"SE"},   XA_S88_set_modules},     // - / -
//  {{"SX"},   XA_S88_TimerCountSrc},   // - / - / - / Timers and Counters not supported by OpenDCC
//  {{"ST"},   XA_S88_Timer},           // - / - / - / Timers and Counters not supported by OpenDCC
//  {{"SC"},   XA_S88_Count},           // - / - / - / Timers and Counters not supported by OpenDCC
    {{"SO"},   XA_SpecialOption},       // i / s / - / OpenDCC currently allows for read and write of whole EEPROM
    {{"?"},    XA_Help },               // i / -
    {{"HALT"}, XA_Halt },               // i / -    
    {{"HELP"}, XA_Help },               // i / s / - /
    {{"HL"},   XA_Help_on_Lok },        // i / -
    {{"HT"},   XA_Help_on_Turnout },    // i / -
    {{"HF"},   XA_Help_on_Function },   // i / -
    {{"H"},    XA_Help },               // i / -
    {{"B"},    XA_Baud },               // i / s / - / Baud up to 115k erlaubt
    {{"V"},    XA_ShowVersion },        // i / s / - / only string OpenDCC
//  {{"PAR"},  XA_PoM_Acc_Read },       // *           CV read on the main, accessory decoder (only with BiDi)
//  {{"PDR"},  XA_PoM_Lok_Read },       // *           CV read on the main, lok decoder
    {{"PA"},   XA_PoM_Acc },            // n / - / - / Programm in the main, accessory decoder
    {{"PD"},   XA_PoM_Lok },            // n / - / - / Programm in the main, lok decoder
    {{"PTRD"}, XA_ProgReadByte },       // i /   /   / Read using direct byte mode
    {{"PTWD"}, XA_ProgWriteByte },      // i /   /   / Write using direct byte mode
    {{"PT"},   XA_ProgTrack },          // i /   /   / Turn on or off programming
//  {{"RT"},   XA_SetCTSMode },         // - / - / - / special handling od CTS in case of Power Off, not with OpenDCC
//  {{"CT"},   XA_SetextraCR  },        // - / - / - / no addional leading <cr>, not supported by OpenDCC
//  {{"NR"},   XA_Norefresh  },         // - / - / - / makes no sense - not supported by OpenDCC
//  {{"ID"},   XA_SetIdle },            // - / - / - / OpenDCC does always DCC - not supported
//  {{"PX"},   XA_SetXChar },           // - / - / - / not supported by OpenDCC
//  {{"@@"},   XA_Coldstart },          // - / - / - / only IBOX needs such gimiks, we don't crash ;-))
//  {{"@"},    XA_Warmstart },          // - / - / - / Restart and keep Lok and Turnout Status - not impl. 
    {{""},     XA_unknown},             // i / s / - / empty command is always true
};

//--------------------------------------------------------------------------------
// This routine checks for p50x-opcode and calls the corresponding function
//
void parse_p50x_ascii (void)
  {
    unsigned char *arg;
    unsigned char i, j;
    void (*fp)(void);                   // declare pointer to function


    j=0;
    if (*ibcmd == 0)
      {
        XA_Empty();                     // this command is used as Baudrate Check (really dumb!)
        return;
      }
    while(1)
      {
        arg = ibcmd;                    // reset pointer to arg
        for (i=0; i<5; i++)
          {
            if (pgm_read_byte(&AsciiParseTab[j].command[i]) == 0)   // target found
              {
                ibcmd = ibcmd+i;        // advance parse pointer to arguments

                parse_parameters();
                
                fp = (void *) pgm_read_word(&AsciiParseTab[j].fnAction);   // assign function addr read from flash
                (*fp)();                  // now call this function
                         
                return;
              }
            if (*arg++ != pgm_read_byte(&AsciiParseTab[j].command[i]) )
              {
                break;                // this target is wrong
              }
          }
        j++;
      }
  }

//----------------------------------------------------------------------------------
// this routine parses the old 6050 commands
//
// Note: this is only implemented as far as P50X is affected
// the following list is taken from tams easy control 
//
//
//P50-Commands (M?rklin 6050):
//-----------------------------
//0x00 .. 0x1F: Geschwindigkeit und 'Function' (2 Bytes)
//   Die unteren 4 Bits des ersten Bytes sind die Geschwindigkeit (0 .. 14). Der
//   Wert 15 bedeutet Richtungsumkehr. Das Bit 4 (0x10) schaltet die 'Function'
//   (Licht). Ist es gesetzt so wird das Licht eingeschaltet, sonst aus.
//   Das zweite Byte ist die Lokadresse. G?ltige Adressen: 1 .. 255.
//
//0x20 .. 0x22: Weichen (2 Bytes)
//   Das erste Byte kodiert die Funktion f?r die Weiche: 0x20 = Weichenantrieb
//   aus, 0x21 = Weiche gerade ('gr?n'), 0x22 = Weiche auf Abzweig ('rot').
//   Das zweite Byte ist die Weichenadresse 1 .. 255.
//   Anmerkungen: Das Kommando 0x23 ist eigentlich nicht vorgesehen. Es verh?lt
//   sich jedoch genauso wie 0x20 (Weichenantrieb aus). Im Kommandobereich bis
//   0x3F wiederholen sich die Weichenkommandos alle 4 Werte (0x24, 0x28, ... 0x3C
//   sind also das Gleiche wie 0x20).
//   ACHTUNG: die Magnete sollten sicherheitshalber automatisch abschalten ....
//   (oder nicht???)
//
//0x40 .. 0x4F: Funktionen F1 bis F4 (2 Bytes)
//   Das erste Byte enth?lt in den unteren 4 Bits den neuen Zustand der Funktionen
//   F1 bis F4. Dabei wird F1 durch Bit 0 (0x01) und F4 durch Bit 4 (0x08) darge-
//   stellt.
//   Das zweite Byte ist die Lokadresse 1 .. 255.
//   Anmerkung: Das Kommando wiederholt sich im Bereich 0x50 .. 0x5F.
//   AUSNAHME: 0x58 = 'X' dient als Prefix f?r erweiterte Kommandos (P50x)
//
//0x60 .. 0x61: Start / Stop (1 Byte)
//   0x60 entspricht dem Drcken der Taste "GO", 0x61 entspricht dem Drcken der
//   Taste "STOP". Die Kommandos wiederholen sich alle 2 Kommandowerte bis 0x7F
//   wobei jeder gerade Wert einem "GO" entspricht und jeder ungerade einem
//   "STOP".
//   AUSNAHME: 0x78 = 'x' dient als Prefix f?r Erweiterte Kommandos (P50x)
//
//0x80 .. 0x9F: Ausgabe einer Anzahl S88-Units / S88 Reset aus (1 Byte)
//   Die unteren 5 Bits sind die Anzahl der auszugebenden S88 Module (1 bis 31).
//   Wird als Anzahl 0 angegeben (Kommando 0x80) so wird das S88-Autoreset
//   abgeschaltet.
//
//0xA0 .. 0xBF: unbenutzt (1 Byte)
//
//0xC0 .. 0xDF Ausgabe einer S88-Unit / S88 Reset ein (1 Byte)
//   Die unteren 5 Bits ist die Nummer des auszugebenden S88-Moduls (1 bis 31).
//   Wird als Anzahl 0 angegeben (Kommando 0xC0) so wird das S88-Autoreset
//   eingeschaltet.
//
//0xE0 .. 0xFF: unbenutzt (1 Byte)

void parse_p50(void)
  {
    if (*ibcmd == 0x60)
      { // go command;
        set_opendcc_state(RUN_OKAY);
      }
    else if (*ibcmd == 0x61)
      { // stop command;
        set_opendcc_state(RUN_OFF);
      }
    else if (*ibcmd < 0xC0)
      { // void command;
        pc_send(0x00);
      }
    else if (*ibcmd < 0xE0)
      {
        // S88 readout - Hier nur 0x0, damit Autobaud der IBOX funktioniert.
        pc_send(0x00);
        pc_send(0x00);
      }
    else 
      {
        // unused
      }
  }



//====================================================================================================
/// run_parser: multitask replacement, must be called in loop
//  run_parser frgt die Schnittstelle ab und verzweigt je nach Protokollart in die unterschiedlichen
//  States. Es gibt drei Protokollarten:
//  a) P50:     Kurzbefehle des alten Mrklin 6050, ein oder zwei Bytes lang
//              -> parse_p50 (ist aber nur rudimentrst implementiert)
//  b) P50Xa:   Erweiterung wie Intellibox bzw. Tams Easy Control,
//              Steuerung ber ASCII-Sequenzen;
//              Befehlslnge ist max. 62 Bytes, gefolgt von <cr>
//              Antwort ist Befehlsabhngig, immer temininiert von "<cr>]" 
//  c) P50xb:   Binrbefehle, Befehlslnge gem Tabelle.

// internal, but static:
enum parser_states
  {                            // actual state
     IDLE,                     // Wait for p50 Message
     XIDLE,                    // Waiting for p50X message
     WAIT_FOR_XB_PARAM,        // Wait for P50X Binary Parameters
     WAIT_FOR_XA_PARAM,         // Wait for P50X ASCII
     WAIT_FOR_P50_COMMAND2
  } parser_state;



//------------------------------------------------------------------------
// 

bool input_ready(void)
  {
    if(rs232_break_detected == 1)
      {
        init_parser();
        init_rs232(DEFAULT_BAUD);  // reinit rs232 and get rid of old connection
      }
    return(rx_fifo_ready());
  }


unsigned char input_read(void)
  {
    return(rx_fifo_read());
  }



void run_parser(void)
  {
    switch (parser_state)
      {
        case IDLE:                                      // standard
            if (!input_ready()) return;
            pcc[0] = input_read();                     // read first char
            ibcmd = &pcc[0];
            if (toupper(pcc[0]) == 'X')
              {
                parser_state = XIDLE;
                    return;
              }
            // we got a dumb old P50 Command
            
            if (pcc[0] < 0x60)
              { // command with two bytes
                parser_state = WAIT_FOR_P50_COMMAND2;
              }
            else
              {
                parse_p50();
              }
            break;
        case XIDLE:
            if (!input_ready()) return;
            pcc_index = 1;
            pcc[1] = input_read();                    // read p50x

            if (pcc[1] >= 0x80)
              { // binary command
                pcc_size = get_p50xb_length(pcc[1]);
                if (pcc_size == 0)
                  {
                    // command is unknown - discard
                    if (P50X_mode) parser_state = XIDLE;
                    else parser_state = IDLE;
                  }
                else
                  {
                    no_timeout.parser = 250000L / TICK_PERIOD;   // 250ms - see status.c
                    parser_state = WAIT_FOR_XB_PARAM;
                  }
              }
            else
              {
                no_timeout.parser = 250000L / TICK_PERIOD;   // 250ms - see status.c
                pcc[1]=toupper(pcc[1]);
                ibcmd = &pcc[1];
                parser_state = WAIT_FOR_XA_PARAM;
              }
            break;

        case WAIT_FOR_XA_PARAM:
            if (*ibcmd == 0x0d)
              {
                *ibcmd = 0;                                 // string terminator
                ibcmd = &pcc[1];
                parse_p50x_ascii();
                if (P50X_mode) parser_state = XIDLE;
                else parser_state = IDLE;
                return;
              }
            if (!input_ready()) return;
            if (pcc_index < 63)
              {
                pcc_index++;
                ibcmd++;                                    // advance read pointer
              }
            else 
              {
                // - fatal error - string too long
              }
            *ibcmd = toupper(input_read());                    // read p50xa
            break;


        case WAIT_FOR_XB_PARAM:
            if (pcc_index == pcc_size)
              {
                // message complete, do the job
                parse_p50x_binary (pcc[1]);
                
                if (P50X_mode) parser_state = XIDLE;      // parser_state = P50X_mode?XIDLE:IDLE;  
                else  parser_state = IDLE;
                return;
              }
            if (!input_ready())
              {
                if (no_timeout.parser)  return;
                else
                  {
                    if (P50X_mode) parser_state = XIDLE;
                    else parser_state = IDLE;                 
                    return;
                  }
              }
            pcc_index++;
            pcc[pcc_index] = input_read();
            break;

        case WAIT_FOR_P50_COMMAND2:                      // standard p50, 2 byte commands
            if (!input_ready()) return;
            pcc[1] = input_read();                     // read second char
            ibcmd = &pcc[0];
                        parse_p50();
            parser_state = IDLE;
            break;
     }
  }




void init_parser(void)
  {
    unsigned char mymode;

    mymode = eeprom_read_byte((void *)eadr_startmode_ibox);
    if (mymode == 0)
      {
         parser_state = IDLE;
         P50X_mode = 0;
      }
    else
      {
         parser_state = XIDLE;
         P50X_mode = 1;
      } 
  }


//-------------------------------------------------------------------------------------
#ifdef PARSER_TEST
         
void write_rx(char *c)
  {
    while(*c !=0) 
      {
        push_to_rx(*c++);
      }
  }
  
void test_lokcfg(void)
  { 
    // unsigned char i;
    write_rx("XV\r");

    write_rx("XSO 513\r");
    write_rx("XSO 513 12\r");
    write_rx("XSO 513, 23\r");
    write_rx("XSO 513\r");
    
    write_rx("XV\r");
    write_rx("XSO 513\r");
    write_rx("XSO 512\r");
    
    PORTA=4;

  }

void test(void)
  {
    test_lokcfg();
  }
#endif  // PARSER_TEST


#endif // (PARSER == INTELLIBOX)
