//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006, 2007 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      parser.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2006-04-13 V0.01 started
//            2006-05-19 V0.02 added baud, test of comm an status
//            2006-07-31 V0.03 error in functions call corrected
//            2006-09-30 V0.04 do_loco_speed umbenannt
//                             Umbau der internen Speed auf 0..127
//                             Parser nun als Compilevariante
//            2006-11-19 V0.05 Bugfix beim parser(); hat jetzt locale
//            2007-01-11 V0.06 Lenz scheint rot-grn zu vertauschen
//                             hier mit einer EEPROM-CV (invert_accessory)
//                             gefixed.
//            2007-03-21 V0.07 turnout feedback
//                             !!! noch ungetestet !!!
//            2007-04-20 V0.08 changed programming messages according to log
//                             from Rainer
//            2007-06-08 V0.09 Bugfix in change baud
//            2008-07-23 V0.10 Loco > 100 handled with 0xc000;
//            2008-07-27 V0.11 Change in request accessory info
//
// 2do:       Programming on the main!
//            !!! means in Arbeit!
//            PROGMODE, Broadcasts
//            !!! Speed umsetzung!
//
// issues:    Lenz has no command to set no of s88-modules
//            must do it with cv in eeprom
//
//-----------------------------------------------------------------
//
// purpose:   lowcost central station for dcc
// content:   reads pc-commands from rs232 and generates calls
//            to organizer.c
//
// interface upstream:
//            init_parser(void)         // set up the queue structures
//            run_parser(void)          // multitask replacement, must be called
//                                      // every 20ms (approx)
//            event_send(t_BC_message)
//
// interface downstream:
//            rx_fifo_read()            // to rs232
//            rx_fifo_ready()
//            tx_fifo_write()
//            tx_fifo_full()
//
//            pc_send_single_rm(addr);  // to s88
//
//                                      // to organizer
//
// 2do:       speed umcodieren
//            Schaltbefehle
//            Programmierbefehle
//
//-----------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <string.h>

#include "config.h"                // general structures and definitions
#include "dccout.h"
#include "status.h"                // timeout engine
#include "organizer.h"
#include "rs232.h"
#include "s88.h"



#if (PARSER == LENZ)

#include "lenz_programmer.h"            // service mode
#include "lenz_parser.h"


//-----------------------------------------------------------------
// internal to this module, but static:
enum parser_states
  {                            // actual state
     IDLE,
     WF_MESSAGE,
     WF_XOR
  } parser_state;

unsigned char pcc[16];          // pc_message speicher

unsigned char pcc_size, pcc_index;

unsigned char invert_accessory = 0;  // Lenz seems to invert red and green
                                     // with accessory commands  -->
                                     // we do the same!
                                     // this variable is loaded during init
                                     // from eeprom 


//------------------------------------------------------------------------------
// predefined pc_messages:

unsigned char *pars_pcm;

//-- communication

unsigned char pcm_timeout[] = {0x01, 0x01};                 // Timeout
unsigned char pcm_overrun[] = {0x01, 0x06};                 // too many commands
unsigned char pcm_ack[] = {0x01, 0x04};                     // ack

// generell fixed messages

unsigned char pcm_datenfehler[] = {0x61, 0x80};             // xor wrong
unsigned char pcm_busy[] = {0x61, 0x81};                    // busy
unsigned char pcm_unknown[] = {0x61, 0x82};                 // unknown command
unsigned char pcm_BC_alles_aus[] = {0x61, 0x00};            // Kurzschluabschaltung
unsigned char pcm_BC_alles_an[] = {0x61, 0x01};             // DCC wieder einschalten
unsigned char pcm_BC_progmode[] = {0x61, 0x02};             // Progmode
unsigned char pcm_BC_locos_aus[] = {0x81, 0x00};            // Alle Loks gestoppt

unsigned char pcm_version[] = {0x63, 0x21, 0x30, 0x00};     // LZ100 Zentrale in Version 3.0

// variable messages

unsigned char pcm_status[] = {0x62, 0x22, 0x45};                  //wird von pc_send_status gebaut.

unsigned char pcm_build[] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};  // max 6 bytes




//-------------------------------------------------------------------------------
//
//   Support Routines for parser
//
//--------------------------------------------------------------------------------
//
/// pc_send_lenz:   generell purpose send answer to pc
/// pc_send_lenz_P: generell purpose send answer stored in program memory to pc

void pc_send_lenz(unsigned char *str)      // vorlufig kein Timeout
  {
    unsigned char n, total, my_xor;
 
    n = 0;
    my_xor = str[0];
    total = str[0] & 0x0F;
    
    while (!tx_fifo_ready()) ;             // busy waiting!

    tx_fifo_write(str[0]);                   // send header

    while (n != total)
      {
         n++;
         my_xor ^= str[n];
         tx_fifo_write(str[n]);              // send data
      }    

    tx_fifo_write(my_xor);                   // send xor

  }





//-----------------------------------------------------------------------------------
// Interface fr events, die in der Zentrale passieren und an den PC gemeldet werden.
// Dies besteht aus 2 Typen:
// a) Zustandsnderungen
// b) Rckmeldungen
//
// a) Zustandsnderungen

void event_send(t_BC_message mess)
  {
    switch(mess)
          {
            case ALL_OFF:
                pc_send_lenz(pars_pcm = pcm_BC_alles_aus);  
                break;
            case ALL_ON:
                pc_send_lenz(pars_pcm = pcm_BC_alles_an);
                break;
            case ALL_LOCO_OFF:
                pc_send_lenz(pars_pcm = pcm_BC_locos_aus);  
                break;
            case PROGMODE:
                pc_send_lenz(pars_pcm = pcm_BC_progmode);  
                break;
          }
  }


// b) Rckmeldungen  (ausgelagert nach s88.c)

// void pc_send_rueckmeldung(void)             // broadcast: rckmeldung siehe S88.c


// Alle Broadcast messages aus status werden hier abgesendet.



void pc_send_status(void)
  {
    // Format: Headerbyte Daten 1 Daten 2 X-Or-Byte
    // Hex : 0x62 0x22 S X-Or-Byte
    // S:
    // Bit 0: wenn 1, Anlage in Nothalt
    // Bit 1: wenn 1, Anlage in Notaus
    // Bit 2: Zentralen-Startmode (0 = manueller Start, 1 = automatischer Start)
    // Bit 3: wenn 1, dann Programmiermode aktiv
    // Bit 4: reserviert
    // Bit 5: reserviert
    // Bit 6: wenn 1, dann Kaltstart in der Zentrale
    // Bit 7: wenn 1, dann RAM-Check-Fehler in der Zentrale
    // Besonderheiten: siehe bei Lenz
    unsigned char my_status = 0;
    if (opendcc_state == RUN_STOP) my_status |= 0x01;
    if (opendcc_state == RUN_OFF) my_status |= 0x02;
    // my_status &= ~0x04;  // manueller Start
    if ( (opendcc_state == PROG_OKAY)
       | (opendcc_state == PROG_SHORT)
       | (opendcc_state == PROG_OFF)
       | (opendcc_state == PROG_ERROR) ) my_status |= 0x08;          // Programmiermode
    my_status |= 0x40;  // wir behaupten mal "Kaltstart"
    pcm_status[2] = my_status;
    pc_send_lenz(pars_pcm = pcm_status);
  }


static unsigned char scan_locobuffer(unsigned int addr)       // Hilfsroutine: locobuffer durchsuchen
  {
    register unsigned char i;
    
    // im locobuffer nach dieser Lok suchen
    
    for (i=0; i<SIZE_LOCOBUFFER; i++)
      {
        if (locobuffer[i].address == addr)
          {
            return(i);
          }
      }
    return (SIZE_LOCOBUFFER);   // max if not found
  }


void pc_send_lokdaten(unsigned int addr)
  {
    register unsigned char i, data;
    register unsigned char speed;
    
    i = scan_locobuffer(addr);

    
    
    pcm_build[0] = 0xE4;               // Headerbyte = 0xE4
    pcm_build[1] = 0x00;               // Byte1 = Kennung = 0000BFFF:  B=0: nicht besetzt
                                       //   FFF=Fahrstufen: 000=14, 001=27, 010=28, 100=128
    if (i==SIZE_LOCOBUFFER)            // not found - was not used yet
      {
        pcm_build[1] |= get_loco_format(addr);  // ask eeprom about speed steps
        pcm_build[2] = 0;                      // no Speed
        pcm_build[3] = 0;                      // no functions
        pcm_build[4] = 0; 
      }
    else
      {
        speed = convert_speed_to_rail(locobuffer[i].speed, locobuffer[i].format);
        switch(locobuffer[i].format)
          {
            case DCC14:
                pcm_build[2] = speed;    //Byte2 = Speed = R000 VVVV;
                break;
            case DCC27:
                pcm_build[1] |= 0b001;
                if (speed < 1)
                  {
                    pcm_build[2] = speed; 
                  }
                else
                  {          
                    data = (speed & 0x1F) + 2;    // map internal speed 2..29 to external 4..31
                    data = (data>>1) | ((data & 0x01) <<4);
                    pcm_build[2] = data | (speed & 0x80); 
                  }
                break;
            case DCC28:
                pcm_build[1] |= 0b010;           
                if (speed < 1)
                  {
                    pcm_build[2] = speed; 
                  }
                else
                  {          
                    data = (speed & 0x1F) + 2;    // map internal speed 2..29 to external 4..31
                    data = (data>>1) | ((data & 0x01) <<4);
                    pcm_build[2] = data | (speed & 0x80); 
                  }
                break;
            case DCC128:
                pcm_build[1] |= 0b100;          
                pcm_build[2] = speed;    //Byte2 = Speed = RVVV VVVV;
                break;
          }
        pcm_build[3] = (locobuffer[i].fl << 4) | locobuffer[i].f4_f1;
        pcm_build[4] = (locobuffer[i].f12_f9 << 4) | locobuffer[i].f8_f5;
        //F0: Zustand der Funktionen 0 bis 4. 0 0 0 F0 F4 F3 F2 F1
        //F1: Zustand der Funktionen 5 bis 12 F12 F11 F10 F9 F8 F7 F6 F5
      }
    pc_send_lenz(pars_pcm = pcm_build);
  }


///------------------------------------------------------------------------------
/// parse_commands analyzes the command received by run_parser()
/// currently only the most important commands are implemented
//

void parse_command(void)
//---------------------------------------- Befehlsliste lt. Lenz
// x _ Alles An 0x21 0x81 0xA0
// x _ Alles Aus 0x21 0x80 0xA1
// x _ Alle Loks anhalten 0x80 0x80
// x _ Eine Lok anhalten ab V3 0x92 ADR High ADR Low X-Or
// x _ Prog.-Lesen Registermode 0x22 0x11 REG X-Or
// x _ Prog.-Lesen CV 0x22 0x15 CV X-Or
// x _ Prog.-Lesen Paging 0x22 0x14 CV X-Or
// x _ Prog.-Ergebnis anfordern 0x21 0x10 0x31
// x _ Prog.-Schreiben Register 0x23 0x12 REG DAT X-Or
// x _ Prog.-Schreiben CV 0x23 0x16 CV DAT X-Or
// x _ Prog.-Schreiben Paging 0x23 0x17 CV DAT X-Or
// x _ Softwareversion anfordern 0x21 0x21 0x00
// x _ Status Zentrale anfordern 0x21 0x24 0x05
// x _ Schaltinformation anfordern 0x42 ADR Nibble X-Or
// x _ Schaltbefehl 0x52 ADR DAT X-Or
// x _ Lokdaten anfordern ab V3 0xE3 0x00 ADR High ADR Low X-Or
// x _ Fkt-Status anfordern ab V3 0xE3 0x07 ADR High ADR Low X-Or
// x _ Lok Fahrbefehl ab V3 0xE4 Kennung ADR High ADR Low Speed X-Or
// x _ Lok Funktionsbefehl ab V3 0xE4 Kennung ADR High ADR Low Gruppe X-Or
// x _ Funktionsstatus setzen ab V3 0xE4 Kennung ADR High ADR Low Gruppe X-Or
// x _ DTR-Befehle ab V3 0xE5 0x43 ADR1 H ADR1 L ADR2 H ADR2 L X-Or
// x _ Prog. on Main Byte ab V3 0xE6 0x30 ADR High ADR Low 0xEC + C CV DAT X-Or
// x _ Prog. on Main Bit ab V3 0xE6 0x30 ADR High ADR Low 0xE8 + C CV DAT X-Or
// x _ Lok zu MTR hinzufgen ab V3 0xE4 0x40 + R ADR High ADR Low MTR X-Or
// x _ Lok aus MTR entfernen ab V3 0xE4 0x42 ADR High ADR Low MTR X-Or
// x _ Adresssuche Lok in Mtr ab V3 0xE4 0x01 + R MTR ADR High ADR Low X-Or
// x _ Adresssuche MTR ab V3 0xE2 0x03 + R MTR X-Or
// x _ Stacksuche Lok ab V3 0xE3 0x05 + R ADR High ADR Low X-Or
// x _ Lok aus Stack lschen ab V3 0xE3 0x44 ADR High ADR Low X-Or
// !!! = not yet implemented


  {
    unsigned int addr;
    unsigned char speed = 0;
    t_format format;
        
    
    switch(pcc[0] / 16)   // this is opcode
      {
       case 0x2:
           switch(pcc[1])
             {
               case 0x10: 
                   // Prog.-Ergebnis anfordern 0x21 0x10 0x31
                   if (opendcc_state >= PROG_OKAY)
                     {
                       lprog_send_prog_result_again();
                       return;
                     }
                   // else: Command void -> end of case
                   break;
               case 0x11: 
                   // Prog.-Lesen Register 0x22 0x11 REG X-Or
                   // REG contains the Resister (1...8), this Command has no answer
                   lprog_read_register(pcc[2]);
                   pc_send_lenz(pars_pcm = pcm_BC_progmode);    // a log shows 2 BC-messages; we do one here, 
                                                                // the other from state engine
                
                   // pc_send_lenz(pars_pcm = pcm_ack);
                   return;
                   break;
               case 0x12: 
                   // Prog.-Schreiben Register 0x23 0x12 REG DAT X-Or
                   lprog_write_register(pcc[2], pcc[3]);
                   pc_send_lenz(pars_pcm = pcm_BC_progmode);    // a log shows 2 BC-messages; we do one here, 
                                                                // the other from state engine
                   return;
                   break;
               case 0x14:
                   // Prog-lesen Pagemode Hex : 0x22 0x14 CV X-Or-Byte
                   lprog_read_paged(pcc[2]);
                   pc_send_lenz(pars_pcm = pcm_BC_progmode);    // a log shows 2 BC-messages; we do one here, 
                                                                // the other from state engine
                   return;
                   break;
               case 0x15: 
                   // Prog.-Lesen CV 0x22 0x15 CV X-Or
                   lprog_read_cv(pcc[2]);
                   pc_send_lenz(pars_pcm = pcm_BC_progmode);    // a log shows 2 BC-messages; we do one here, 
                                                                // the other from state engine
                   return;
                   break;
               case 0x16: 
                   // Prog.-Schreiben CV 0x23 0x16 CV DAT X-Or
                   lprog_write_cv(pcc[2], pcc[3]);
                   pc_send_lenz(pars_pcm = pcm_BC_progmode);    // a log shows 2 BC-messages; we do one here, 
                                                                // the other from state engine
                   return;  // no answer to host - doesnt work 18.05.2007
                   break;
               case 0x17: 
                   // Prog.-Schreiben Paging 0x23 0x17 CV DAT X-Or
                   lprog_write_paged(pcc[2], pcc[3]);
                   pc_send_lenz(pars_pcm = pcm_BC_progmode);    // a log shows 2 BC-messages; we do one here, 
                                                                // the other from state engine
                   return;
                   break;
               case 0x21:
                   // Softwareversion anfordern 0x21 0x21 0x00
                   pc_send_lenz(pars_pcm = pcm_version);
                   return;
                   break;
               case 0x24:
                   // Status Zentrale anfordern 0x21 0x24 0x05 ---> wird von TC benutzt!
                   pc_send_status();                          // Statusbyte zurckliefern 
                   return;            
                   break;
               case 0x80:
                   // Alles Aus 0x21 0x80 0xA1 ---> Nothalt
                   pc_send_lenz(pars_pcm = pcm_ack);    // buggy?
                   set_opendcc_state(RUN_OFF);
                   // pc_send_lenz(pars_pcm = pcm_BC_alles_aus);  // passiert mit state
                   return;
                   break;
               case 0x81:
                   // Alles An 0x21 0x81 0xA0
                   pc_send_lenz(pars_pcm = pcm_ack);    // buggy?
                   set_opendcc_state(RUN_OKAY);    
                   // pc_send_lenz(pars_pcm = pcm_BC_alles_an);    // das wird bereits von set_opendcc_state gemacht
                   return;
                   break;
             }
           break;

        case 0x4:
            // not yet tested: Schaltinformation anfordern 0x42 ADR Nibble X-Or
            // Hex : 0x42 Adresse 0x80 + N X-Or-Byte
            // fr Weichen: Adresse = Adr / 4; N=Nibble
            // fr Rckmelder: Adresse = Adr / 8; N=Nibble
            // Antwort:
            // Hex : 0x42 ADR ITTNZZZZ X-Or-Byte
            // ADR = Adresse mod 4
            // I: 1=in work; 0=done -> bei uns immer 0
            // TT = Type: 00=Schaltempf. 01=Schaltempf. mit RM, 10: Rckmelder, 11 reserved
            // N: 0=lower Nibble, 1=upper
            // ZZZZ: Zustand; bei Weichen je 2 Bits: 00=not yet; 01=links, 10=rechts, 11:void
            // Bei Rckmeldern: Direkt die 4 Bits des Nibbles
            // TC interpretiert das alles als direkt bereinanderliegend;
            // also hier und in s88.c folgende Notlsung:
            //  Adressen 0..63  werden als DCC-Accessory interpretiert, aus dem Turnout-Buffer geladen
            //                  und mit TT 01 oder 00 quittiert. Das bedeutet 256 mgliche Weichen
            //  Adressen 64-127 werden als Feedback interpretiert, das bedeutet 512 mgliche Melder

            if (pcc[1] < 0x40)
              {
                // Nur fr Schaltinfo:
                addr = (pcc[1] << 2) + (unsigned char)((pcc[2] & 0x01) << 1);
                pcm_build[0] = 0x42;
                create_xpressnet_schaltinfo(addr, &pcm_build[1]);
                pc_send_lenz(pars_pcm = pcm_build);           
              }
            else
              { // request feedback info, shift addr locally down
                addr = ((pcc[1] & 0x4f) << 3) + ((pcc[2] & 0x01) << 2);
                pcm_build[0] = 0x42;
                create_xpressnet_feedback(addr, &pcm_build[1]);
                pc_send_lenz(pars_pcm = pcm_build);
              }
            return;
            break;

        case 0x5:
            // Schaltbefehl 0x52 ADR DAT X-Or
            // Hex: 0x52 Adresse 0x80 + SBBO; 
            // Adresse: = Decoder;
            // S: 1=activate, 0=deactivate,
            // BB=local adr,
            // O=Ausgang 0 (red) / Ausgang 1 (grn)
            // (das wrde eigentlich schon passend fr DCC vorliegen, aber lieber sauber bergeben)
            addr = (unsigned int) (pcc[1] << 2) + ((pcc[2] >> 1) & 0b011);
            unsigned char activate, coil;
            activate = (pcc[2] & 0b01000) >> 3;
            coil = pcc[2] & 0b01;
            coil = coil ^ invert_accessory;
            do_accessory(addr, coil, activate);
            pc_send_lenz(pars_pcm = pcm_ack);
            if (pcc[1] < 0x40)                              // only xpressnet feedback if < 256
              { // 28.07.2008
                pcm_build[0] = 0x42;
                create_xpressnet_schaltinfo(addr, &pcm_build[1]);
                pc_send_lenz(pars_pcm = pcm_build);
              }
            return;
            break;

       case 0x8:
           // Alle Loks anhalten 0x80 0x80
           do_all_stop();                                   // from organizer.c   
           pc_send_lenz(pars_pcm = pcm_ack);
           return;
           break;
        case 0x9:
            // Eine Lok anhalten ab V3 0x92 ADR High ADR Low X-Or (Nothalt)
            addr = (pcc[1] & 0xc0) * 256 + pcc[2];
            pc_send_lenz(pars_pcm = pcm_ack);

            // format dieser Lok rausfinden
            register unsigned char i;
            i = scan_locobuffer(addr);
            if (i==SIZE_LOCOBUFFER)     // not found
              { 
                format = get_loco_format(addr);
              }
            else
              {
                format = locobuffer[i].format;
              }
            if (do_loco_speed_f(addr, 1, format))   // speed = 1: Nothalt!
              {
                 // pc_send_lenz(wait);
              }    
            // !!! fehlt Frage, ob voll!! und wie starten wir wieder?
            return;
            break;
        case 0xE:
            switch(pcc[1] & 0xf0)      // high nibble von pcc[1]:
              {
                case 0x00:             // 0: Adresssuche, Lokdaten anfordern, Fct anfordern, Stacksuche
                    // Lokdaten anfordern ab V3 0xE3 0x00 ADR High ADR Low X-Or
                    addr = (pcc[1] & 0xc0) * 256 + pcc[2];
                    pc_send_lokdaten(addr);
                    return;
                    break;
                case 0x01:            // !!! Adresssuche Lok in Mtr ab V3 0xE4 0x01 + R MTR ADR High ADR Low X-Or
                    break;
                    // !!! Adresssuche MTR ab V3 0xE2 0x03 + R MTR X-Or
                    // !!! Stacksuche Lok ab V3 0xE3 0x05 + R ADR High ADR Low X-Or
                    // !!! Fkt-Status anfordern ab V3 0xE3 0x07 ADR High ADR Low X-Or (tastend-nicht tastend)
                case 0x10:           
                    // !!! Lok Fahrbefehl ab V3 0xE4 Kennung ADR High ADR Low Speed X-Or
                    // if (pcc[1] & 0b00010000 == 0b00010000)
                    format = pcc[1] & 0x03;   // 0=14, 1=27, 2=28, 3=128 see t_format Definition
                    switch(format)
                      {
                        case DCC14:
                            speed = (pcc[4] & 0x80) | (pcc[4] & 0x0F);   
                            break;
                        case DCC27:
                        case DCC28:
                            if ((pcc[4] & 0x0F) <= 1)               // map 0x?0 to 0 and 0x?1 to 1
                                 speed = pcc[4] & 0x81;             // stop or nothalt
                            else 
                              {
                                speed = ((pcc[4] & 0x0F) << 1) | ((pcc[4] & 0x10) >> 4);
                                speed = speed - 2;                  // map 4..31 to 2..29
                                speed = speed | (pcc[4] & 0x80);    // direction
                              }
                            break;
                        case DCC128:
                            speed = pcc[4];
                            break;
                      }
                    addr = (pcc[1] & 0xc0) * 256 + pcc[2];
                    if (organizer_ready())
                      {
                        unsigned char myspeed;
                        myspeed = convert_speed_from_rail(speed, format); // map lenz to internal 0...127
                        do_loco_speed_f(addr, myspeed, format);
                        pc_send_lenz(pars_pcm = pcm_ack);
                      }
                    else
                      {
                        pc_send_lenz(pars_pcm = pcm_busy);
                      }
                    return;
                    break;
                case 0x20:
                    addr = (pcc[1] & 0xc0) * 256 + pcc[2];
                    switch(pcc[1] & 0x0F)   // !!! Lok Funktionsbefehl ab V3 0xE4 Kennung ADR High ADR Low Gruppe X-Or
                      {
                        case 0:          // Hex : 0xE4 0x20 AH AL Gruppe 1 X-Or-Byte   (Gruppe 1: 000FFFFF) f0, f4...f1
                            if (organizer_ready())
                              {
                                do_loco_func_grp0(addr, pcc[4]>>4); // light, f0
                                do_loco_func_grp1(addr, pcc[4]);
                                pc_send_lenz(pars_pcm = pcm_ack);
                              }
                            else
                              {
                                pc_send_lenz(pars_pcm = pcm_busy);
                              }
                            return;
                            break;
                        case 1:          // Hex : 0xE4 0x21 AH AL Gruppe 2 X-Or-Byte   (Gruppe 2: 0000FFFF) f8...f5
                            if (organizer_ready())
                              {
                                do_loco_func_grp2(addr, pcc[4]);
                                pc_send_lenz(pars_pcm = pcm_ack);
                              }
                            else
                              {
                                pc_send_lenz(pars_pcm = pcm_busy);
                              }
                            return;
                            break;
                        case 2:          // Hex : 0xE4 0x22 AH AL Gruppe 3 X-Or-Byte   (Gruppe 3: 0000FFFF) f12...f9
                            if (organizer_ready())
                              {
                                do_loco_func_grp3(addr, pcc[4]);
                                pc_send_lenz(pars_pcm = pcm_ack);
                              }
                            else
                              {
                                pc_send_lenz(pars_pcm = pcm_busy);
                              }
                            return;
                            break;
                        case 4:         // !!! Funktionsstatus setzen ab V3 0xE4 Kennung ADR High ADR Low Gruppe X-Or
                                        // Hex : 0xE4 0x24 AH AL Gruppe 1 (0000SSSS)  S=1: Funktion ist tastend
                                        // Hex : 0xE4 0x25 AH AL Gruppe 2 (0000SSSS)
                                        // Hex : 0xE4 0x26 AH AL Gruppe 3 (0000SSSS)
                            break;
                      }
                    break;
                case 0x30:
                    // !!! Prog. on Main Byte ab V3 0xE6 0x30 ADR High ADR Low 0xEC + C CV DAT X-Or
                    // !!! Prog. on Main Bit ab V3 0xE6 0x30 ADR High ADR Low 0xE8 + C CV DAT X-Or
                    break;
                case 0x40:   //Lokverwaltung
                    // !!! Lok zu MTR hinzufgen ab V3 0xE4 0x40 + R ADR High ADR Low MTR X-Or
                    // !!! Lok aus MTR entfernen ab V3 0xE4 0x42 ADR High ADR Low MTR X-Or
                    // !!! DTR-Befehle ab V3 0xE5 0x43 ADR1 H ADR1 L ADR2 H ADR2 L X-Or
                    // !!! Lok aus Stack lschen ab V3 0xE3 0x44 ADR High ADR Low X-Or
                    break;
              }
            break;
        case 0xF:
            switch(pcc[1])
              {
                case 0x02:    // setze Baud (getestet 19.05.2006)
                             // Antwort: F2 02 Baud, wie Aufruf (Antwort noch in der alten Baudrate, dann umschalten)
                             // BAUD = 1 19200 baud (Standardeinstellung)
                             // BAUD = 2 38400 baud
                             // BAUD = 3 57600 baud
                             // BAUD = 4 115200 baud
                             // nach BREAK (wird als 000) empfangen sollte Interface default auf auf 19200
                             // schalten
                if ((pcc[2] < 1) && (pcc[2] > 4)) pcc[2] = 1;
                pc_send_lenz(&pcc[0]);

                while(!tx_all_sent());                  // busy waiting until all sent
                        
                init_rs232(pcc[2]);             // jetzt umschalten und fifos flushen
                return;
              }
      }
    pc_send_lenz(pars_pcm = pcm_unknown);   // wer bis hier durchfllt, ist unbekannt!
  }



//====================================================================================================
/// run_parser: multitask replacement, must be called in loop

void run_parser(void)
  {
    unsigned char i, my_check;

    if (s88_event & (1<<S88_HW_CHANGE)) 
      {
        make_s88_lenz_report();
      }
    switch (parser_state)
      {
        case IDLE:
            if (!rx_fifo_ready()) return;
            pcc[0] = rx_fifo_read();                     // read header
            pcc_size = pcc[0] & 0x0F;
            pcc_index = 0;                             // message counter
            no_timeout.parser = 250000L / TICK_PERIOD;   // 250ms - see status.c
            parser_state = WF_MESSAGE;
            break;

       case WF_MESSAGE:
            if (pcc_index == pcc_size)
              {
                parser_state = WF_XOR;
                return;
              }
            if (!rx_fifo_ready())
              {
                if (no_timeout.parser)  return;
                else
                  {
                    parser_state = IDLE;
                    pc_send_lenz(pars_pcm = pcm_timeout);      // throw exception, if timeout reached
                    return;
                  }
              }
            pcc_index++;
            pcc[pcc_index] = rx_fifo_read();
            break;

       case WF_XOR:
            if (!rx_fifo_ready())
              {
                if (no_timeout.parser)  return;
                else
                  {
                    parser_state = IDLE;
                    pc_send_lenz(pars_pcm = pcm_timeout);        // throw exception, if timeout reached
                    return;
                  }
              }
            my_check = 0;  
            for (i=0; i<=pcc_size; i++) my_check ^= pcc[i];   
            if (my_check != rx_fifo_read())
              {
                // XOR is wrong!
                pc_send_lenz(pars_pcm = pcm_datenfehler);
                parser_state = IDLE;
                return;
              }
           
            parse_command();       // analyze received message and send code
            parser_state = IDLE;
            break;
     }
  }



void init_parser(void)
  {
    parser_state = IDLE;
    invert_accessory = eeprom_read_byte((void *)eadr_invert_accessory); 
  }

#endif // (PARSER == LENZ)
