//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      programmer.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2006-06-02 V0.1 started
//            2006-10-24 V0.2 bugfixes in register mode and
//                            write routines (reported by Volker Bosch)
//            2007-01-24 V0.3 IB routines removed and put
//                            in ibox_programmer.c
//            2007-04-05 V0.4 added programmer_busy()
//            2007-04-20 V0.5 test with Train Programmer successfully
//            2008-01-28 V0.6 key_go_at_start_pressed added
//            
//           
//-----------------------------------------------------------------
//
// purpose:   lowcost central station for dcc
// content:   genereates the dcc servicemode messages and checks for
//            acknowledge.
//
// interface upstream:
//            init_programmer(void)     // 
//
// interface downstream:
//
//                                      // to organizer
//
// 2do:       
//-----------------------------------------------------------------

// LINK:      http://ruppweb.dyndns.org/xray/comp/decoder.htm
//            (calculate long addresses)

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <avr/pgmspace.h>        // put var to program memory

#include <string.h>

#include "config.h"                // general structures and definitions
#include "dccout.h"                // next message

#include "status.h"                // timeout engine, set_state
#include "organizer.h"

#if (PARSER == LENZ)

#include "lenz_parser.h"
#include "lenz_programmer.h"



#define NEW_PROG 1
    


//=====================================================================
//
// Data Structures
//
//--------------------------------------------------------------------
// internal, but static:

enum prog_seq_states
  {
     PS_IDLE,                   // fertig, Ergebnis kann abgeholt werden
     PS_START,                  // Call 
     PS_RUNNING                    
  } prog_seq_state;

enum
  {
    PSC_DCCRD,                  // read mit byte verify in loop

  } ps_command;                 // hier wird das Kommand hinterlegt.

unsigned int ps_cv;             // aktuelle cv;
  
unsigned char ps_step;          // lokaler Programmschritt 

enum programmer_states
  {                            // actual state for this CV
     IDLE,
     DO_1ST_RESET,
     DO_PAGE_PRESET,
     DO_2ND_RESET,
     DO_PROG_MESSAGE,
     SETUP_3RD_RESET,
     DO_3RD_RESET
  } programmer_state;


struct t_prog_result                    // Bitfeld: MSB: result ready, LSB: ack received
  {
    unsigned char ack_received: 1;    // LSB = ack_received;counter for repeat or refresh
    unsigned char reserved: 6;        // 2 .. 7
    unsigned char result_ready: 1;    // MSB  .
  } prog_result;


// Overview of Programming Modes:
//
// Mode:        Leading PagePre Mid     Command Post
//              resets          resets          resets (r/w)
//-------------------------------------------------------------------
// Direct         3       -       -       5       1 / 6
// Addr, only     3       5       9       7       1 / 10
// Reg. Mode      3       5       9       7       1 / 10
// Paged Mode     same as Reg. Mode, but page access in advance        
//

typedef enum {P_IDLE, P_WRITE, P_READ}  t_prog_mode;

typedef struct                      // this is the structure where we handle different modes
  {
    t_prog_mode mode;
    unsigned char cycles[5];        // we have 5 different cycles - see table above
  } t_prog_ctrl;

t_prog_ctrl prog_ctrl;           
 
t_prog_ctrl direct_ctrl PROGMEM       = {P_READ, {3, 0, 0, 5,  6}};
t_prog_ctrl registermode_ctrl PROGMEM = {P_READ, {3, 5, 9, 7, 10}};


//--------------------------------------------------------------------
//                            rep, size, type, DCC
t_message pDCC_Reset     = {1, {{ 2, is_void}}, {0x00, 0x00}};    // DCC-Reset-Paket
t_message *dcc_reset_ptr = &pDCC_Reset;
t_message prog_message   = {1, {{ 2, is_prog}}, {0x00, 0x00, 0x00, 0x00, 0x00}};    // DCC-Programming
t_message *prog_message_ptr = &prog_message;

t_message page_preset    = {1, {{ 2, is_void}}, {0b01111101, 0b00000001}};
t_message *page_preset_ptr = &page_preset;



//xxxx
//===================================================================================
//
//   Support Routines for programmer
//
//--------------------------------------------------------------------------------
//
//--------------------------------------------------------------------------------------
/// run_programmer: multitask replacement, must be called in loop
//
// howto:   run_programmer runs the inner loop
//
// input:   prog_ctrl must be loaded with the desired operation
//          (mode and no. of cycles for each state)
//
// output:  prog_result is updated, dcc commands are feed to prog_queue
//          
// requires: organizer is running!


void run_programmer(void)
  {
    
    switch (programmer_state)
      {
        case IDLE:
            if (prog_ctrl.mode != P_IDLE)                   // something to do
              {
                prog_result.ack_received = 0;               // clear result
                prog_result.result_ready = 0;               // clear result
                pDCC_Reset.repeat = prog_ctrl.cycles[0];    // send reset packets
                put_in_queue_prog(dcc_reset_ptr);
                programmer_state = DO_1ST_RESET;
              }
            break;

        case DO_1ST_RESET:
            if (!queue_prog_is_empty()) return;

            if (prog_ctrl.cycles[1] > 0)
              {                                             // page presets required
                page_preset.repeat = prog_ctrl.cycles[1];     
                put_in_queue_prog(page_preset_ptr);                                            
              }
            programmer_state = DO_PAGE_PRESET;
            break;

        case DO_PAGE_PRESET:
            if (!queue_prog_is_empty()) return;             

            if (prog_ctrl.cycles[2] > 0)
              {          
                pDCC_Reset.repeat = prog_ctrl.cycles[2];    // send reset packets
                put_in_queue_prog(dcc_reset_ptr);
              }
            programmer_state = DO_2ND_RESET;
            break;

        case DO_2ND_RESET:
            if (!queue_prog_is_empty()) return;

            prog_message.repeat = prog_ctrl.cycles[3];
            put_in_queue_prog(&prog_message);                                            
            programmer_state = DO_PROG_MESSAGE;
            break;

        case DO_PROG_MESSAGE:                           // now wait for ack or timeout
            if (!queue_prog_is_empty()) return;
            // command has been taken, now check for ACK
            if (ACK_IS_DETECTED) 
              {
                unsigned char i;
                for (i=0; i<5; i++)
                  {
                    _mydelay_us(100);                   // ACK must be present for 500us
                    if (!ACK_IS_DETECTED) return;       // Note: with 1ms sometimes read error
                  }

                cli();
                // skip further repetitions -> fool dccout - this is dirty! 
                if (next_message_count>1) next_message_count=1;     
                sei();
                
                prog_result.ack_received = 1;
                programmer_state = SETUP_3RD_RESET;
                return;
              }
                 
            if (next_message_count > 1) return;   // again dirty: we ask the communication flag
                                                  // our message goes with rep 5, so 5...1
                                                  // is our time to wait for ACK; 
         
            programmer_state = SETUP_3RD_RESET;
            break;

        case SETUP_3RD_RESET:
            if (prog_ctrl.mode == P_READ)
                pDCC_Reset.repeat = 1;
            else
                pDCC_Reset.repeat = prog_ctrl.cycles[4];    // if WRITE: send more reset packets
            put_in_queue_prog(dcc_reset_ptr);
            programmer_state = DO_3RD_RESET;
            break;

        case DO_3RD_RESET:
            if (queue_prog_is_empty()) return;

            prog_result.result_ready = 1;                    // tell caller that we are ready
            prog_ctrl.mode = P_IDLE;                         // clear request
            programmer_state = IDLE;
            break;

     }
  }



void init_programmer(void)
  {
    programmer_state = IDLE;
    prog_ctrl.mode = P_IDLE;
  }

/*
void run_prog_sequenzer(void)
  {
  
    switch (prog_seq_state)
      {
        case PS_IDLE:                                       // all done
            break;

        case PS_START:
            // make result void
            // get command
            switch (ps_command)
              {
                case PSC_DCCRD:
                    break;
              }
            // perform

            prog_seq_state = PS_RUNNING;
            break;

        case PS_RUNNING:
            run_programmer();
            if (prog_result.result_ready == 1) prog_seq_state = PS_IDLE;
      }
  }
*/

// return true, if busy
// we are always false - Lenz Programmer is busy waiting

unsigned char programmer_busy(void)
  {
    return(0);
  }

//================================================================================


#if (NEW_PROG == 1)
unsigned char proceed_direct_mode(t_message *myprog, unsigned char recover)
  {
    prog_result.result_ready = 0;
    programmer_state = IDLE;
    while (prog_result.result_ready == 0) run_programmer();    // busy loop
    return(prog_result.ack_received);
  }   

#else
unsigned char proceed_direct_mode(t_message *myprog, unsigned char recover)
  {
     unsigned char retval = 0;
	 
	 while (queue_prog_is_empty() != 0);   // wait for dccout engine

     pDCC_Reset.repeat = 3;             // 3 reset packets
     put_in_queue_prog(dcc_reset_ptr);

     while (queue_prog_is_empty() != 0);   // wait for dccout engine
	 
     myprog->repeat = 5;
     put_in_queue_prog(myprog);
     
     // now wait for ack or timeout

     while (next_message_count != 0)    // hier direkt auf dcc_out warten
       {
         if (ACK_IS_DETECTED) 
           {
             retval = 1;
             next_message_count = 1;   // stop further repetitions - this is dirty! 
           }
	    }

     while (queue_prog_is_empty() != 0);   // wait for dccout engine
     if (recover == 0) pDCC_Reset.repeat = 1;
     else pDCC_Reset.repeat = 6;      
        
     put_in_queue_prog(dcc_reset_ptr);

     return(retval);
  }        

#endif  


/// CV-Mode, write
//
//  Write Paket
//  preamble 0 0111 CCAA 0 AAAAAAAA 0 DDDDDDDD 0 EEEEEEEE 1 mit CC=11

void direct_mode_write(unsigned int cv, unsigned char data)
  {
     unsigned int cv_adr;

     cv_adr = cv-1;
     prog_message.repeat = 1;
     prog_message.size   = 3;
     prog_message.type   = is_prog;
     prog_message.dcc[0] = 0b01110000 | 0b00001100 | (unsigned char)((cv_adr >> 8) & 0b11);
     prog_message.dcc[1] = (unsigned char)(cv_adr & 0xFF);
     prog_message.dcc[2] = data;
     
     memcpy_P (&prog_ctrl, &direct_ctrl, sizeof(prog_ctrl));
     prog_ctrl.mode = P_WRITE;
     
     proceed_direct_mode(prog_message_ptr, 1);    
  }

/// CV-Mode, verify
//
//  Verify Paket
//  preamble 0 0111 CCAA 0 AAAAAAAA 0 DDDDDDDD 0 EEEEEEEE 1 mit CC=01

unsigned char direct_mode_verify(unsigned int cv, unsigned char data)
  {
     unsigned int cv_adr;

     cv_adr = cv-1;
     prog_message.repeat = 1;
     prog_message.size   = 3;
     prog_message.type   = is_prog;
     prog_message.dcc[0] = 0b01110000 | 0b00000100 | (unsigned char)((cv_adr >> 8) & 0b11);
     prog_message.dcc[1] = (unsigned char)(cv_adr & 0xFF);
     prog_message.dcc[2] = data;
     
     memcpy_P (&prog_ctrl, &direct_ctrl, sizeof(prog_ctrl));
     
     if (proceed_direct_mode(prog_message_ptr, 0))
       {
         // ACK received
		 return(1);
       }
	 return(0);    
  }


/// CV-Mode, verify single bit
//
//  Verify Paket
//  preamble 0 0111 CCAA 0 AAAAAAAA 0 111KDBBB 0 EEEEEEEE 1 mit CC=10
//  K = (1=write, 0=verify) D = Bitvalue, BBB = bitpos
//  return: true if verified

unsigned char direct_mode_bit_verify(unsigned int cv, unsigned char bitpos, unsigned char mybit)
  {
     unsigned int cv_adr;
     unsigned char retval = 0;

     cv_adr = cv-1;
     prog_message.repeat = 1;
     prog_message.size   = 3;
     prog_message.type   = is_prog;
     prog_message.dcc[0] = 0b01110000 | 0b00001000 | (unsigned char)((cv_adr >> 8) & 0b11);
     prog_message.dcc[1] = (unsigned char)(cv_adr & 0xFF);
     prog_message.dcc[2] = 0b11100000 | ((mybit &0b1) << 3) | (bitpos & 0b111);
     
     memcpy_P (&prog_ctrl, &direct_ctrl, sizeof(prog_ctrl));
     
     if (proceed_direct_mode(prog_message_ptr, 0))
       {
         retval = 1; // ACK received
       }  
     return(retval);  
  }

/// CV-Mode, write single bit
//
//  Write Paket
//  preamble 0 0111 CCAA 0 AAAAAAAA 0 111KDBBB 0 EEEEEEEE 1 mit CC=10
//  K = (1=write, 0=verify) D = Bitvalue, BBB = bitpos

void direct_mode_bit_write(unsigned int cv, unsigned char bitpos, unsigned char mybit)
  {
     unsigned int cv_adr;

     cv_adr = cv-1;
     prog_message.repeat = 1;
     prog_message.size   = 3;
     prog_message.type   = is_prog;
     prog_message.dcc[0] = 0b01110000 | 0b00001000 | (unsigned char)((cv_adr >> 8) & 0b11);
     prog_message.dcc[1] = (unsigned char)(cv_adr & 0xFF);
     prog_message.dcc[2] = 0b11110000 | ((mybit &0b1) << 3) | (bitpos & 0b111);
     
     memcpy_P (&prog_ctrl, &direct_ctrl, sizeof(prog_ctrl));
     prog_ctrl.mode = P_WRITE;
     
     proceed_direct_mode(prog_message_ptr, 1);
  }



// returns byte;
// returns -1 if verify of byte failed

signed int direct_mode_read(unsigned int cv)
  {
     unsigned char i;
     unsigned char result=0;
     for (i=0; i<8; i++)
       {
         if (direct_mode_bit_verify(cv, i, 1))
           {
             result |= (1<<i);
           }
       }
     if (direct_mode_verify(cv, result)) return(result);
     else return(-1);
  }


unsigned char test_direct_mode(void)
  {
     // test auf MSB von CV8
     if (direct_mode_bit_verify(8, 7, 1)) return(1);
     if (direct_mode_bit_verify(8, 7, 0)) return(1);
     return(0);
  }


//-------------------------------------------------------------------- Address Only Mode


// recover gibt an, dass der Decoder Zeit zum recover bekommt
// notwendig immer nach write-Operationen



#if (NEW_PROG == 1)
unsigned char proceed_addressonly_mode(t_message *myprog, unsigned char recover)
  {
    prog_result.result_ready = 0;
    programmer_state = IDLE;
    while (prog_result.result_ready == 0) run_programmer();    // busy loop
    return(prog_result.ack_received);
  }   

#else
unsigned char proceed_addressonly_mode(t_message *myprog, unsigned char recover)
  {
     unsigned char retval = 0;

     while (queue_prog_is_empty() != 0);   // wait for dccout engine
     
     pDCC_Reset.repeat = 3;               // 3 reset packets
     put_in_queue_prog(dcc_reset_ptr);

     while (queue_prog_is_empty() != 0);   // wait for dccout engine
     page_preset.repeat = 5;             // 5 page preset packets
     put_in_queue_prog(page_preset_ptr);

     while (queue_prog_is_empty() != 0);   // wait for dccout engine
     pDCC_Reset.repeat = 9;               // 3+6 reset packets
     put_in_queue_prog(dcc_reset_ptr);

     myprog->repeat = 7;
     put_in_queue_prog(myprog);
     
     // now wait for ack or timeout

     while (next_message_count != 0)
       {
         if (ACK_IS_DETECTED) 
           {
             retval = 1;
             next_message_count = 1;   // stop further repetitions - this is dirty! 
           }
	   }

    while (queue_prog_is_empty() != 0);   // wait for dccout engine
     if (recover == 0) pDCC_Reset.repeat = 1;
     else pDCC_Reset.repeat = 10;      
        
     put_in_queue_prog(dcc_reset_ptr);

     return(retval);
  }

#endif  


//long-preamble 0 0111C000 0 0DDDDDDD 0 EEEEEEEE 1
//mit: C=0 Verify Address, C=1 Write Address


unsigned char address_mode_verify(unsigned char testadr)
  {
     unsigned char retval = 0;

     prog_message.repeat = 1;
     prog_message.size   = 2;
     prog_message.type   = is_prog;
     prog_message.dcc[0] = 0b01110000;
     prog_message.dcc[1] = (testadr & 0x7F);
     
     memcpy_P (&prog_ctrl, &registermode_ctrl, sizeof(prog_ctrl));


     if (proceed_addressonly_mode(prog_message_ptr, 0))
       {
         retval = 1; // ACK received
       }  
     return(retval);  
  }

unsigned char address_mode_write(unsigned char testadr)
  {
     unsigned char retval = 0;

     prog_message.repeat = 1;
     prog_message.size   = 2;
     prog_message.type   = is_prog;
     prog_message.dcc[0] = 0b01111000;
     prog_message.dcc[1] = (testadr & 0x7F);

     memcpy_P (&prog_ctrl, &registermode_ctrl, sizeof(prog_ctrl));
     prog_ctrl.mode = P_WRITE;
     
     if (proceed_addressonly_mode(prog_message_ptr, 1))
       {
         retval = 1; // ACK received
       }  
     return(retval);  
  }



//-------------------------------------------------------------------- Register Mode
// Note: regadr means binary coded (0..7), regno means coded like nmra (1..8)
// ist identisch zu Adress only


#if (NEW_PROG == 1)
unsigned char proceed_register_mode(t_message *myprog, unsigned char recover)
  {
    prog_result.result_ready = 0;
    programmer_state = IDLE;
    while (prog_result.result_ready == 0) run_programmer();    // busy loop
    return(prog_result.ack_received);
  }   

#else
unsigned char proceed_register_mode(t_message *myprog, unsigned char recover)
  {
     unsigned char retval = 0;

     while (queue_prog_is_empty() != 0);   // wait for dccout engine
     
     pDCC_Reset.repeat = 3;               // 3 reset packets
     put_in_queue_prog(dcc_reset_ptr);

     while (queue_prog_is_empty() != 0);   // wait for dccout engine
     page_preset.repeat = 5;             // 5 page preset packets
     put_in_queue_prog(page_preset_ptr);

     while (queue_prog_is_empty() != 0);   // wait for dccout engine
     pDCC_Reset.repeat = 9;               // 3+6 reset packets
     put_in_queue_prog(dcc_reset_ptr);

     myprog->repeat = 7;
     put_in_queue_prog(myprog);
     
     // now wait for ack or timeout

     while (next_message_count != 0)
       {
         if (ACK_IS_DETECTED) 
           {
             retval = 1;
             next_message_count = 1;   // stop further repetitions - this is dirty! 
           }
	   }

    while (queue_prog_is_empty() != 0);   // wait for dccout engine
     if (recover == 0) pDCC_Reset.repeat = 1;
     else pDCC_Reset.repeat = 10;      
        
     put_in_queue_prog(dcc_reset_ptr);

     return(retval);
  }

#endif  



// long-preamble 0 0111CRRR 0 DDDDDDDD 0 EEEEEEEE 1
// RRR=Register address
// C=0 Verify, C=1 Write register
// RRR=000 ist wie Adress only!
unsigned char register_mode_verify(unsigned char regadr, unsigned data)
  {
    unsigned char retval = 0;

    prog_message.repeat = 1;
    prog_message.size   = 2;
    prog_message.type   = is_prog;
    prog_message.dcc[0] = 0b01110000 | (regadr & 0b111);
    prog_message.dcc[1] = data;

    memcpy_P (&prog_ctrl, &registermode_ctrl, sizeof(prog_ctrl));
    prog_ctrl.mode = P_READ;

    if (proceed_register_mode(prog_message_ptr, 0))
      {
        retval = 1; // ACK received
      }  
    return(retval);
  }

// regadr: 0..7; regno: 1..8;
unsigned char register_mode_write(unsigned char regadr, unsigned data)
  {
     unsigned char retval = 0;

     prog_message.repeat = 1;
     prog_message.size   = 2;
     prog_message.type   = is_prog;
     prog_message.dcc[0] = 0b01111000 | (regadr & 0b111);
     prog_message.dcc[1] = data;

     memcpy_P (&prog_ctrl, &registermode_ctrl, sizeof(prog_ctrl));
     prog_ctrl.mode = P_WRITE;

     prog_result.result_ready = 0;
     programmer_state = IDLE;
     while (prog_result.result_ready == 0) run_programmer();    // busy loop

     return(prog_result.ack_received);
     
     if (proceed_register_mode(prog_message_ptr, 1))
       {
          retval = 1; // ACK received
        }  
     return(retval);  
  }

signed int register_mode_read_byte(unsigned char regadr)
  {
     unsigned int i;

     for (i=0; i<256; i++)
       {
         if (register_mode_verify(regadr, i)) return(i);
       }

     return(-1);   // failed
  }



//-------------------------------------------------------------------- Paged Mode

// paged mode ist wie register mode, nur wird voher in Register 6 die Seite
// geschrieben, und ber Register 1..4 darauf zugegriffen;
//
// CV = 1 + (RRR + 4 * (mypage-1))
//
// This results in rollover at page 256!
// see http://www.users.bigpond.com/pbhandary/dcc/service.html
//
void paged_mode_write(unsigned int cv, unsigned char data)
  {
    unsigned char regdata, mypage;

    // Berechnet mypage und regdata
    mypage = (unsigned char)(((cv-1) >> 2) + 1);   
	regdata = (unsigned char)((cv-1) & 0b11);
	
	register_mode_write(0b101, mypage);   // set register 6 (RRR=101) as page
	register_mode_write(regdata, data); 
  } 


unsigned char paged_mode_verify(unsigned int cv, unsigned char data)
  {
    unsigned char regdata, mypage;

    // Berechnet mypage und regdata
    mypage = (unsigned char)(((cv-1) >> 2) + 1);   
	regdata = (unsigned char)((cv-1) & 0b11);
	
	register_mode_write(0b101, mypage);   // set register 6 (RRR=101) as page
	return(register_mode_verify(regdata, data)); 
  } 


signed int paged_mode_read_byte(unsigned int cv)
  {
    unsigned char regdata, mypage;

    // Berechnet mypage und regdata
    mypage = (unsigned char)(((cv-1) >> 2) + 1);   
	regdata = (unsigned char)((cv-1) & 0b11);

	register_mode_write(0b101, mypage);   // set register 6 (RRR=101) as page
    return(register_mode_read_byte(regdata)); 

    
  }





//
// Packet Sequence for Command Stations/Programmers using Paged Mode Addressing
// 3 or more Reset Packets
// 5 or more Writes to the Page Register
// 6 or more reset packets (Decoder recovery Time from write to Page Register)
// 3 or more Reset Packets
//[ 5 or more Verifies to a single Data Register 1-4 ] or 5 or more Writes to a single Data Register 1-4
//] ] 6 or more identical Write or Reset packets
//(Decoder Recovery Time from write to Data
//Register) ] ]
//Optional Power Off
//Within a verify or write sequence, the Command Station/Programming may cease sending packets and continue





//===========================================================================================
//
// Lenzinterface (upstream)
//
// Achtung: Lenz kann nur 256 Adressen!
// Liste der untersttzten Lenzbefehle (x=nyi):

// x _ Prog.-Lesen Register 0x22 0x11 REG X-Or
// x _ Prog.-Lesen CV 0x22 0x15 CV X-Or
// x _ Prog.-Lesen Paging 0x22 0x14 CV X-Or
// x _ Prog.-Ergebnis anfordern 0x21 0x10 0x31
// x _ Prog.-Schreiben Register 0x23 0x12 REG DAT X-Or
// x _ Prog.-Schreiben CV 0x23 0x16 CV DAT X-Or
// x _ Prog.-Schreiben Paging 0x23 0x17 CV DAT X-Or

// x _ Prog. on Main Byte ab V3 0xE6 0x30 ADR High ADR Low 0xEC + C CV DAT X-Or
// x _ Prog. on Main Bit ab V3 0xE6 0x30 ADR High ADR Low 0xE8 + C CV DAT X-Or

t_lenz_result lenz_result;                                  // This stores the last programming result
unsigned char lenz_result_adr;
unsigned char lenz_result_data;

unsigned char pcm_lenz_prog[] = {0x63, 0x21, 0x30, 0x00};     // LZ100 Zentrale in Version 3.0
unsigned char *pcm_lenz_ptr;


//-----------------------------------------------------------------------------------------------
/// switch opendcc to progmode
/// does nothing if already in progmode, else puts a power cycle on the track

void enter_progmode(void)
  {
    switch(opendcc_state)
      {
        case RUN_OKAY:
        case RUN_STOP:             // DCC Running, all Engines Emergency Stop
        case RUN_OFF:              // Output disabled (2*Taste, PC)
        case RUN_SHORT:            // Kurzschlu
        case RUN_PAUSE:            // DCC Running, all Engines Speed 0
            while (next_message_count != 0);    // busy wait for current message to terminate
                                                // do not allow organizer to load next command!
	        set_opendcc_state(PROG_OKAY);
            pDCC_Reset.repeat = 20;             // 20 reset packets -> power on cycle	 
	        put_in_queue_prog(dcc_reset_ptr);
            break;

        case PROG_OKAY:            // nothing to change, we are already in prog_mode
            break;
        case PROG_SHORT:           //
        case PROG_OFF:
        case PROG_ERROR:
            while (next_message_count != 0);    // busy wait for current message to terminate


	        set_opendcc_state(PROG_OKAY);
            pDCC_Reset.repeat = 20;             // 20 reset packets -> power on cycle	 
	        put_in_queue_prog(dcc_reset_ptr);
            break;
      }
  }

void lprog_send_prog_result(t_lenz_result ergebnis)
  {
    #if (PARSER == LENZ)
    
    // Messages:
	// 61 11: ready
	// 61 12: short - Kurzschluss
	// 61 13: cant read - Daten nicht gefunden
	// 61 1f: busy
	// 63 10 EE D: EE=adr, D=Daten; nur fr Register oder Pagemode, wenn bei cv diese Antwort, dann kein cv!
	// 63 14 CV D: CV=cv, D=Daten: nur wenn cv gelesen wurde
	switch (ergebnis)
	  {
	    case VOID:
		    pcm_lenz_prog[0] = 0x61;
		    pcm_lenz_prog[1] = 0x13;
		    pc_send_lenz(pcm_lenz_ptr = pcm_lenz_prog); 
		    break; 
	    case READY:
		    pcm_lenz_prog[0] = 0x61;
		    pcm_lenz_prog[1] = 0x11;
		    pc_send_lenz(pcm_lenz_ptr = pcm_lenz_prog); 
		    break; 
	    case BUSY:
		    pcm_lenz_prog[0] = 0x61;
		    pcm_lenz_prog[1] = 0x1f;
		    pc_send_lenz(pcm_lenz_ptr = pcm_lenz_prog); 
		    break; 
	    case REGMODE:
		    pcm_lenz_prog[0] = 0x63;
		    pcm_lenz_prog[1] = 0x10;
		    pcm_lenz_prog[2] = lenz_result_adr;
		    pcm_lenz_prog[3] = lenz_result_data;
		    pc_send_lenz(pcm_lenz_ptr = pcm_lenz_prog); 
		    break; 
	    case CVMODE:
		    pcm_lenz_prog[0] = 0x63;
		    pcm_lenz_prog[1] = 0x14;
		    pcm_lenz_prog[2] = lenz_result_adr;
		    pcm_lenz_prog[3] = lenz_result_data;
		    pc_send_lenz(pcm_lenz_ptr = pcm_lenz_prog); 
		    break; 
	    case SHORT:
		    pcm_lenz_prog[0] = 0x61;
		    pcm_lenz_prog[1] = 0x12;
		    pc_send_lenz(pcm_lenz_ptr = pcm_lenz_prog); 
		    break;  
	
      }
    #endif  // PARSER LENZ
  }


void lprog_send_prog_result_again(void)
  {
    lprog_send_prog_result(lenz_result);
  }


void lprog_read_register(unsigned char regnr)
  {
     unsigned int retval;
     enter_progmode();
	 lenz_result = BUSY;
	 lenz_result_adr = regnr;
	 if ((retval=register_mode_read_byte(regnr-1)) == -1)       // regadr:0..7, regnr: 1..8
       {
	     // failed
		 lenz_result = VOID;
	   }
     else
	   {
	     lenz_result_data = (unsigned char) retval;
		 lenz_result = REGMODE;
	   }
  }


void lprog_write_register(unsigned char regnr, unsigned char data)
  {
    enter_progmode();
	lenz_result = BUSY;
	lenz_result_adr = regnr;
    register_mode_write(regnr - 1, data);
	lenz_result = REGMODE;
  }




// lenz_cv = 1...256, 256 is coded as 0
//
void lprog_read_cv(unsigned char lenz_cv)
  {
     unsigned int i, retval;
     enter_progmode();
	 lenz_result = BUSY;
	 lenz_result_adr = lenz_cv;
     // i = (unsigned int)(lenz_cv-1) + 1;                       // das geht leider nicht, optimiert gnu weg
	 if (lenz_cv == 0) i = 256;
	 else i = lenz_cv;
 
    if (key_go_at_start_pressed)                                 // if GO was pressed - redirect to internal cv
      {
        lenz_result_data = eeprom_read_byte((void *)(eadr_OpenDCC_Version + lenz_cv));
        lenz_result = CVMODE;
        return;
      }

	 if ((retval=direct_mode_read(i)) == -1)                     // cv as is
       {
	     // failed, try again in page mode
         if ((retval=paged_mode_read_byte(i)) == -1)   
		   {
      		 lenz_result = VOID;
		   }
		 else
		   {
		     lenz_result_data = (unsigned char) retval;
		     lenz_result = REGMODE;
	       }
	   }
     else
	   {
	     lenz_result_data = (unsigned char) retval;
		 lenz_result = CVMODE;
	   }
  }


void lprog_write_cv(unsigned char lenz_cv, unsigned char data)
  {
    unsigned int i;
    enter_progmode();
	lenz_result = BUSY;
	lenz_result_adr = lenz_cv;
	 if (lenz_cv == 0) i = 256;
	 else i = lenz_cv;

    if (key_go_at_start_pressed)                                 // if GO was pressed - redirect to internal cv
      {
        eeprom_write_byte((void *)(eadr_OpenDCC_Version + lenz_cv), data);
        lenz_result = CVMODE;
        return;
      }

    direct_mode_write(i, data);
	lenz_result = CVMODE;
  }



// lenz_cv = 1...256, 256 is coded as 0
//
void lprog_read_paged(unsigned char lenz_cv)
  {
     unsigned int i, retval;
     enter_progmode();
	 lenz_result = BUSY;
	 lenz_result_adr = lenz_cv;
	 if (lenz_cv == 0) i = 256;
	 else i = lenz_cv;


     if ((retval=paged_mode_read_byte(i)) == -1)   
		   {
      		 lenz_result = VOID;
		   }
		 else
		   {
		     lenz_result_data = (unsigned char) retval;
		     lenz_result = REGMODE;
	       }

  }

void lprog_write_paged(unsigned char lenz_cv, unsigned char data)
  {
    unsigned int i;
    enter_progmode();
	lenz_result = BUSY;
	lenz_result_adr = lenz_cv;
	 if (lenz_cv == 0) i = 256;
	 else i = lenz_cv;

    paged_mode_write(i, data);
	lenz_result = REGMODE;
  }




#endif

