//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      main.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2006-04-13 V0.1 started
//            2006-10-24 V0.2 check jumpers
//            2007-01-11 V0.3 clean up
//            2008-01-16 V0.4 call to s88 only if not prog_state
//            2008-07-19 V0.5 external stop
//
//-----------------------------------------------------------------
//
// purpose:   lowcost central station for dcc
//            a replacement for Lenz, Intellibox or HSI88
//            
//            See CONFIG.H about more information and how to
//            configure the system
//
// content:   init Atmel AVR ATmega
//            check jumpers
//            init states
//            startup ISR for DCC-OUT
//            idle-loop
//
//-----------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <string.h>

#include "config.h"                // general structures and definitions - make your changes here
#include "status.h"                // led, key, state
#include "rs232.h"                 // interface to pc
#include "dccout.h"                // make dcc
#include "dmxout.h"                // make dmx (light)
#include "organizer.h"             // manage commands
#if (PARSER == LENZ)
  #include "lenz_parser.h"         // talk to pc (same as ibox_parser.h)
  #include "lenz_programmer.h"     // DCC service mode
#endif
#if (PARSER == INTELLIBOX)
  #include "ibox_parser.h"         
  #include "ibox_programmer.h"     // DCC service mode
#endif
#include "s88.h"                   // s88-bus



//-------------------------------------------------------------------------------------------


void init_timer2(void)
  {
    #if (PROCESSOR == ATMEGA644P)
        // Timer/Counter 2 initialization
        // Clock source: System Clock / 32 -> 2us
        TCCR2A = (0<< COM2A1)   // 00 = normal port mode
               | (0<< COM2A0)
               | (0<< COM2B1)   // 00 = normal port mode
               | (0<< COM2B0)
               | (0<< WGM21)    // WGM = 000: normal mode
               | (0<< WGM20);
        TCCR2B = (0<< FOC2A)    // 0 = no forced compare match
               | (0<< FOC2B)    // 0 = no forced compare match
               | (0<< WGM22)    // WGM = 000: normal mode
               | (0<<CS22)      // CS = 000: stopped
               | (1<<CS21)      //      001 = run, 010 = div8, 011=div32, 100=div64, 101=div128. 
               | (1<<CS20);     //      110 = div256, 111 = div1024
        TCNT2=0x00;
    #elif (PROCESSOR == ATMEGA644)
        // Timer/Counter 2 initialization
        // Clock source: System Clock / 32 -> 2us
        // Mode: Normal top=FFh
        // OC2 output: Disconnected
        TCCR2 =  (0<<  FOC2)    // 0 = no forced compare match
               | (0<<  WGM20)   // WGM = 00: normal mode  
               | (0<<  COM21)   // 00 = normal port mode
               | (0<<  COM20)
               | (0<<  WGM21)
               | (0<<  CS22)    // CS = 000: stopped
               | (1<<  CS21)    //      001 = run, 010 = div8, 011=div32, 100=div64, 101=div128. 
               | (1<<  CS20);   //      110 = div256, 111 = div1024
        TCNT2=0x00;
    #elif (PROCESSOR == ATMEGA32)
        // Timer/Counter 2 initialization
        // Clock source: System Clock / 32 -> 2us
        // Mode: Normal top=FFh
        // OC2 output: Disconnected
        TCCR2 =  (0<<  FOC2)    // 0 = no forced compare match
               | (0<<  WGM20)   // WGM = 00: normal mode  
               | (0<<  COM21)   // 00 = normal port mode
               | (0<<  COM20)
               | (0<<  WGM21)
               | (0<<  CS22)    // CS = 000: stopped
               | (1<<  CS21)    //      001 = run, 010 = div8, 011=div32, 100=div64, 101=div128. 
               | (1<<  CS20);   //      110 = div256, 111 = div1024
        TCNT2=0x00;
    #else
      #error: counter2 undefinded
    #endif
  }



void init_main(void)
  {
  // Input/Output Ports initialization
  PORTA = (1<<S88DIN1)       // high = pull up
        | (1<<S88DIN2)       // high = pull up
        | (1<<S88DIN3)       // high = pull up
        | (1<<EXT_STOP)      // high = pull up
        | (1<<LED_GO)        // low = turn on
        | (1<<LED_STOP)      // low = turn on
        | (1<<KEY_STOP)      // high = pull up
        | (1<<KEY_GO);       // high = pull up
  DDRA  = (0<<S88DIN1)       // in
        | (0<<S88DIN2)       // in
        | (0<<S88DIN3)       // in
        | (0<<EXT_STOP)      // in
        | (1<<LED_GO)        // out
        | (1<<LED_STOP)      // out
        | (0<<KEY_STOP)      // in
        | (0<<KEY_GO);       // in
  // Port B
  PORTB = (0<<S88_CLK)       // 0 clk = low
        | (0<<S88_RESET)     // 1 reset = low
        | (1<<ACK_DETECT)    // 2 pullup
        | (1<<S88_LOAD)      // 3 load = high
        | (1<<DMX_OUT)       // 4 high = inactiv
                             // 5 MOSI
        | (0<<MY_CTS)        // 6
        | (1<<MY_RTS);       // 7 high = pullup

  DDRB  = (1<<S88_CLK)       // out
        | (1<<S88_RESET)     // out
        | (0<<ACK_DETECT)    // in
        | (1<<S88_LOAD)      // out
        | (1<<DMX_OUT)       // out
        | (1<<MY_CTS)        // out
        | (0<<MY_RTS);       // in
// Port C
  PORTC = (1<<SHORT_MAIN)    // pull up
        | (1<<SHORT_PROG)    // pull up
        | (1<<LED_CTRL)      // low = turn on
        | (1<<LED_RS232);    // low = turn on

  DDRC  = (0<<SHORT_MAIN)    // in
        | (0<<SHORT_PROG)    // in
        | (1<<LED_CTRL)      // out
        | (1<<LED_RS232);    // out
// Port D
  PORTD = (1<<MY_RXD)        // in, pullup
        | (1<<MY_TXD)        // out
        | (1<<JUMPER1)       // pullup
        | (1<<JUMPER2)       // pullup
        | (0<<NDCC)          // out
        | (1<<DCC)           // out
        | (0<<ENA_PROG)      // out, disabled
        | (0<<ENA_MAIN);     // out, disabled

  DDRD  = (0<<MY_RXD)        // in
        | (1<<MY_TXD)        // out
        | (0<<JUMPER1)       // in
        | (0<<JUMPER2)       // in
        | (1<<NDCC)          // out
        | (1<<DCC)           // out
        | (1<<ENA_PROG)      // out
        | (1<<ENA_MAIN);     // out

    // Timer/Counter 0 initialization
    // Clock source: System Clock
    // Clock value: Timer 0 Stopped
    // Mode: Normal top=FFh
    // OC0 output: Disconnected
    ASSR=0x00;
    TCCR0=0x08;
    TCNT0=0x00;
    OCR0=0xE6;

    // Timer1: done in init_dccout();

    init_timer2();


    // Analog Comparator initialization
    // Analog Comparator: Off
    // Analog Comparator Input Capture by Timer/Counter 1: Off
    ACSR=0x80;
    SFIOR=0x00;   // PUD = 0: no "pull up disabled"

  }

void init_interrupt(void)
  {
  // External Interrupt(s) initialization
  // INT0: Off
  // INT1: Off
  // INT2: Off
  // INT3: Off
  // INT4: Off
  // INT5: Off
  // INT6: Off
  // INT7: Off
  // INT7 Mode: Low level

  // Timer(s)/Counter(s) Interrupt(s) initialization
  
  TIMSK = (0<<OCIE2)       // Output Compare
        | (0<<TOIE2)
        | (0<<TICIE1)      // Input Capture Int
        | (1<<OCIE1A)      // Output Compare A
        | (0<<OCIE1B)      // Output Compare B
        | (0<<TOIE1)       // war 1::::: Timer1/Counter1 Overflow Int Enable
        | (0<<OCIE0)
        | (0<<TOIE0);      // Timer0 Overflow


    sei();              // Global enable interrupts

  }



void dcc_startup_messages (void)
  {
    t_message testmess;
    t_message *testmessptr;
    testmessptr = &testmess;

    // 20 Reset Pakete

    testmessptr = &DCC_Reset;
    set_next_message(testmessptr);
    next_message_count = 20;

    while (next_message_count > 0)      // wait
      {
        if (next_message_count % 2)
          {
            LED_CTRL_ON;
          }
        else
          {
            LED_CTRL_OFF;
          }
      }
    // 10 Idle Pakete

    testmessptr = &DCC_Idle;
    set_next_message(testmessptr);
    next_message_count = 10;

    while (next_message_count > 0)      // wait
      {
        if (next_message_count % 2)
          {
            LED_CTRL_ON;
          }
        else
          {
            LED_CTRL_OFF;
          }
      }
  }



//===============================================================================
//
//  Main Program: There are several branches - depending on jumper settings
//
//  Jumpers:
//  0  0     Run as OpenDCC, either Lenz or Intellibox mode
//
//===============================================================================


// Run as DCC central station, emulating Lenz LI101 or Uhlenbrock Intellibox
// see config.h for compile swtiches

#if (WATCH_S88_SIZE1 == 1) 
  volatile unsigned char my_stored_size;
#endif

void main_opendcc(void)
  {
    init_interrupt();
    init_dccout();                      // timing engine for dcc
    init_rs232(DEFAULT_BAUD);           // 19200 is default for Lenz 3.0
    
    init_tick();                        // 5ms timer tick 
    init_parser();                      // command parser

    init_organizer();                   // engine for command repetition, 
                                        // memory of loco speeds and types

    init_programmer();                  // State Engine des Programmers


    #if (WATCH_S88_SIZE1 == 1) 
       my_stored_size = eeprom_read_byte((void *)eadr_s88_size1);
    #endif

    
    #if (S88_ENABLED == 1)
     init_s88(READ_FROM_EEPROM);  
    #endif

    #if (DMX_ENABLED == 1)
     init_dmxout();
    #endif

    if (eeprom_read_byte((void *)eadr_OpenDCC_Version) != OPENDCC_VERSION)
      {
        // oops, no data loaded or wrong version! 
        error_blink_leds();
        while(1) {};
      }

    if (eeprom_read_byte((void *)eadr_ext_stop_enabled) != 0)
      {
        ext_stop_enabled = 1;           // enable external OC Input for emergency stop
      }

    set_opendcc_state(RUN_OFF);         // start up with power off
    // set_opendcc_state(RUN_OKAY);         // start up with power enabled
    
    // dcc_startup_messages();          // issue defined power up sequence on tracks
    
    while(1)
     {
       run_state();                     // check short and keys
       run_organizer();                 // run command organizer, depending on state,
                                        // it will execute normal track operation
                                        // or programming
       run_programmer();
       run_parser();                    // check commands from pc
       #if (S88_ENABLED == 1)
         if (!is_prog_state()) run_s88(); // s88 has busy loops, we block it when programming
       #endif

       #if (WATCH_S88_SIZE1 == 1) 
          if (eeprom_read_byte((void *)eadr_s88_size1) != my_stored_size)
              set_alarm_led(1);
       #endif


     }
  }

/// run as s88 feedback engine, emulating Littfinski HSI88

void main_hsi88(void)
  {
    init_interrupt();
    init_tick();  
    init_rs232(BAUD_9600);
    s88_size1 = 2;                      // this is default for HSI88
    s88_size2 = 2;
    s88_size3 = 2;
    init_s88(KEEP_SIZES);
    LED_GO_ON;
    while(1)
     {
       HSI88_parser();
     }
  }


// run as stand alone room light controller
// (note: use a different program to emulate a miniDMX)
void main_dmx(void)                    
  {
    LED_GO_ON;
    init_interrupt();
    init_tick();
    #if (DMX_ENABLED == 1)
      init_dmxout();  
    #endif
    init_rs232(BAUD_19200);
    while(1)
     {
        #if (DMX_ENABLED == 1)
        run_dmxout();
        run_dmxkey();
        #endif
     }
    
  }


void main_reserved(void)               // currently this is hardware test mode
  {
    unsigned char echo;


    init_interrupt();
    init_tick();                       // 5ms timer tick 
    init_rs232(BAUD_19200);             // isrs+fifos from/to pc

    test_blink_leds();
    
    while(1)
      {
       if (rx_fifo_ready())
         {
            echo = rx_fifo_read();     
            tx_fifo_write(echo);
         }
      }
  }


#if (DEBUG == 3)
#warning DEBUG activated
void main_simulat(void)
  {
    init_interrupt();
    init_dccout();                      // timing engine for dcc
    init_rs232(DEFAULT_BAUD);           // 19200 is default for Lenz 3.0
    
    init_tick();                        // 5ms timer tick 
    init_parser();                      // command parser

    init_organizer();                   // engine for command repetition, 
                                        // memory of loco speeds and types
    
    init_programmer();                  // State Engine des Programmers
    

    #if (S88_ENABLED == 1)
     init_s88(READ_FROM_EEPROM);  
    #endif    

    //set_opendcc_state(RUN_OKAY);         // eingeschaltet  
    set_opendcc_state(RUN_OFF);         // abgeschaltet 

    void push_to_rx(unsigned char d);

    // Some IB command used for simulation
    // Note: these require start_mode_ibox = 1 (no leading 'x')
    // push_to_rx(0x9C); push_to_rx(0x00);                     // X88PGet(0)
    // push_to_rx(0x85); push_to_rx(0x04); push_to_rx(0x00);   // XlokCfg
    // push_to_rx(0x94); push_to_rx(0x01); push_to_rx(0x00);   // XTrnoutSTS 
    // push_to_rx(0x90); push_to_rx(0x8C); push_to_rx(0xc0);   // XTrnout #140
    // push_to_rx(0x90); push_to_rx(0x7C); push_to_rx(0xc0);   // XTrnout #124
    // push_to_rx(0x80); push_to_rx(0x0a); push_to_rx(0x00); push_to_rx(0x3f); push_to_rx(0x85); // XLOK #10, S50%, func 3+1 
    // push_to_rx(0xDE); push_to_rx(0x03); push_to_rx(0x00); push_to_rx(0x44); push_to_rx(0x00); push_to_rx(0x0A);  // POM 
    // push_to_rx(0xF0); push_to_rx(0x0A); push_to_rx(0x00); // XPTRD: read cv10
    // push_to_rx(0xCE); // XPTreport

    // Note: xb19200
    // push_to_rx(0x78);  push_to_rx(0x62);
    // push_to_rx(0x31);  push_to_rx(0x39);  push_to_rx(0x32);  push_to_rx(0x30); push_to_rx(0x30); push_to_rx(0x0D);
    
    // push_to_rx(0x58); push_to_rx(0x50); push_to_rx(0x54); // XPTRD1 (ascii)
    // push_to_rx(0x52); push_to_rx(0x44); push_to_rx(0x31); push_to_rx(0x0d);
    
    // Xtrnout
    push_to_rx(0x78);  push_to_rx(0x90); push_to_rx(0x0D); push_to_rx(0x40);   // XTrnout on #13, red
    push_to_rx(0x78);  push_to_rx(0x90); push_to_rx(0x0D); push_to_rx(0x00);   // XTrnout off #13, red
    

    // Some Lenz Commands
    // push_to_rx(0x21); push_to_rx(0x24); push_to_rx(0x05);   // Status 
    // push_to_rx(0x22); push_to_rx(0x14); push_to_rx(0x01); push_to_rx(0x37);   // read CV1 



    #if (DMX_ENABLED == 1)
     init_dmxout();
    #endif

    // dcc_startup_messages();          // issue defined power up sequence on tracks
    
    while(1)
     {
       run_state();                     // check short and keys
       run_organizer();                 // run command organizer, depending on state,
                                        // it will execute normal track operation or
                                        // programming commands
       run_programmer();
       run_parser();                    // check commands from pc
       #if (S88_ENABLED == 1)
         run_s88();
       #endif
     }


  }
#endif

//================================================================================================
int main(void)                   
  {
    // note: for debugging set the following IO's
    // PINA: 11-- ----  -> no key strokes
    // PINC: ---- --11  -> no short curcuits
    // PIND: ---- 11--  -> no jumper
   
    init_main();                                  // all io's

    
    #if (DEBUG == 3)
      PINA |= 0xC0;
      PINC |= 0x03;
      PIND |= 0x0C;
      main_simulat();
    #endif

    if (KEY_GO_PRESSED) key_go_at_start_pressed = 1;
    
    if (CHECK_JUMPER1) 
      {
        if (CHECK_JUMPER2) main_reserved();       // if both JUMPER: Test Mode 
        else main_dmx();                          // if JUMPER1: DMX
      }
    else 
      {
        if (CHECK_JUMPER2) main_hsi88();          // if JUMPER2: HSI88
        else main_opendcc();                      // no JUMPER: Lenz/IBOX
      }

    return(0);                                    // dummy
  }
   
//===============================================================================
//
// regular code ends here
//
//=============================================================================== 
//===============================================================================
//
// routines for simulation and test - you may ommit them


// test organizer and loco commands

void debug_level1_testing(void)
  {
    unsigned char i;
    t_message testmess;
    t_message *testmessptr;
    testmessptr = &testmess;

    // fill locobuffer

    for (i=1; i<4; i++)
        {
            enter_speed_f_to_locobuffer(i, i, DCC28);    // adr 10, speed 1
          }

    for (i=1; i<4; i++)
          {
            enter_speed_f_to_locobuffer(2+i, 10+i, DCC28);    // versetzt: adr 12, speed 10
          }

    testmessptr = &testmess;
    for (i=0; i<25; i++)
          {
            testmessptr = search_locobuffer();    // und wieder 25 * abholen
           next_message.size = testmessptr->size;
           next_message.dcc[0] = testmessptr->dcc[0];
           next_message.dcc[1] = testmessptr->dcc[1];
          }


        for (i=1; i<4; i++)
          {
            enter_speed_f_to_locobuffer(4+i, 20+i, DCC28);    // versetzt: adr 12, speed 10
          }


    for (i=0; i<250; i++)
      {
        testmessptr = search_locobuffer();    // und wieder abholen
        next_message.size = testmessptr->size;
        next_message.dcc[0] = testmessptr->dcc[0];
        next_message.dcc[1] = testmessptr->dcc[1];
      }

        // test: build messages

    testmessptr = &testmess;
    build_loko_7a28s(3, 10, testmessptr);   // adr 3, speed 10
    put_in_queue_lp(testmessptr);
    enter_speed_f_to_locobuffer(3, 10, DCC28);

    build_loko_7a28s(3, -10, testmessptr);   // adr 3, speed 10

    build_loko_7a128s(3, 50, testmessptr);   // adr 3, speed 50
    build_loko_7a128s(3, -50, testmessptr);   // adr 3, speed 50

    build_loko_14a28s(3, 10, testmessptr);   // adr 3, speed 10
    build_loko_14a28s(3, -10, testmessptr);   // adr 3, speed 10

    build_loko_14a128s(3, 50, testmessptr);   // adr 3, speed 50
    build_loko_14a128s(3, -50, testmessptr);   // adr 3, speed 50

    build_nmra_basic_accessory(333, 1, 1, testmessptr);  // Weiche 333 auf grn, activate

  }



//----------------------------------------------------------------
// test des Organizers
// Hierzu queues fllen
// Interrupts disable, abholen der DCC-messages manuell

unsigned char tdccsize;
unsigned char tdcc0;
unsigned char tdcc1;
unsigned char tdcc2;

void abruf(void)
  { run_organizer(); 
    if ( next_message_count )
          {
            next_message_count = 0;
            tdcc0 = next_message.dcc[0];
            tdcc1 = next_message.dcc[1];
            tdcc2 = next_message.dcc[2];
            tdccsize = next_message.size;
            
          } 
  }


int main_test_organizer(void)                  
  {
    unsigned char i, j;

    init_main();                   // all io's
    init_dccout();                 // timing engine for dcc
    init_organizer();              // engine for command repetition, 
                                   // memory of loco speeds and types

    // init_rs232(BAUD_19200);
    // init_interrupt();
    // init_tick();
    // init_parser();
    // set_led_rs232(STOPPED);
    
    //-- queue fllen 
    for (j=1; j<5; j++) 
       {
           if (do_loco_speed_f(j, j+4, DCC28))
             {
               // queue1 is full!
               i = 0;  // statement for breakpoint
             }
      }
      abruf(); abruf();
    for (j=1; j<4; j++) 
       {
         
           if (do_loco_speed_f(j, 0x80 + j+1, DCC28))   // die sollten in die hp
             {
               // queue1 is full!
               i = 0;  // statement for breakpoint
             }
         }
    
    // jetzt diese abrufen
 
    for (j=0; j<50; j++) 
       {
          abruf(); 
       }
    
    
    while(1)
     {
       run_organizer();                         // let run command organizer
       
       //for (i=0; i<20; i++)  timer1_simu();      //   this to simulate ISR(TIMER1_OVF_vect);

       run_organizer();                         // let run command organizer
       
       //for (i=0; i<20; i++)  timer1_simu();      //   this to simulate ISR(TIMER1_OVF_vect);

       //for (j=0; j<3; j++) 
       //  {
       //    for (i=0; i<50; i++) timer1_simu();      //   this to simulate ISR(TIMER1_OVF_vect);
       //    if (do_loco_speed_f(j+1, j+16, DCC28))
       //      {
       //        // queue1 is full!
       //        i = 0;  // statement for breakpoint
       //      }
       //  }

       run_organizer();                         // let run command organizer
       
       //for (i=0; i<20; i++)  timer1_simu();      //   this to simulate ISR(TIMER1_OVF_vect);

       run_organizer();                         // let run command organizer
       
       //for (i=0; i<20; i++)  timer1_simu();      //   this to simulate ISR(TIMER1_OVF_vect);

       //for (j=0; j<3; j++) 
       //  {
       //    for (i=0; i<50; i++) timer1_simu();      //   this to simulate ISR(TIMER1_OVF_vect);
       //    if (do_loco_speed_f(j+3, j+18, DCC28))
       //      {
       //        // queue1 is full!
       //        i = 0;  // statement for breakpoint
       //      }
       //  }

     }
    return(0);
  }




