//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      rs232.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2006-04-13 V0.1 started
//            2006-04-28 V0.2 added set_led*** and timeout
//            2006-05-15 V0.3 removed early sei();
//            2006-10-17 V0.4 baudrate types for intellibox mode
//            2006-11-16 V0.5 added push_to_rx for easier simulation
//            2007-01-27 V0.6 changed to 2 stop bits to clear some 
//                            trouble with USB-to-Serial Converter
//            2007-06-09 V0.7 new function tx_all_sent
//                            reset txc on data transmission 
//            2008-04-17 V0.8 fixed a bug in init U2X was cleared
//                            unintentionally; added round to baudrate cast!
//
//-----------------------------------------------------------------
//
// purpose:   lowcost central station for dcc
// content:   routines for RS232
//            see also:
//            http://www.roboternetz.de/wissen/index.php/UART_mit_avr-gcc
//
//
//-----------------------------------------------------------------


#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>         // put var to program memory
#include <avr/eeprom.h>
#include <string.h>

#include "config.h"               // general structures and definitions
#include "status.h"
#include "rs232.h"


//
//=====================================================================
//
// RS232
//
// purpose:   send and receive messages from pc
//
// how:       uart acts with interrupt on fifos.
//            ohter programs access only the fifos.
//            hardware handshake with RTS and CTS.
//            display of status with one LED (see status.c)
//
// uses:      set_led_rs232
//            no_timeout.rs232 (see status.c)
// interface: see rs232.h
//
// 2do:       zur Zeit wird RTS nur als connected Erkennung benutzt,
//            das Senden wird NICHT angehalten.
//
//-----------------------------------------------------------------


// FIFO-Objekte und Puffer fr die Ein- und Ausgabe

#define RxBuffer_Size  64              // mind. 16
unsigned char RxBuffer[RxBuffer_Size];


#define TxBuffer_Size  64
unsigned char TxBuffer[TxBuffer_Size];

unsigned char rx_read_ptr = 0;        // point to next read
unsigned char rx_write_ptr = 0;       // point to next write
unsigned char rx_fill = 0;
unsigned char tx_read_ptr = 0;
unsigned char tx_write_ptr = 0;
unsigned char tx_fill = 0;

const unsigned long baudrate[] PROGMEM =    // ordered like in Lenz Interface!
    {9600L,   //  = 0,
     19200L,  //  = 1,
     38400L,  //  = 2,
     57600L,  //  = 3,
     115200L, //  = 4,
     2400L,   //  = 5,  // used for Intellibox 
	};

t_baud actual_baudrate;                      // index to field above

volatile unsigned char rs232_break_detected = 0;      // flag, that a break was detected
                                                    // volatile, weil aus ISR bearbeitet wird.

//-------------------------------------------------------------------

void init_rs232(t_baud new_baud)
  {
    uint16_t ubrr;
    uint8_t sreg = SREG;

    cli();

    UCSRB = 0;                  // stop everything

    actual_baudrate = new_baud;

    switch(new_baud)				// calculation with gain 100
	  {                                     // added 50 before cast to get rounding
	    default:
		case BAUD_2400:
		    // ubrr = (uint16_t) ((uint32_t) F_CPU/(16*2400L) - 1);
                    ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*24L) - 100L + 50L) / 100);
            UBRRH = (uint8_t) (ubrr>>8);
            UBRRL = (uint8_t) (ubrr);
			UCSRA = (1 << RXC) | (1 << TXC);
		    break;
		case BAUD_4800:
		    // ubrr = (uint16_t) ((uint32_t) F_CPU/(16*4800L) - 1);
                    ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*48L) - 100L + 50L) / 100);
            UBRRH = (uint8_t) (ubrr>>8);
            UBRRL = (uint8_t) (ubrr);
			UCSRA = (1 << RXC) | (1 << TXC);
		    break;
		case BAUD_9600:
		    //   UBRRL = 103; // 19200bps @ 16.00MHz
		    // ubrr = (uint16_t) ((uint32_t) F_CPU/(16*9600L) - 1);
                    ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*96L) - 100L + 50L) / 100);
            UBRRH = (uint8_t) (ubrr>>8);
            UBRRL = (uint8_t) (ubrr);
			UCSRA = (1 << RXC) | (1 << TXC);
		    break;
		case BAUD_19200:
		    //   UBRRL = 51; // 19200bps @ 16.00MHz
		    // ubrr = (uint16_t) ((uint32_t) F_CPU/(16*19200L) - 1);
                    ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*192L) - 100L + 50L) / 100);
            UBRRH = (uint8_t) (ubrr>>8);
            UBRRL = (uint8_t) (ubrr);
			UCSRA = (1 << RXC) | (1 << TXC);
		    break;
		case BAUD_38400:
		    //   UBRRL = 25; // 38400bps @ 16.00MHz
		 // ubrr = (uint16_t) ((uint32_t) F_CPU/(16*38400L) - 1);
                    ubrr = (uint16_t) ((uint32_t) (F_CPU/(16*384L) - 100L + 50L) / 100);
            UBRRH = (uint8_t) (ubrr>>8);
            UBRRL = (uint8_t) (ubrr);
			UCSRA = (1 << RXC) | (1 << TXC);
		    break;
		case BAUD_57600:
		    // ubrr = (uint16_t) ((uint32_t) F_CPU/(8*57600L) - 1);
                    ubrr = (uint16_t) ((uint32_t) (F_CPU/(8*576L) - 100L + 50L) / 100);
            UBRRH = (uint8_t) (ubrr>>8);
            UBRRL = (uint8_t) (ubrr);
			UCSRA = (1 << RXC) | (1 << TXC) | (1 << U2X);  // High Speed Mode, nur div 8
		    break;
		case BAUD_115200:
		    // ubrr = (uint16_t) ((uint32_t) F_CPU/(8*115200L) - 1);
                    ubrr = (uint16_t) ((uint32_t) (F_CPU/(8*1152L) - 100L + 50L) / 100);
            UBRRH = (uint8_t) (ubrr>>8);
            UBRRL = (uint8_t) (ubrr);
			UCSRA = (1 << RXC) | (1 << TXC) | (1 << U2X);  // High Speed Mode
		    break;
	  }
   
    // FIFOs fr Ein- und Ausgabe initialisieren
    rx_read_ptr = 0;      
    rx_write_ptr = 0;
    rx_fill = 0;      
    tx_read_ptr = 0;
    tx_write_ptr = 0;
    tx_fill = 0;

	
	// UART Receiver und Transmitter anschalten, Receive-Interrupt aktivieren
    // Data mode 8N1, asynchron
    
    UCSRB = (1 << RXEN) | (1 << TXEN) | (1 << RXCIE);
    UCSRC = (1 << URSEL)        // must be one
          | (0 << UMSEL)        // 0 = asyn mode
          | (0 << UPM1)         // 0 = parity disabled
          | (0 << UPM0)         // 0 = parity disabled
          | (1 << USBS)         // 1 = tx with 2 stop bits
          | (1 << UCSZ1)        // 00 = 8 or 9 bits
          | (1 << UCSZ0);

    
    #ifdef RS232_EXTRA_STOP
        UCSRB |= (1<<TXB8);     // set 9th bit to 1
        UCSRB |= (1<<UCSZ2);    // set frame length to 9bit
    #endif

    // Flush Receive-Buffer
    do
      {
        UDR;
      }
    while (UCSRA & (1 << RXC));

    // Rcksetzen von Receive und Transmit Complete-Flags
    // UCSRA = (1 << RXC) | (1 << TXC);
    UCSRA |= (1 << RXC);
    UCSRA |= (1 << TXC);    

    UDR;

    rs232_break_detected = 0;
    
    SREG = sreg;

    set_led_rs232(DISCONNECTED);


}


//---------------------------------------------------------------------------
// Empfangene Zeichen werden in die Eingabgs-FIFO gespeichert und warten dort
// Wenn bis auf einen Rest von 10 gefllt ist: CTS senden.
//
ISR(SIG_UART_RECV)
{
    if (UCSRA & (1<< FE))
	  { // Frame Error
        rs232_break_detected = 1;       // always on error
		if (UDR == 0)                   // this is a break sent!
	      {
		     rs232_break_detected = 1;  // set flag for parser and discard
		  }
	  }
    else
      {
        if (UCSRA & (1<< DOR))
          { // DATA Overrun -> Fatal
            // !!! 
          }

       	RxBuffer[rx_write_ptr] = UDR;

        rx_write_ptr++;
        if (rx_write_ptr == RxBuffer_Size) rx_write_ptr=0;
        rx_fill++;
        if (rx_fill > (RxBuffer_Size - 10))
          {
            // we are full, stop remote Tx -> set CTS off !!!!
    		// set_led_rs232(STOPPED);
            // removed here done with polling of CTS in status.c
    		MY_CTS_HIGH;
          }
      }
}

//----------------------------------------------------------------------------
// Ein Zeichen aus der Ausgabe-FIFO lesen und ausgeben
// Ist das Zeichen fertig ausgegeben, wird ein neuer SIG_UART_DATA-IRQ getriggert
// Ist das FIFO leer, deaktiviert die ISR ihren eigenen IRQ.

// Bei 9 Bit knnte noch ein Stopbit erzeugt werden: UCSRB |= (1<<TXB8);



#if (DEBUG == 3)
   unsigned  char tx_char;
#endif

ISR(SIG_UART_DATA)
  {

    if (tx_read_ptr != tx_write_ptr)
      {
        UCSRA |= (1 << TXC);              // writing a one clears any existing tx complete flag
        UDR = TxBuffer[tx_read_ptr];
        #if (DEBUG == 3)
            tx_char = TxBuffer[tx_read_ptr];
        #endif
        tx_read_ptr++;
        if (tx_read_ptr == TxBuffer_Size) tx_read_ptr=0;
        tx_fill--;
      }
    else
        UCSRB &= ~(1 << UDRIE);           // disable further TxINT
  }



//=============================================================================
// Upstream Interface
//-----------------------------------------------------------------------------
// TX:
bool tx_fifo_ready (void)
{
    if (tx_fill < (TxBuffer_Size-16))     // keep space for one complete message (16)
      {
        return(1);                        // true if enough room
      }
    else
      {
        return(0);
      }
}


// ret 1 if full
bool tx_fifo_write (const unsigned char c)
{
    TxBuffer[tx_write_ptr] = c;

    tx_write_ptr++;
    if (tx_write_ptr == TxBuffer_Size) tx_write_ptr=0;

    unsigned char mysreg = SREG;
    cli();
    tx_fill++;
    SREG = mysreg;

    UCSRB |= (1 << UDRIE);   // enable TxINT

    if (tx_fill < (TxBuffer_Size-16))
      {
        return(1);
      }

    return(0);
}

// ret 1 if all is sent
bool tx_all_sent (void)
{
    if (tx_fill == 0)
      {
        if (!(UCSRA & (1 << UDRE)))  return(0);    // UDR not empty
        if (!(UCSRA & (1 << TXC)))  return(0);    // TX Completed not set
        return(1);                        
      }
    else
      {
        return(0);
      }
}

void uart_puts (const char *s)
{
    do
      {
        tx_fifo_write (*s);
      }
    while (*s++);
}


//------------------------------------------------------------------------------
// RX:
bool rx_fifo_ready (void)
{
    if (rx_read_ptr != rx_write_ptr)
      {
        return(1);     // there is something
      }
    else
      {
        if (no_timeout.rs232)
          {
            // kein timeout -> nichts machen
          }
        else
          {
            if (MY_RTS_DETECT) set_led_rs232(CONNECTED);
            else set_led_rs232(DISCONNECTED);
          } 
        return(0);  
      }
}

//-------------------------------------------------------------------
// rx_fifo_read gets one char from the input fifo
//
// there is no check whether a char is ready, this must be
// done before calling with a call to rx_fifo_ready();

unsigned char rx_fifo_read (void)
  {
    unsigned char retval;

    retval = RxBuffer[rx_read_ptr];
    rx_read_ptr++;
    if (rx_read_ptr == RxBuffer_Size) rx_read_ptr=0;

    unsigned char mysreg = SREG;
    cli();
    rx_fill--;
    SREG = mysreg;

    if (rx_fill < (RxBuffer_Size - 14))
      {
        if (MY_CTS_STATE)
          {
            // CTS is high, but we are no longer full
            // enable remote Tx -> reset CTS 
            set_led_rs232(RUNNING);
            MY_CTS_LOW;
          }
      }
    no_timeout.rs232 = 1000000L / TICK_PERIOD;      // 1 sec
    return(retval);
  }


//-----------------------------------------------------------------------------------
// testroutinen zu Parsertest
// Ersatzroutine, um ein Zeichen in den Empfangsbuffer abzulegen
// Sendepuffer kann im Simulator mit einen Databreakpoint auf UDR berwacht werden.
void push_to_rx(unsigned char d)
  {
	RxBuffer[rx_write_ptr] = d;

    rx_write_ptr++;
    if (rx_write_ptr == RxBuffer_Size) rx_write_ptr=0;
    rx_fill++;
    if (rx_fill > (RxBuffer_Size-5))
      {
        // we are full, stop remote Tx -> set CTS off !!!!
		set_led_rs232(STOPPED);
		MY_CTS_HIGH;
        // 
      }
  }


