//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006, 2007, 2008 Kufer
//
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      s88.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2006-04-28 V0.01 started
//            2006-05-24 V0.02 added Lenz-If and task scheduler
//            2006-05-25 V0.03 added HSI88 Mode;
//            2006-05-28 V0.04 swapped high-low Byte in HSI88 answer 
//            2006-11-01 V0.05 added code for Intellibox *_ib_*
//            2006-11-01 V0.06 read s88_size from EEPROM
//            2006-12-04 V0.07 call s88_init with typedef
//            2007-02-04 V0.08 make_s88_ib_report (bitorder invertiert)
//            2007-03-20 V0.09 added turnout feedback and s88_mode
//            2007-06-06 V0.10 turnout feedback - range shifted down
//            2008-01-11 V0.11 s88 clk pulse with subroutine, length
//                             defined by a CV.
//            2008-01-16 V0.12 s88 reading split up into bit loops
//                             to avoid too long blocking
//            2008-02-02 V0.13 feedback_size added 
//            2008-07-26 V0.14 s88-Timing based on counter2 - no more polling 
//                             Reports for xpressnet are restricted
//                             to real hardware detectors - no turnout feedback.
//        
//
//-----------------------------------------------------------------
//
// purpose:   lowcost central station for dcc
// content:   s88 interface
//            
// issues:    S88 read in is busy waiting.           
//
// tests:     HSI88-Mode successfully tested
//            S88 Mode with Intellibox tested
//
// 2do:       Timing fr Rckmelder optimieren (Interrupt!)
//            S88 Mode with LENZ not yet tested
//            
// Note:      S88-Timing is defined with S88_CLOCK_PULSE
//
//-----------------------------------------------------------------

#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>               // using sscanf and sprintf increases prog. by 4k!
#include <inttypes.h>
#include <avr/pgmspace.h>        // put var to program memory
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include <string.h>

#include "hardware.h"             
#include "config.h"              // general structures and definitions
#include "rs232.h"               // tx ready
#include "organizer.h"           // recall_turnout

#if (PARSER == LENZ)
  #include "lenz_parser.h"         // tell pc about status changes       
#endif
#if (PARSER == INTELLIBOX)
  #include "ibox_parser.h"         
#endif

#if (TURNOUT_FEEDBACK_ENABLED == 1)
#include "dccout.h"              // feedback variables
#endif

#include "s88.h"     


#define HSI88_ENABLED     1         // if 1: add code to run as HSI88

#define SPECIAL_STDIO               // if defined: the memory consuming functions from stdio
                                    // are replaced by small special functions tailored for
                                    // HSI88;

//-----------------------------------------------------------------------
// Variables and Defines for S88

#define S88_TIMER_TICK   2              // 2us -> this is one tick on Timer2 - 256us maximum

unsigned char s88_mode;                 // Bit 0: 0 = hw read disabled
                                        //        1 = hw read enabled
                                        // Bit 1: 0 = feedback disabled
                                        //        1 = feedback enabled

unsigned char xpressnet_feedback_mode;  // defines the address mapping of feedbacks in Xpressnet
                                        // read from CV29

#define S88_HW_READ     0
#define S88_FEEDBACK    1

unsigned char s88_pulsduration;         // every s88_clk_pulse is constructed of 10us segements
                                        // a value 2 means: 20us high 20us low
                                        // max. 20 = 200us

unsigned char s88_size1, s88_size2, s88_size3;  // sizes of hw_buffer

unsigned char s88_sizemax;              // so viele shifts ntig (=max(size1,size2,size3)
unsigned char s88_sizesum;              // all HW-Melder (=sum(size1,size2,size3)
unsigned char s88_total;                // max. 2048 bits macht 256 bytes
                                        // this value is calculated at power up

unsigned char feedback_s88_offset;      // given in bytes
unsigned char feedback_s88_size;
unsigned char feedback_s88_type;        // howto handle feedback bits
                                        // 0: positive ack feedback (set s88, if feedback was received)
                                        // 1: error feedback (set s88, if no feedback was received)
                                        // 2: coil position



#ifndef SIZE_S88_MAX                       // should be defined in config.h
  #warning local definition of SIZE_S88_MAX
  #define  SIZE_S88_MAX  128               // das Lenzprotokoll lt nur 1024 Rckmelder zu
#endif

unsigned char s88_data[SIZE_S88_MAX];       // Hier sind die aktuellen Zustnde gespeichert
                                            // s88_data[0..feedback_s88_offset-1]: echte Hardwaremelder
                                            // s88_data[feedback_s88_offset..max-1]: Weichen Feedback

unsigned char s88_change[SIZE_S88_MAX];     // Hier ist gespeichert, ob es seit dem letzten
                                            // Lesen einen Change gab.
                                            // Achtung: Lenz macht den Change nibbleweise

unsigned char s88_event;                    // Global: da war irgendeine nderung (see s88.h)


#if (SIZE_S88_MAX > 255)
# warning -> Error, SIZE_S88_MAX to large
#endif

#if (SIZE_S88_1+SIZE_S88_2+SIZE_S88_3 > SIZE_S88_MAX)
# warning -> Error, S88 to large
#endif

// Variables for s88 Task
enum s88_states
  {                              // actual state
     IDLE,
     DO_S88_LEAD_IN,
     DO_READ_BYTE,
     DO_LAST_CLK_PULSE,
     DO_CHECK,
     DO_MESSAGE,
     DO_MESSAGE2
  } s88_state;                  

unsigned char s88_current_i;    // dieses Byte wird gerade gelesen
unsigned char s88_current_b;    // dieses Bit wird gerade gelesen
signed char s88_TCNT;           // Zeitpunkt der nchsten Hardwareausgabe

//---------------------------------------------------------------------------
//
// read_s88
// Action:      liest die Rckmelder nach s88_data[] ein. Es erfolgt noch
//              keine Auswertung.
//              s88_change (complete array) is cleared.
//              s88_event is cleared.
// Parameter:   none
//              done with software delays
//
//
//-----------------------------------------------------------------------
// timing for read
//
// Load  _______/---------------------\______________________________________
//
// Clk   ___________/---\_____________________/---\___/---\___/---\___/---\__
//
// Reset ____________________/---\___________________________________________
//
// Data  ___________/x1.x1.x1.x1.x1.x1.x1.x1../\x2.x2./\x3.x3./\x3.x3./\x4.x4.
//
// Durations:
// with 30m of cable: t(pd)=200ns;
// with 44pF/m and Rout=250Ohm: t(rc)=11ns/m;
// Note: 20us are 320 cycles with a 16 MHz CPU
//
unsigned char s88_clock_phase;

void s88_clk_pulse(void)
  {
    unsigned char i;
    
    S88_CLK_HIGH;
    for (i=0; i<s88_pulsduration; i++) _mydelay_us(10); 
    S88_CLK_LOW; 
    for (i=0; i<s88_pulsduration; i++) _mydelay_us(10); 
  }  

// call until return is TRUE
unsigned char s88_clk_pulse_phased(void)
  {
    unsigned char retval = 0;

    switch(s88_clock_phase)
      {
        default:
        case 0: S88_CLK_HIGH;
                s88_TCNT = TCNT2 + (unsigned char)(s88_pulsduration * 10 / S88_TIMER_TICK);
                s88_clock_phase++;
                break;
        case 1: if ((signed char)(TCNT2 - s88_TCNT) > 0)
                  {
                    S88_CLK_LOW;
                    s88_TCNT = TCNT2 + (unsigned char)(s88_pulsduration * 10 / S88_TIMER_TICK);
                    s88_clock_phase++;
                  }
                break;
        case 2: if ((signed char)(TCNT2 - s88_TCNT) > 0)
                  {
                    s88_clock_phase = 0;
                    retval = 1;                 // completed clock cycle: ^^^^^^.......
                  }
                break;
      }
    return(retval);        
  }  


// old: #define S88_CLK_PULSE   S88_CLK_HIGH; _mydelay_us(30); S88_CLK_LOW; _mydelay_us(20); 

//---------------------------------------------------------------------------------
/// make_s88_reset
/// Action:     macht einen Reset am S88 Bus.
///
/// Note:       Code ist nur kurz busy waiting (je fr ein Byte, 400us).
///
void make_s88_reset(void)
  {
    if (s88_mode & (1<<S88_HW_READ))
      {
        S88_RESET_LOW; _mydelay_us(10);
        S88_LOAD_HIGH; _mydelay_us(100);
        s88_clk_pulse();   // old: S88_CLK_PULSE;                               // load info to shifters
        S88_LOAD_LOW;
        S88_RESET_HIGH; _mydelay_us(100);            // reactivate latches
        S88_RESET_LOW;  _mydelay_us(20); 
      }
  }

//---------------------------------------------------------------------------------
/// make_s88_reset_phased
/// Action:     macht einen Reset am S88 Bus. Diese Routine mu hierzu so lange
///             aufgerufen werden, bis Sie true zurckliefert.
///
/// Return:     true if reset completed
///
unsigned char s88_reset_phase;


unsigned char make_s88_reset_phased(void)
  {
    unsigned char retval = 0;

    if (s88_mode & (1<<S88_HW_READ))
      {
        switch(s88_reset_phase)
          {
            default:
            case 0: S88_RESET_LOW; 
                    s88_TCNT = TCNT2 + (10 / S88_TIMER_TICK);
                    s88_reset_phase++;
                    break;
            case 1: if ((signed char)(TCNT2 - s88_TCNT) > 0)
                      {
                        S88_LOAD_HIGH;                              
                        s88_TCNT = TCNT2 + (100 / S88_TIMER_TICK);
                        s88_reset_phase++;
                      }
                    break;
            case 2: if ((signed char)(TCNT2 - s88_TCNT) > 0)
                        s88_reset_phase++;
                        s88_clock_phase = 0;
                    break;
            case 3: if (s88_clk_pulse_phased())                  // load info to shifters
                            s88_reset_phase++;
                    break;
            case 4: S88_LOAD_LOW;
                    S88_RESET_HIGH;                              // reactivate latches
                    s88_TCNT = TCNT2 + (100 / S88_TIMER_TICK);          
                    s88_reset_phase++;
                    break;
            case 5: if ((signed char)(TCNT2 - s88_TCNT) > 0)
                      {
                        S88_RESET_LOW;
                        s88_TCNT = TCNT2 + (20 / S88_TIMER_TICK);
                        s88_reset_phase++;
                      }
                    break;
            case 6: if ((signed char)(TCNT2 - s88_TCNT) > 0)
                      {
                        s88_reset_phase=0;
                        retval = 1;
                      }
                    break;
          }
      }
    else
        retval = 1;
    return(retval); 
  }

   
void sense_off_s88(void)
  {
    unsigned char i;
    
    for (i=0; i<s88_total; i++) s88_data[i]=0;   // komplett lschen
    for (i=0; i<s88_total; i++) s88_change[i]=0;
    s88_event = 0;
  } 
 

void read_s88(void)  // This code is busy waiting!
  {
    unsigned char i;
    unsigned char b;                    // bit shifter
    
    sense_off_s88();                     // vorher komplett lschen
    
    make_s88_reset();                   // am S88 die Rckmelder laden

    if (s88_mode & (1<<S88_HW_READ))
      {
        for (i=0; i<s88_sizemax; i++)       // Schleife ber maximale Module
          {
            for(b=1; b!=0; b=b<<1)          // loop over bits, use no counter but shifter!      
              {
                if (i<s88_size1)
                  {
                    if (READ_S88_D1) s88_data[i] |= b;
                  }
                if (i<s88_size2)
                  {
                    if (READ_S88_D2) s88_data[i+s88_size1] |= b;
                  }
                if (i<s88_size3)
                  {
                    if (READ_S88_D3) s88_data[i+s88_size1+s88_size2] |= b;
                  }

                s88_clk_pulse();   // old: S88_CLK_PULSE;
              }
          }
      }
  }




//---------------------------------------------------------------------------------
/// check_s88_bit
/// Action:     liest ein Bit am S88 Bus ein und vergleicht dabei auf nderungen
///             Die nderungen werden in s88_change[] aufgesammelt;
///             Alte Changeflags werden nicht bearbeitet.
///
/// Return:     True, if any new change was detected (on hardware)
///
/// Parameter:  i: read this byte
///             b: bitmask (binary)
//
unsigned char check_s88_bit(unsigned char i, unsigned char b)
  {
    unsigned char any_change = 0;

  
    if (s88_mode & (1<<S88_HW_READ))
      {
        if (i<s88_size1)
          {
            if (READ_S88_D1)
              {
                 if ((s88_data[i] & b) == 0) 
                   {
                      // 1 gelesen, alter Zustand 0
                      s88_data[i] |= b;
                      s88_change[i] |= b;
                      any_change |= b;
                   }
              }
            else
              {
                 if ((s88_data[i] & b) != 0) 
                   {
                      // 0 gelesen, alter Zustand 1
                      s88_data[i] &= ~b;
                      s88_change[i] |= b;
                      any_change |= b;
                   }
              }
          }
        if (i<s88_size2)
          {
            if (READ_S88_D2)
              {
                if ((s88_data[i+s88_size1] & b) == 0) 
                  {
                    // 1 gelesen, alter Zustand 0
                    s88_data[i+s88_size1] |= b;
                    s88_change[i+s88_size1] |= b;
                    any_change |= b;
                  }
              }
            else
              {
                if ((s88_data[i+s88_size1] & b) != 0) 
                  {
                    // 0 gelesen, alter Zustand 1
                    s88_data[i+s88_size1] &= ~b;
                    s88_change[i+s88_size1] |= b;
                    any_change |= b;
                  }
              }
          }
        if (i<s88_size3)
          {
            if (READ_S88_D3)
              {
                if ((s88_data[i+s88_size1+s88_size2] & b) == 0) 
                  {
                    // 1 gelesen, alter Zustand 0
                    s88_data[i+s88_size1+s88_size2] |= b;
                    s88_change[i+s88_size1+s88_size2] |= b;
                    any_change |= b;
                  }
              }
            else
              {
                if ((s88_data[i+s88_size1+s88_size2] & b) != 0) 
                  {
                    // 0 gelesen, alter Zustand 1
                    s88_data[i+s88_size1+s88_size2] &= ~b;
                    s88_change[i+s88_size1+s88_size2] |= b;
                    any_change |= b;
                  }
              }
          }

        // removed here: s88_clk_pulse();   // old: S88_CLK_PULSE; 
      } 
    return(any_change);
  }



//---------------------------------------------------------------------------------
/// check_s88_byte
/// Action:     liest ein Byte am S88 Bus ein und vergleicht dabei Bit fr Bit auf nderungen
///             Die nderungen werden in s88_change[] aufgesammelt;
///             Alte Changeflags werden nicht bearbeitet.
///
/// Note:       Code ist busy waiting und blockiert OpenDCC (kurzfristig).
///
/// Return:     True, if any new change was detected (on hardware)
//
unsigned char check_s88_byte(unsigned char i)
  {
    unsigned char b;                    // bit shifter
    unsigned char any_change = 0;

  
    if (s88_mode & (1<<S88_HW_READ))
      {
        for(b=1; b!=0; b=b<<1)          // Schleife ber bits, hier kein Zhler, shifter!      
          {
            any_change |= check_s88_bit(i,b);
            s88_clk_pulse();
          }
      } 
    return(any_change);
  }


//------------------------------------------------------------------------------------
// Turnout FEEDBACK
//------------------------------------------------------------------------------------

unsigned char is_turnout_feedbacked(unsigned int trnt_addr)
  {
    #if (TURNOUT_FEEDBACK_ENABLED == 1)
        unsigned char index;
        if (s88_mode & (1<<S88_FEEDBACK))
          {
            index = trnt_addr / 8;
            if (index < feedback_s88_size) return(1);
          }
        return(0);
    #else
        return(0);
    #endif
  }

#if (TURNOUT_FEEDBACK_ENABLED == 1)
// enter_feedback_to_s88 saves a feedback from accessory to s88-buffer.
// 
// parameters:
//    coil_addr   = turnout_addr * 2 + coil                 
//                  turnout range is from 4 to 2047 (decoder 0 is broadcast)
//    state:        if 0: no ack was received
//                  if 1: ack was received
// effects:
//    s88_event is updated


void enter_feedback_to_s88(unsigned int coil_addr, unsigned char state)
  {
    unsigned int index;
    unsigned char mask;
    unsigned char coil;

    coil = coil_addr & 0b01;
    index = (coil_addr) / 2;                        // now turnout_addr
    index -= 4;                                     // shift range down to 0..2046
    mask = (unsigned char) (index) & 0b0111;
    mask = 1 << mask;                               // bit mask

    index = index / 8;                              // byte addr
    index += feedback_s88_offset;


    switch(feedback_s88_type)                       // determine bit type
          {
            case 0:                 // positive ack
                break;
            case 1:                 // error
                state = !state;
                break;
            case 2:                 // position
                if (coil == 0) state = !state;
                break;
            default:
                break;
          }

    if (index < SIZE_S88_MAX)
      {  
        if (state)
          {
            if (!(s88_data[index] & mask))
              {
                s88_data[index] |= mask;            // save new state
                s88_change[index] |= mask;          // mark change
                s88_event |= (1<<S88_TRNT_CHANGE);
              }
          }
        else
          {
            if (s88_data[index] & mask)
              {
                s88_data[index] &= ~mask;           // save new state
                s88_change[index] |= mask;
                s88_event |= (1<<S88_TRNT_CHANGE);
              }
          }
      }
  }

#endif

#if (HSI88_ENABLED == 1)
//---------------------------------------------------------------------------------
/// check_s88
/// Action:     liest den S88 Bus ein und vergleicht dabei Bit fr Bit auf nderungen
///             Die nderungen werden in s88_change[] aufgesammelt;
///             Alte Changeflags werden nicht bearbeitet.
///
/// Note:       Code ist busy waiting und blockiert OpenDCC. Benutzung nicht empfohlen,
///             besser ber die Task einlesen.
///
/// Return:     True, if any new change was detected
//
unsigned char check_s88(void)
  {
    unsigned char i;
    unsigned char any_change = 0;

    
    make_s88_reset();                       // am S88 die Rckmelder laden

    for (i=0; i<s88_sizemax; i++)           // Schleife ber maximale Module
      {
        any_change |= check_s88_byte(i);    // ein Byte einlesen
      }
    return(any_change);
  }
#endif


//================================================================================
//
// Upstream Interface
//
//--------------------------------------------------------------------------------



//=====================================================================================
// Interfaceroutinen fr das Xpressnet Interface (Lenz)
//=====================================================================================
//
#if (PARSER == LENZ  ||  XPRESSNET_ENABLED == 1)
//
// Format: Headerbyte Adr 1 Dat1 Adr 2 Dat 2 ....... X-Or-Byte
// Hex :   0x40 + N [Adr Dat] X-Or-Byte
// In einem Broadcast wird mindestens ein Adresszustand, maximal 7 Zustnde bertragen.
// (Je Adresse ein Datenbyte, d.h. max 15 Byte (ohne header und xor)
// ADR_x und DAT_x haben das Format wie unter bei "Schaltinformationen" beschrieben.
//
// Parameter: index: Bitadresse: range 0..1023
//            *message: dort wird die info gem Format eingetragen (ein Byte-Prchen)

void create_xpressnet_feedback(unsigned int index, unsigned char *message)
  {
    register unsigned char nibble, addr;
      
    addr = index >> 3;   // range 0...127

    switch(xpressnet_feedback_mode)
      {
        case 0:                             // lenz feedback is shifted up by 512!    
            *message++ = addr | 0x40;       // or 0x40 -> shift 512 up
            break;
        default:                            // restrict to baseline
            *message++ = addr;
            break;
      }

    // build ITTNZZZZ -> real hardware feedback (only)
    // I = not ready
    // TT = Type: 00=Schaltempf. 01=Schaltempf. mit RM, 10: Rckmelder, 11 reserved -> hier 10
    // N: 0=lower Nibble, 1=upper
    // Bei Rckmeldern: Direkt die 4 Bits des Nibbles

    nibble = ((unsigned char) index & 0x04 );
    
    if (nibble == 0)
      {  // lower nibble
        *message = 0b01000000 | (s88_data[addr] & 0b00001111);
      }
    else
      {
        *message = 0b01010000 | ((s88_data[addr] & 0b11110000) >> 4);
      }
  }

unsigned char recall_feedback(unsigned int addr)  // no size checking is performed
  {
    unsigned int index;  unsigned char mask;

    index = addr / 8;
    mask = (unsigned char) (addr) & 0x7;
    mask = 1 << mask;

    if ((s88_data[index + feedback_s88_offset] & mask) == 0) return(0);
    else return(1);
  }


void create_xpressnet_schaltinfo(unsigned int trnt_addr, unsigned char *message)
  {
    unsigned char nibble;
    // Flle:
    // a) keine Rckmeldung: die Weiche wird eingelaufen gemeldet
    // b) Rckmeldung: 
    // b1) Wenn feedback == recall -> eingelaufen
    // b2) Wenn feedback != recall -> im Umlauf
   
    *message++ = trnt_addr >> 2;   // range 0...255
    nibble = (unsigned char)trnt_addr & 0x2;

    if (is_turnout_feedbacked(trnt_addr))
      {
        // ITTN ZZZZ = ittn xxxx; tt = 01 - mit feedback

        if (nibble == 0)  *message = 0b00100000;
        else              *message = 0b00110000;

        if (recall_turnout(trnt_addr & 0xfffe))
          {
            if (!recall_feedback(trnt_addr & 0xfffe)) *message |= 0x80;   // I-Bit setzen	
            *message |= 0b0010;   // Weiche steht auf 1
          }
        else
          {
            if (recall_feedback(trnt_addr & 0xfffe)) *message |= 0x80;   // I-Bit setzen		
            *message |= 0b0001;   // Weiche steht auf 0
          }
        if (recall_turnout(trnt_addr | 0x0001))
          {
            if (!recall_feedback(trnt_addr | 0x0001)) *message |= 0x80;   // I-Bit setzen
            *message |= 0b1000;   // Weiche steht auf 1
          }
        else
          {
            if (recall_feedback(trnt_addr | 0x0001)) *message |= 0x80;   // I-Bit setzen
            *message |= 0b0100;   // Weiche steht auf 0
          }       
      }
    else
      {
        // ITTN ZZZZ = 000n xxxx
        if (nibble == 0)  *message = 0b00000000;
        else              *message = 0b00010000;

        if (recall_turnout(trnt_addr & 0xfffe))
          {
            *message |= 0b0010;   // Weiche steht auf 1
          }
        else
          {
            *message |= 0b0001;   // Weiche steht auf 0
          }
        if (recall_turnout(trnt_addr | 0x0001))
          {
            *message |= 0b1000;   // Weiche steht auf 1
          }
        else
          {
            *message |= 0b0100;   // Weiche steht auf 0
          }
      }
  }

#endif // (PARSER == LENZ  ||  XPRESSNET_ENABLED == 1)

#if (PARSER == LENZ)

#define S88_REPORT_SIZE   7        // laut Lenz max. 7 Adr Dat Prchen

unsigned char s88_message[S88_REPORT_SIZE*2 + 1];
unsigned char *s88_mes_ptr;

/// das s88_change array durchsuchen und nderungen an den PC melden
/// hier nur die echten Hardwaremelder - im Falle Xpressnet werden die anderen per Schaltinfo gemeldet.
/// return true = ja, report gemacht
///        false = kein neuer report
bool make_s88_lenz_feedback_report(unsigned char *message)
  {
    unsigned char i=0, report_i=0, report_max;                 

    report_max = s88_sizesum;               // limit to real hw only

    while((i<report_max) && (report_i<S88_REPORT_SIZE*2))     
      {                                    // Schleife ber ganzes Array
        if (s88_change[i])
          {
            if (s88_change[i] & 0b11110000)
              {
                // change in upper nibble
                create_xpressnet_feedback(i*8 + 4, message+report_i+1);        //  was  s88_mes_ptr+report_i+1);
                s88_change[i] &= 0b00001111;   // clear change
                report_i += 2;
              }  
            else
              {
                // change in lower nibble
                create_xpressnet_feedback(i*8, message+report_i+1);
                s88_change[i] &= 0b11110000;   // clear change
                report_i += 2;
              }  
          }
        i++;
      }
  
    if (report_i == 0)
      {
        s88_event &= ~(1<<S88_HW_CHANGE);
        return 0;               // false, no message built
      }

     /* s88_message[0] */
     *message =  0x40 | report_i;          // im header die Lnge eintragen

     return 1;
  }

bool make_s88_lenz_schalt_report(unsigned char *message)
  {
    #if (TURNOUT_FEEDBACK_ENABLED == 1)    
        unsigned char i=0, report_i=0, report_max;

        if (s88_mode & (1<<S88_FEEDBACK))
          {
            report_max = feedback_s88_size;  // limit to feedback turnouts

            while((i<report_max) && (report_i<S88_REPORT_SIZE*2))     
              {                                    // Schleife ber ganzes Array
                if (s88_change[i])
                  {
                    if (s88_change[i] & 0b00000011)
                      {
                        // change in lower pair
                        create_xpressnet_schaltinfo(i*4, message+report_i+1); 
                        s88_change[i] &= 0b11111100;   // clear change
                        report_i += 2;
                      }  
                    else if (s88_change[i] & 0b00001100)
                      {
                        create_xpressnet_feedback(i*4 + 1, message+report_i+1);
                        s88_change[i] &= 0b11110011;   // clear change
                        report_i += 2;
                      } 
                    else if (s88_change[i] & 0b00110000)
                      {
                        create_xpressnet_feedback(i*4 + 2, message+report_i+1);
                        s88_change[i] &= 0b11001111;   // clear change
                        report_i += 2;
                      }
                    else
                      {
                        // change in upper pair
                        create_xpressnet_feedback(i*4 + 2, message+report_i+1);
                        s88_change[i] &= 0b00111111;   // clear change
                        report_i += 2;
                      }  
                  }
                i++;
              }

            if (report_i == 0)
              {
                s88_event &= ~(1<<S88_TRNT_CHANGE);
                return 0;                           // false, no message built
              }

            *message =  0x40 | report_i;          // im header die Lnge eintragen

            return 1;
         }
    #endif
    return 0; // no message built
  }


void make_s88_lenz_report(void)
  {
    while(tx_fifo_ready() && make_s88_lenz_feedback_report(s88_message))
      { 
        pc_send_lenz(s88_mes_ptr = s88_message);
      }
  /*  while(tx_fifo_ready() && make_s88_lenz_schalt_report(s88_message))
      { 
        pc_send_lenz(s88_mes_ptr = s88_message);
      } */
  }

#endif // (PARSER == LENZ)

//==============================================================================
// Intelliboxroutinen
//
/// make_s88_ib_report
/// Action:     das s88_change array durchsuchen und nderungen an den PC melden
///             es wird jeweils ein Modul (d.h. 16 Bit bertragen)
/// Parameter:  none
/// Return:     true = ja, report gemacht (s88_ib_report)
///             false = kein neuer report
//
unsigned char s88_ib_report[3];

unsigned char invert_order(unsigned char d)
  {
    unsigned char i, retval = 0;

    for (i=0; i<8; i++)
      {
        retval <<= 1;
        if (d & 0x01) retval |= 1;
        d >>= 1;
      }
    return(retval);
  }

void make_s88_ib_single_report(unsigned char modul)
  {
    // adressing starts with 0
    s88_ib_report[0] = modul+1;
    s88_ib_report[1] = invert_order(s88_data[modul*2]);
    s88_ib_report[2] = invert_order(s88_data[modul*2+1]);
   
    s88_change[modul*2] &= 0;     // clear change
    s88_change[modul*2+1] &= 0;   // clear change 
  }

bool make_s88_ib_report(void)
  {
    
    unsigned char modul;
    s88_ib_report[0] = 0;                  // default: no report

    modul=0;  

    while( (modul*2) < s88_total )     
      {                                    // Schleife ber ganzes Array (hier keine Unterscheidung Feedback-Hardware)
        if (s88_change[modul*2] || s88_change[modul*2 + 1]) 
          {
            // there is a change
            make_s88_ib_single_report(modul);
            
            return 1; 
          }
        modul++;
      }  
     s88_event &= ~(1<<S88_HW_CHANGE);
     s88_event &= ~(1<<S88_TRNT_CHANGE);
     return 0;               // false, no message built
  }


//==============================================================================
// parameter: s88_init_mode
// return true: okay
// false: size too large
unsigned char init_s88(t_s88_init_mode s88_init_mode)
  {
    unsigned int temp1, temp2;

    xpressnet_feedback_mode = eeprom_read_byte((void *)eadr_xpressnet_feedback); 

    s88_mode = eeprom_read_byte((void *)eadr_s88_mode); 
    
    s88_pulsduration = eeprom_read_byte((void *)eadr_s88_clk_timing); 
    if (s88_pulsduration == 0) s88_pulsduration = 1;
    if (s88_pulsduration > 20) s88_pulsduration = 20;   // tick is 2us; 20*10us = 200us -> max. 100 for signed counter-compare
    
    feedback_s88_offset = eeprom_read_byte((void *)eadr_feedback_s88_offset);
    feedback_s88_size = eeprom_read_byte((void *)eadr_feedback_s88_size);
    feedback_s88_type = eeprom_read_byte((void *)eadr_feedback_s88_type);      

    if (s88_mode & (1<<S88_HW_READ))
      {
        S88_RESET_LOW;
        S88_CLK_LOW;
        S88_LOAD_LOW;
      }

    // get S88-stream sizes; 
    
    if (s88_init_mode == READ_FROM_EEPROM)
      {
        s88_size1 = eeprom_read_byte((void *)eadr_s88_size1); 
        s88_size2 = eeprom_read_byte((void *)eadr_s88_size2);
        s88_size3 = eeprom_read_byte((void *)eadr_s88_size3);
      }

    s88_sizemax = s88_size1;
    if (s88_size2 > s88_sizemax) s88_sizemax = s88_size2;
    if (s88_size3 > s88_sizemax) s88_sizemax = s88_size3;

    s88_sizesum = s88_size1 + s88_size2 + s88_size3;
    if ((s88_sizesum) > SIZE_S88_MAX)   // max. is 255 -> 2040 events
      {
        return(0);
      }
 
    temp1 = s88_sizesum;
    // is there something larger?

    temp2 = feedback_s88_size + feedback_s88_offset;
    if (temp2 > temp1) temp1 = temp2;

    temp2 = eeprom_read_byte((void *)eadr_s88_total_from_pc);        // has a larger value been loaded?
    if (temp2 > temp1) temp1 = temp2;

    if (temp1 < SIZE_S88_MAX)  s88_total = temp1;
    else  s88_total = SIZE_S88_MAX;      // limit to max memory

    s88_state = IDLE;

    read_s88();                          // lscht auch s88_event und _alle_ change_flags
                                         // impl. with software delay
    return(1);
  }


// Note: some state transistions are without condition, just to allow
// pseudo multitasking ;-)

void run_s88(void)  
  {
    unsigned char i,flag;

    // check for feedback from dccout

    #if (TURNOUT_FEEDBACK_ENABLED == 1)
    if (s88_mode & (1<<S88_FEEDBACK))
      {
        cli();
        if (feedback_ready & (1<<FB_READY)) 
          {
            feedback_ready &= ~(1<<FB_READY);
            sei();
            enter_feedback_to_s88(feedbacked_accessory, feedback_ready & (1<<FB_OKAY));
          }
        sei();
      }
    #endif

    switch (s88_state)
      {
        case IDLE:
                                                // wait some time??
            if (s88_mode & (1<<S88_HW_READ))
              {
                s88_state = DO_S88_LEAD_IN;
                s88_reset_phase = 0;
              }
            else
              {
                s88_state = DO_CHECK;
              }
            break;
        case DO_S88_LEAD_IN:
            // old: make_s88_reset();                   // am S88 die Rckmelder laden
            // new:
            if (make_s88_reset_phased())
              {
                s88_current_i = 0;
                s88_current_b = 1;
                check_s88_bit(s88_current_i, s88_current_b);    // read first bit
                s88_clock_phase = 0;
                s88_state = DO_READ_BYTE;
              }
            break;
        case DO_READ_BYTE:
            if (s88_clk_pulse_phased())
              {
                s88_current_b = s88_current_b<<1;
                if (s88_current_b == 0)                         // this byte now completely done
                  {
                    s88_current_b = 1;                          // restart byte shifter
                    s88_current_i++;
                    if (s88_current_i == s88_sizemax)  s88_state = DO_LAST_CLK_PULSE;
                    else
                      {
                        check_s88_bit(s88_current_i, s88_current_b);     // read first bit of next byte
                      }                    
                  }
                else
                  {
                    check_s88_bit(s88_current_i, s88_current_b);        // read next bit
                  }
              }
            break;
        case DO_LAST_CLK_PULSE:
            if (s88_clk_pulse_phased())              // to be complete: last clk pulse
                s88_state = DO_CHECK;
            break;

        case DO_CHECK:
            flag = 0;

            for (i=0; i<s88_sizesum; i++)             // scan hw-array for changes
              {
                flag |= s88_change[i];
              }

            if (flag)  s88_event |= (1<<S88_HW_CHANGE);

            s88_state = IDLE;
            break;

// begin: now void!!!

        case DO_MESSAGE:
            #if (PARSER == LENZ)
                if (tx_fifo_ready())                // wir knnten was absenden
                  {
                    if (make_s88_lenz_feedback_report(s88_message))
                      { 
                        pc_send_lenz(s88_mes_ptr = s88_message);
                      }
                    else 
                      {
                        s88_state = DO_MESSAGE2;
                      }
                  }
            #endif // PARSER = LENZ
            #if (PARSER == INTELLIBOX)
                s88_event = 1;                     // hier nur das Flag
                s88_state = IDLE;
            #endif
            break;
        case DO_MESSAGE2:
            #if (PARSER == LENZ)
                if (tx_fifo_ready())                // wir knnten was absenden
                  {
                    if (make_s88_lenz_schalt_report(s88_message))
                      { 
                        pc_send_lenz(s88_mes_ptr = s88_message);
                      }
                    else 
                      {
                        s88_state = IDLE;
                      }
                  }
            #endif // PARSER = LENZ
            break;
      }
  }

//===================================================================================
//
// HSI88 Mode
//
// Emulation des HSI88 Interfaces von Littfinski
// keine Sicherung gegen Timouts, ist auch beim Orginal HSI88 so :-(
//
// getestet unter TC5.5 und Railware 5.01 -> okay
//
//===================================================================================

#if (HSI88_ENABLED == 1)

unsigned char terminalmode = 0;          // if 0: binary answers
                                         // if 1: give answers in ASCII
unsigned int no_of_modules_links;
unsigned int no_of_modules_mitte;
unsigned int no_of_modules_rechts;
unsigned int no_of_modules;

#ifdef SPECIAL_STDIO
unsigned char my_scan(unsigned char *buffer)
  {
    unsigned char retval;
    retval = (*buffer++ & 0x0f) * 10;
    retval += (*buffer & 0x0f);
    return(retval);
  }

void my_sprint_num(char *buffer, unsigned char num)  // fest 2 digits
  {
    *buffer++ = 0x30 + (num / 10);
    *buffer++ = 0x30 + (num % 10);
    *buffer   = 0;
  }

void my_sprint_hex(unsigned char *buffer, unsigned char num)
  {
    unsigned char hex;
    hex = num >> 4;
    if (hex < 10) hex += 0x30;
    else hex += 0x37;   // +offset fr ABCDEF
    *buffer++ = hex;
    hex = num & 0x0f;
    if (hex < 10) hex += 0x30;
    else hex += 0x37;   // +offset fr ABCDEF
    *buffer++ = hex;
    *buffer   = 0;      
  }
#endif


void hsi_send_num(unsigned char nummer)
  {
    char buffer[3];  // 2+1 fr sprintf + 0
    if (terminalmode)
      {
        while (tx_fifo_ready() == 0);
        #ifdef SPECIAL_STDIO
          my_sprint_num(buffer, nummer);
        #else
          sprintf(buffer, "%02d", nummer);
        #endif
        tx_fifo_write(buffer[0]);
        tx_fifo_write(buffer[1]);
      }
    else
      {
        while (tx_fifo_ready() == 0);
        tx_fifo_write(nummer);
      }
  }

void hsi_send_byte(unsigned char nummer)
  {
    unsigned char buffer[3];  // 2+1 fr sprintf + 0
    if (terminalmode)
      {
        while (tx_fifo_ready() == 0);
        #ifdef SPECIAL_STDIO
          my_sprint_hex(buffer, nummer);
        #else
          sprintf(buffer, "%02x", nummer);
        #endif
        tx_fifo_write(buffer[0]);
        tx_fifo_write(buffer[1]);
      }
    else
      {
        while (tx_fifo_ready() == 0);
        tx_fifo_write(nummer);
      }
  }



/// Alle Module melden
/// hier nur der Nummmernteil, inkl. <cr>

void full_hsi_message(void)
  {
    
    unsigned char i;

    hsi_send_num((char) no_of_modules);
    for (i=0; i<no_of_modules; i++)
      {
        hsi_send_num(i+1);
        hsi_send_byte(s88_data[2*i+1]);          // High byte first
        hsi_send_byte(s88_data[2*i]);
      }
    while (tx_fifo_ready() == 0);
    tx_fifo_write(0x0d);
  }




// const char copyright[] PROGMEM = "Ver. x.xx / dd.mm.jj / HSI-88 / (c) LDT\r";
const char copyright[] PROGMEM = "Ver. 0.31 / 29.05.06 / HSI-88 / OpenDCC\r";

void copyright_message(void)
  {
     char zeichen;
     unsigned char i=0;

     while ( (zeichen = pgm_read_byte(&copyright[i++])) )
       {
         while (tx_fifo_ready() == 0);
         tx_fifo_write(zeichen);
       }
  }

//               ;Antwortlnge: 41 Byte


//
void HSI88_parser(void)
  {
     unsigned char cmd;
     unsigned char myzahl[3];  // 2+1 fr sprintf + 0

    //kw terminalmode = 1;
    //kw hsi_send_num(23);

    if (rx_fifo_ready())   // es geht ein neues Kommando los
      {
        cmd = rx_fifo_read();
        switch(cmd)
          {
            case 't':                                         // Command: toggle terminal mode
                while(!rx_fifo_ready());
                if (rx_fifo_read() == 0x0d)                   // discard terminating <cr>
                  {
                    tx_fifo_write('t');
                    if (terminalmode)
                      {
                        terminalmode = 0;
                        tx_fifo_write('0');
                      }
                    else
                      {
                        terminalmode = 1;
                        tx_fifo_write('1');
                      }
                    LED_STOP_OFF;
                    tx_fifo_write(0x0d);
                   }
                 else
                   {
                      LED_STOP_ON;
                   }
                 break;

            case 's':                                          // Command: set modules
                if (terminalmode)
                  {
                    while(!rx_fifo_ready());
                    myzahl[0] = rx_fifo_read();
                    while(!rx_fifo_ready());
                    myzahl[1] = rx_fifo_read();
                    myzahl[2] = 0;
                     #ifdef SPECIAL_STDIO
                         no_of_modules_links = my_scan(myzahl);
                     #else
                         sscanf(myzahl, "%u", &no_of_modules_links);
                     #endif
                     
                    while(!rx_fifo_ready());
                    myzahl[0] = rx_fifo_read();
                    while(!rx_fifo_ready());
                    myzahl[1] = rx_fifo_read();
                    myzahl[2] = 0;
                    #ifdef SPECIAL_STDIO
                         no_of_modules_mitte = my_scan(myzahl);
                    #else
                         sscanf(myzahl, "%u", &no_of_modules_mitte);
                    #endif
             
                    while(!rx_fifo_ready());
                    myzahl[0] = rx_fifo_read();
                    while(!rx_fifo_ready());
                    myzahl[1] = rx_fifo_read();
                    myzahl[2] = 0;
                    #ifdef SPECIAL_STDIO
                         no_of_modules_rechts = my_scan(myzahl);
                    #else
                         sscanf(myzahl, "%u", &no_of_modules_rechts);
                    #endif
                     
                    no_of_modules = no_of_modules_links + no_of_modules_mitte + no_of_modules_rechts;
                    if (no_of_modules > 31)        // 31 ist maximum laut Littfinski, auch TC kommt
                                                    // nicht mit grerer Zahl klar!
                      {
                         no_of_modules_links = 2; 
                         no_of_modules_mitte = 2; 
                         no_of_modules_rechts = 2; 
                         no_of_modules = 6; 
                         LED_STOP_ON;
                       }
                    while(!rx_fifo_ready());
                    if (rx_fifo_read() == 0x0d)                 // discard terminating <cr>
                       {
                         tx_fifo_write('s');
                         hsi_send_num(no_of_modules);
                         tx_fifo_write(0x0d);
                       }
                     else
                       { 
                         LED_STOP_ON;
                       }
                   }
                 else  // terminalmode = 1;
                   {
                     while(!rx_fifo_ready());
                     no_of_modules_links = rx_fifo_read();
                     while(!rx_fifo_ready());
                     no_of_modules_mitte = rx_fifo_read();
                     while(!rx_fifo_ready());
                     no_of_modules_rechts = rx_fifo_read();
                     no_of_modules = no_of_modules_links + no_of_modules_mitte + no_of_modules_rechts;
                     if (no_of_modules > 31)
                       {
                         no_of_modules_links = 2; 
                         no_of_modules_mitte = 2; 
                         no_of_modules_rechts = 2; 
                         no_of_modules = 6; 
                         LED_STOP_ON;
                       }
                    while(!rx_fifo_ready());
                    if (rx_fifo_read() == 0x0d)                   // discard terminating <cr>
                      {
                        tx_fifo_write('s');
                        tx_fifo_write((char) no_of_modules);
                        tx_fifo_write(0x0d);
                      }
                    else
                      {
                         LED_STOP_ON;
                      }
                  }
                // init set modules and performs the first read
                LED_CTRL_ON;
                s88_size1 = no_of_modules_links*2;
                s88_size2 = no_of_modules_mitte*2;
                s88_size3 = no_of_modules_rechts*2;
                init_s88(KEEP_SIZES);
                LED_CTRL_OFF;
                // now send answer 2
                tx_fifo_write('i');
                 
                full_hsi_message();
                break;
             case 'm':                                              // Command: full read
                 while(!rx_fifo_ready());
                 if (rx_fifo_read() == 0x0d)                        // discard terminating <cr>
                   {
                     tx_fifo_write('m');
                     LED_CTRL_ON;
                     read_s88();
                     LED_CTRL_OFF;
                     LED_STOP_OFF;
                     full_hsi_message();
                   }
                 else
                   {
                     LED_STOP_ON;
                   }
                 break;
             case 'v':                                               // Command: Version
                 while(!rx_fifo_ready());
                 if (rx_fifo_read() == 0x0d) 
                   {
                     copyright_message();
                     LED_STOP_OFF;
                   }
                 else
                   {
                      LED_STOP_ON;
                   }
                 break;
          }
      }
    else
      {                                                              // check for push service
         unsigned char i;
         unsigned char changed_modules;

         LED_CTRL_ON;
         check_s88();
         LED_CTRL_OFF;
         changed_modules = 0;
         for (i=0; i<no_of_modules; i++)                            // scan array for changes
           {
             if (s88_change[2*i] | s88_change[2*i+1])
               {
                 changed_modules++;
               }
           }
         if (changed_modules)
           {
             tx_fifo_write('i');
             hsi_send_num(changed_modules);
             for (i=0; i<no_of_modules; i++)                        // report all changed modules to pc
               {
                 if (s88_change[2*i] | s88_change[2*i+1])
                   {
                     s88_change[2*i]=0;
                     s88_change[2*i+1]=0;
                     hsi_send_num(i+1);
                     hsi_send_byte(s88_data[2*i+1]);          // High Byte first
                     hsi_send_byte(s88_data[2*i]);
                   }
               }
             tx_fifo_write(0x0d);
           }
      }
  }

#else

void HSI88_parser(void)
  {
    #warning: HSI88-Support turned off.
  }

#endif  // HSI88_ENABLE

