//----------------------------------------------------------------
//
// OpenDCC
//
// Copyright (c) 2006 Kufer
// 
// This source file is subject of the GNU general public license 2,
// that is available at the world-wide-web at
// http://www.gnu.org/licenses/gpl.txt
//
//-----------------------------------------------------------------
//
// file:      status.c
// author:    Wolfgang Kufer
// contact:   kufer@gmx.de
// history:   2006-04-28 V0.1 started
//            2006-05-15 V0.2 bugfixes in LED-Control
//            2006-06-01 V0.3 debouncer for short curcuit inputs
//            2006-06-20 V0.4 added new timerengines with timerval
//            2007-10-15 V0.5 key_go_at_start_pressed neu dazu.
//            2008-01-08 V0.6 short_turnoff_time is fetched from eeprom
//            2008-01-11 V0.7 prog_short_turnoff_time is fetched from eeprom
//            2008-01-16 V0.8 is_prog_state() added
//                            moved to XP
//            2008-07-18 V0.9 ext_stop added
//            2008-07-18 V0.9 ext_stop added; also DMX out is parallel controlled
//
//-----------------------------------------------------------------
//
// purpose:   lowcost central station for dcc
// content:   control of LED via state engines
//            timertick
//            check for key strokes and output short
//            shorts are reported with "event_send" and it is up to
//            the selected parser, to report this event to the host.
//            
//
// 2do:       output short for prog mode not yet impl.
// 
//-----------------------------------------------------------------


#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>           // put var to program memory
#include <avr/eeprom.h>
#include <string.h>

#include "config.h"                 // general structures and definitions
#include "status.h"                

#if (PARSER == INTELLIBOX)
 #include "ibox_parser.h"           // tell pc about status changes
#endif
#if (PARSER == LENZ)
 #include "lenz_parser.h"             
#endif


//---------------------------------------------------------------------------
// data for timeout and debounce
// every task loads its  corresponding member with the desired timeout value
// ISR timer_tick decrements this value every tick (when not 0)

volatile signed int timerval;    // is incremented every timertick, wrap around

t_no_timeout no_timeout;


unsigned char key_go_at_start_pressed;

#define DEBOUNCE_TIMEOUT    15000L

#define TURN_OFF_IF_SHORT    1      // 0: don't turn off output in case of short
                                    //    WARNING: only for debugging!
                                    // 1: in case of short, turn off output

#if (TURN_OFF_IF_SHORT != 1)
#warning: SHORT Detection not activated - only for debugging hardware!
#endif

unsigned char short_turnoff_time;   // time from detect of a short until disable output
                                    // is loaded during init from eeprom
unsigned char prog_short_turnoff_time;   // is loaded during init from eeprom


//-------------------------------------------------------------------------------------
// Timer 0 is used to create a 5ms Timertick

void init_tick(void)
  {
    timerval = 0;

    // clear all timeouts

    no_timeout.rs232 = 0;
    no_timeout.key_go = 0;
    no_timeout.key_stop = 0; 
    no_timeout.parser = 0; 
    no_timeout.main_short = 0; 
    no_timeout.prog_short = 0; 

    // load timing values from eeprom

    short_turnoff_time = eeprom_read_byte((void *)eadr_short_turnoff_time);
    if (short_turnoff_time == 0) short_turnoff_time = 1;

    prog_short_turnoff_time = eeprom_read_byte((void *)eadr_prog_short_toff_time);
    if (prog_short_turnoff_time == 0) prog_short_turnoff_time = 1;

    #if (PROCESSOR == ATMEGA644P)
        // Timer/Counter 0 initialization
        // Clock source: System Clock / 1024
        // Clock value: 200Hz; TICK_PERIOD defined in config.h
        // Mode: Output Compare   0x05
        // OC0 output: Disconnected
        TCCR0A = (0<< COM0A1)   // 00 = normal port mode
               | (0<< COM0A0)
               | (0<< COM0B1)   // 00 = normal port mode
               | (0<< COM0B0)
               | (0<< WGM01)    // WGM = 010: CTC
               | (0<< WGM00);
        TCCR0B = (0<< FOC0A)    // 0 = no forced compare match
               | (0<< FOC0B)    // 0 = no forced compare match
               | (0<< WGM02)    // WGM = 000: normal mode
               | (1<<CS02)      // CS = 000: stopped
               | (0<<CS01)      //      001 = run, 010 = div8, 011=div64, 100=div256, 101=div1024. 
               | (1<<CS00);     //      110 = ext, 111 = ext
        OCR0A =  F_CPU / 1024L * TICK_PERIOD / 1000000L;               // 8 Bit compare on A
        TCNT0=0x00;
        TIMSK0 |= (1<<OCIE0A);   // Timer0 Compare Int.
    
    
    #else
        // Timer/Counter 0 initialization
        // Clock source: System Clock / 1024
        // Clock value: 200Hz; TICK_PERIOD defined in config.h
        // Mode: Output Compare   0x05
        // OC0 output: Disconnected
        TCCR0 = (0<<FOC0)     // 0 = no forced compare match
              | (0<<WGM00)    // WGM = 10: CTC-Mode (clear on compare match)
              | (0<<COM01)    // 00 = OC0 disconnected.
              | (0<<COM00)
              | (1<<WGM01)    //
              | (1<<CS02)     // CS = 000: stopped
              | (0<<CS01)     // 001 = run, 010 = div8, 011=div64
              | (1<<CS00);    // 100 = div256, 101 = div1024
        OCR0 =  F_CPU / 1024L * TICK_PERIOD / 1000000L;               // 8 Bit compare 
        TCNT0=0x00;
        TIMSK |= (1<<OCIE0);      // Timer0 Compare Int.
    #endif
  }

void stop_tick(void)
  {
    #if (PROCESSOR == ATMEGA644P)
        TIMSK0 &= ~(1<<OCIE0A);      // Disable Timer0 Compare
    #else
        TIMSK &= ~(1<<OCIE0);      // Disable Timer0 Compare
    #endif
  }

//----------------------------------------------------------------
// Statusanzeigen
//
// Legende:  - = off;  + = on;   O = blinkend
//
// Es gibt 3 LED: Ctrl Stop Go 
//
// diese werden wie folgt gesteuert:
//------------+----+----+----+-------------------------------------
// Zustand:   |CTRL|Stop| Go | Bedeutung:
//            |    |    |    |
// Farbe:     | or | rd | gn |
//            |    |    |    | 
// RUN_OKAY   | -  | -  | +  | DCC Running
// RUN_STOP   | -  | +  | +  | DCC Running, all Engines Emergency Stop
// RUN_OFF    | -  | +  | -  | Output disabled (2*Taste, PC)
// RUN_SHORT  | -  | O  | O  | Kurzschlu
// RUN_PAUSE  | -  | -  | O  | DCC Running, all Engines Speed 0
//-----------------------------------------------------------------
// PROG_OKAY  | +  | -  | +  | DCC Programming
// PROG_SHORT | +  | O  | O  | Kurzschlu
// PROG_OFF   | +  | +  | -  | diabled
// PROG_ERROR | O  | -  | +  | ACK fehlt
//
//--------------------------------------------------------------
//
// RS232_LED zeigt den Zustand der PC-Verbindung an
// 
//
// NO CONN (RTS) ++++++------++++++
// CONNECTED: ++++++++++++++++++
// RUNNING: +++++-+++++-+++++
// STOPPED (CTS) +-+-+-+-+-+-
// 
//

t_led_pwm led_pwm;


void led_tick(void)                // call only inside ISR
  {           
    if (led_pwm.rs232_rest != 0)
      {
        if (--led_pwm.rs232_rest == 0)
          {
             if (LED_RS232_STATE)
              {
                led_pwm.rs232_rest = led_pwm.rs232_offtime;
                LED_RS232_OFF;
              }
            else
              {
                led_pwm.rs232_rest = led_pwm.rs232_ontime;
                LED_RS232_ON;
              }
          }
      }
    if (led_pwm.go_rest != 0)
      {
        if (--led_pwm.go_rest == 0)
          {
            if (LED_GO_STATE)
              {
                led_pwm.go_rest = led_pwm.go_offtime;
                LED_GO_OFF;
              }
            else
              {
                led_pwm.go_rest = led_pwm.go_ontime;
                LED_GO_ON;
              }
          }
      }
    if (led_pwm.stop_rest != 0)
      {
        if (--led_pwm.stop_rest == 0)
          {
            if (LED_STOP_STATE)
              {
                led_pwm.stop_rest = led_pwm.stop_offtime;
                LED_STOP_OFF;
              }
            else
              {
                led_pwm.stop_rest = led_pwm.stop_ontime;
                LED_STOP_ON;
              }
          }
      }
    if (led_pwm.ctrl_rest != 0)
      {
        if (--led_pwm.ctrl_rest == 0)
          {
            if (LED_CTRL_STATE)
              {
                led_pwm.ctrl_rest = led_pwm.ctrl_offtime;
                LED_CTRL_OFF;
              }
            else
              {
                led_pwm.ctrl_rest = led_pwm.ctrl_ontime;
                LED_CTRL_ON;
              }
          }
      }

  }
  
  


void timeout_tick(void)
  {
    if (no_timeout.rs232) no_timeout.rs232--;  
    if (no_timeout.key_go) no_timeout.key_go--;  
    if (no_timeout.key_stop) no_timeout.key_stop--;  
    if (no_timeout.parser) no_timeout.parser--;  
    if (no_timeout.main_short) no_timeout.main_short--;  
    if (no_timeout.prog_short) no_timeout.prog_short--;  
  }


///-------------------------------------  This interrupt ticks every 5ms 

#if (PROCESSOR == ATMEGA644P)
        ISR(TIMER0_COMPA_vect)
#else
        ISR(TIMER0_COMP_vect) 
#endif

  {
    sei();                        // allow further ints
    timerval++; 
    led_tick();                   // run the leds
    timeout_tick();               // run all timeout-ticks and debouncer
  }


//===================================================================================
// keyboard
// only consists of two tracers STOP and GO
//

#define UNPRESSED 0
#define DEBOUNCE  1
#define PRESSED   2

#define GO_STROKE    1
#define STOP_STROKE  2


unsigned char key_stop = UNPRESSED;
unsigned char key_go = UNPRESSED;


// returns 0 if no keystroke
// returns 1 if stop
// returns 2 if run

unsigned char run_key(void)
  {
    #if (DEBUG == 3)
      return(UNPRESSED);
    #endif
    switch(key_stop)
      {
        case UNPRESSED:
            if (KEY_STOP_PRESSED)
              {
                key_stop = DEBOUNCE;
                no_timeout.key_stop = DEBOUNCE_TIMEOUT / TICK_PERIOD;   // 15ms
              }
            break;
        case DEBOUNCE:
            if (!KEY_STOP_PRESSED)
              {
                key_stop = UNPRESSED;  // nochmal von vorn
              }
            else
              {
                if (no_timeout.key_stop) {}
                else
                  {
                    key_stop = PRESSED;
                    return(STOP_STROKE);     // exit here
                  } 
              }
            break;
        case PRESSED:
            if (!KEY_STOP_PRESSED)
              {
                key_stop = UNPRESSED;  // nochmal von vorn
              }
            break;
      }
    switch(key_go)
      {
        case UNPRESSED:
            if (KEY_GO_PRESSED)
              {
                key_go = DEBOUNCE;
                no_timeout.key_go = DEBOUNCE_TIMEOUT / TICK_PERIOD;   // 15ms
              }
            break;
        case DEBOUNCE:
            if (!KEY_GO_PRESSED)
              {
                key_go = UNPRESSED;  // nochmal von vorn
              }
            else
              {
                if (no_timeout.key_go) {}
                else
                  {
                    key_go = PRESSED;
                    return(GO_STROKE);
                  }
                
              }
            break;
        case PRESSED:
            if (!KEY_GO_PRESSED)
              {
                key_go = UNPRESSED;  // nochmal von vorn
              }
            break;
      }
    return(UNPRESSED);
  }



//===================================================================================
/// output short protection
/// this routine debounces short curcuit message from output stage
/// act on MAIN_IS_SHORT and PROG_IS_SHORT
//
// a filter is performed on MAIN_IS_SHORT:
// every scan: if MAIN is short, the mean value is increased by 4
//          

#define UNPRESSED 0
#define DEBOUNCE  1
#define PRESSED   2

#define IS_SHORT    1
#define IS_OKAY     0


unsigned char main_short = UNPRESSED;
signed char main_short_mean = 0;

unsigned char prog_short = UNPRESSED;
signed char prog_short_mean = 0;

// returns 0 if no short
// returns 1 if short is detected and debounced

unsigned char run_main_short_check(void)
  {
    #if (DEBUG == 3)
      return(IS_OKAY);
    #endif

    switch(main_short)
      {
        case UNPRESSED:
            if (MAIN_IS_SHORT)
              {
                main_short = DEBOUNCE;
                no_timeout.main_short = short_turnoff_time;   // default 15ms
                main_short_mean = 0;
              }
            break;
        case DEBOUNCE:
            if (!MAIN_IS_SHORT)
              {
                main_short_mean -= 1;
                if (main_short_mean < -100) main_short = UNPRESSED;  // nochmal von vorn
              }
            else
              {
                main_short_mean += 4;
                if (main_short_mean > 100) main_short_mean = 100;   // limit
                
                if (no_timeout.main_short) {}
                else
                  {
                    main_short = PRESSED;
                    return(IS_SHORT);     // exit here
                  } 
              }
            break;
        case PRESSED:
            if (!MAIN_IS_SHORT)
              {
                main_short_mean -= 1;
                if (main_short_mean < -100) main_short = UNPRESSED;  // nochmal von vorn
              }
            else
              {
                main_short_mean += 2;
                if (main_short_mean > 100) main_short_mean = 100;   // limit

                return(IS_SHORT);        // still short - exit here
              }
            break;
      }
    return(IS_OKAY);
  }


unsigned char run_prog_short_check(void)
  {
    #if (DEBUG == 3)
      return(IS_OKAY);
    #endif

    switch(prog_short)
      {
        case UNPRESSED:
            if (PROG_IS_SHORT)
              {
                prog_short = DEBOUNCE;
                no_timeout.prog_short = prog_short_turnoff_time;   // default 40ms
              }
            break;
        case DEBOUNCE:
            if (!PROG_IS_SHORT)
              {
                prog_short = UNPRESSED;  // nochmal von vorn
              }
            else
              {
                if (no_timeout.prog_short) {}
                else
                  {
                    prog_short = PRESSED;
                    return(IS_SHORT);     // exit here
                  } 
              }
            break;
        case PRESSED:
            if (!PROG_IS_SHORT)
              {
                prog_short = UNPRESSED;  // nochmal von vorn
              }
            else
              {
                return(IS_SHORT);        // still short - exit here
              }
            break;
      }
    return(IS_OKAY);
  }



//===================================================================================
// Interface-Routines to other programs
// Alle Zeiten in us

t_rs232_led_state current_led_state = 0xFF;


#if (WATCH_S88_SIZE1 == 1) 
void set_alarm_led(unsigned char next)
  {

    if (next == current_led_state) return;     // das kennen wir schon
    current_led_state = next;  
    switch(next)
      {
        case 0:
            break;
        case 1:
            LED_RS232_ON;
            led_pwm.rs232_rest = 200000L / TICK_PERIOD;
            led_pwm.rs232_ontime = 100000L / TICK_PERIOD;
            led_pwm.rs232_offtime = 100000L / TICK_PERIOD;
            break;
      }
  }

#endif


void set_led_rs232(t_rs232_led_state next)
  {

    #if (WATCH_S88_SIZE1 == 1) 
      return;                // block rs232 calls to led
    #endif

    if (next == current_led_state) return;     // das kennen wir schon
    current_led_state = next;  
    unsigned char mysreg = SREG;
    cli();
    switch(next)
      {
        case DISCONNECTED:
            LED_RS232_OFF;
            led_pwm.rs232_rest = 50;
            led_pwm.rs232_ontime = 200000L / TICK_PERIOD;   
            led_pwm.rs232_offtime = 1000000L / TICK_PERIOD;
            break;
        case CONNECTED:
            LED_RS232_ON;
            led_pwm.rs232_rest = 0;              // forever
            // led_pwm.rs232_ontime = 50;
            // led_pwm.rs232_offtime = 50;
            break;
        case RUNNING:
            LED_RS232_ON;
            led_pwm.rs232_rest = 125000L / TICK_PERIOD;
            led_pwm.rs232_ontime = 250000L / TICK_PERIOD;
            led_pwm.rs232_offtime = 100000L / TICK_PERIOD;
            break;
        case STOPPED:
            LED_RS232_OFF;
            led_pwm.rs232_rest = 20000L / TICK_PERIOD;
            led_pwm.rs232_ontime = 50000L / TICK_PERIOD;
            led_pwm.rs232_offtime = 50000L / TICK_PERIOD;
            break;
      }
    SREG = mysreg;
  }


void set_led_state(t_opendcc_state next)
  {
    // Zugriff auf .rest und PORT muss atomar sein.
    
    unsigned char mysreg = SREG;
    cli();
    switch(next)
      {
//------------+----+----+----+-------------------------------------
// Zustand:   |CTRL|Stop| Go | Bedeutung:
//            |    |    |    |
// Farbe:     | or | rd | gn |
//            |    |    |    | 
// RUN_OKAY   | -  | -  | +  | 
// RUN_STOP   | -  | +  | +  | DCC Running, all Engines Emergency Stop
// RUN_OFF    | -  | +  | -  | Output disabled (2*Taste, PC)
// RUN_SHORT  | -  | O  | O  | Kurzschlu
// RUN_PAUSE  | -  | -  | O  | DCC Running, all Engines Speed 0
//-----------------------------------------------------------------
// PROG_OKAY  | +  | -  | +  |
// PROG_SHORT | +  | O  | O  |
// PROG_OFF   | +  | +  | -  |
// PROG_ERROR | O  | -  | +  |

        case RUN_OKAY:            // RUN_OKAY   | -  | -  | +  | 
            LED_GO_ON;
            led_pwm.go_rest = 0;                // forever
            LED_STOP_OFF;
            led_pwm.stop_rest = 0;                // forever
            LED_CTRL_OFF;
            led_pwm.ctrl_rest = 0;                // forever
             break;
        case RUN_STOP:             // DCC Running, all Engines Emergency Stop
            LED_GO_ON;
            led_pwm.go_rest = 0;                // forever
            LED_STOP_ON;
            led_pwm.stop_rest = 0;                // forever
            LED_CTRL_OFF;
            led_pwm.ctrl_rest = 0;                // forever
            break;
        case RUN_OFF:                    // Output disabled (2*Taste, PC)
            LED_GO_OFF;
            led_pwm.go_rest = 0;                // forever
            LED_STOP_ON;
            led_pwm.stop_rest = 0;                // forever
            LED_CTRL_OFF;
            led_pwm.ctrl_rest = 0;                // forever
            break;
       case RUN_SHORT:                    // Kurzschluss, hektisch blinken im Wechsel
            LED_GO_OFF;
            led_pwm.go_rest = 100000L / TICK_PERIOD; 
            led_pwm.go_ontime = 150000L / TICK_PERIOD;
            led_pwm.go_offtime = 150000L / TICK_PERIOD;
            LED_STOP_ON;
            led_pwm.stop_rest = 100000L / TICK_PERIOD; 
            led_pwm.stop_ontime = 150000L / TICK_PERIOD;
            led_pwm.stop_offtime = 150000L / TICK_PERIOD;
            LED_CTRL_OFF;
            led_pwm.ctrl_rest = 0;                // forever
             break;
        case RUN_PAUSE:                   // DCC Running, all Engines Speed 0
            LED_GO_OFF;                    // rot langsam blinken
            led_pwm.go_rest = 0; 
            LED_STOP_ON;
            led_pwm.stop_rest = 500000L / TICK_PERIOD; 
            led_pwm.stop_ontime = 500000L / TICK_PERIOD;
            led_pwm.stop_offtime = 500000L / TICK_PERIOD;
            LED_CTRL_OFF;
            led_pwm.ctrl_rest = 0;                // forever
            break;
        case PROG_OKAY:
            LED_GO_ON;
            led_pwm.go_rest = 0;                // forever
            LED_STOP_OFF;
            led_pwm.stop_rest = 0;                // forever
            LED_CTRL_ON;
            led_pwm.ctrl_rest = 0;                // forever
             break;
        case PROG_SHORT:           // Kurzschlu, hektisch blinken
            LED_GO_OFF;
            led_pwm.go_rest = 100000L / TICK_PERIOD; 
            led_pwm.go_ontime = 100000L / TICK_PERIOD;
            led_pwm.go_offtime = 100000L / TICK_PERIOD;
            LED_STOP_ON;
            led_pwm.stop_rest = 100000L / TICK_PERIOD; 
            led_pwm.stop_ontime = 100000L / TICK_PERIOD;
            led_pwm.stop_offtime = 100000L / TICK_PERIOD;
            LED_CTRL_ON;
            led_pwm.ctrl_rest = 0;                // forever
             break;
        case PROG_OFF:
            LED_GO_OFF;
            led_pwm.go_rest = 0;                // forever
            LED_STOP_ON;
            led_pwm.stop_rest = 0;                // forever
            LED_CTRL_ON;
            led_pwm.ctrl_rest = 0;                // forever
            break;
        case PROG_ERROR:         // wie go, CTRL blinkt
            LED_GO_OFF;
            led_pwm.go_rest = 0;                // forever
            LED_STOP_ON;
            led_pwm.stop_rest = 0;                // forever
            LED_CTRL_ON;
            led_pwm.ctrl_rest = 100000L / TICK_PERIOD;
            led_pwm.ctrl_ontime = 200000L / TICK_PERIOD;
            led_pwm.ctrl_offtime = 200000L / TICK_PERIOD;
            break;
      }
    SREG = mysreg;
  }

t_opendcc_state opendcc_state;            // this is the current running state


void test_blink_leds(void)
  {
    LED_GO_OFF;
    led_pwm.go_rest = 50000L / TICK_PERIOD; 
    led_pwm.go_ontime = 300000L / TICK_PERIOD;
    led_pwm.go_offtime = 300000L / TICK_PERIOD;
    LED_STOP_OFF;
    led_pwm.stop_rest = 150000L / TICK_PERIOD; 
    led_pwm.stop_ontime = 300000L / TICK_PERIOD;
    led_pwm.stop_offtime = 300000L / TICK_PERIOD;
    LED_CTRL_OFF;
    led_pwm.ctrl_rest = 250000L / TICK_PERIOD; 
    led_pwm.ctrl_ontime = 300000L / TICK_PERIOD;
    led_pwm.ctrl_offtime = 300000L / TICK_PERIOD;
  }

void error_blink_leds(void)
  {
    LED_GO_OFF;
    led_pwm.go_rest = 50000L / TICK_PERIOD; 
    led_pwm.go_ontime = 150000L / TICK_PERIOD;
    led_pwm.go_offtime = 150000L / TICK_PERIOD;
    LED_STOP_OFF;
    led_pwm.stop_rest =  50000L / TICK_PERIOD; 
    led_pwm.stop_ontime = 150000L / TICK_PERIOD;
    led_pwm.stop_offtime = 150000L / TICK_PERIOD;
    LED_CTRL_OFF;
    led_pwm.ctrl_rest =  50000L / TICK_PERIOD; 
    led_pwm.ctrl_ontime = 150000L / TICK_PERIOD;
    led_pwm.ctrl_offtime = 150000L / TICK_PERIOD;
  }


//---------------------------------------------------------------------------------
// set_opendcc_state(t_opendcc_state next)
//
// Einstellen des nchsten Zustand mit:
// - Stellen entsprechender Bits in der Hardware und Rcksetzen der
//   berwachungstasks ->
//   damit wird nach "power on" wieder Strom auf den Ausgang gelegt.
// - Stellen der LEDs
// - Setzen der globalen Variablen, fallweise Nachricht an den Host
//

void main_on(void)
  {
    MAIN_TRACK_ON;
    if (ext_stop_enabled) DMX_OUT_HIGH;
  }
void main_off(void)
  {
    MAIN_TRACK_OFF;
    if (ext_stop_enabled) DMX_OUT_LOW;
  }

void set_opendcc_state(t_opendcc_state next)
  {
    opendcc_state = next;
    set_led_state(next);
    switch(next)
      {
        case RUN_OKAY:             // DCC running
            event_send(ALL_ON);
            PROG_TRACK_OFF;
            main_on();
            main_short = UNPRESSED;  // reset short detection
            break;
        case RUN_STOP:             // DCC Running, all Engines Emergency Stop
            event_send(ALL_OFF);

            PROG_TRACK_OFF;
            main_on();
            break;
        case RUN_OFF:              // Output disabled (2*Taste, PC)
            event_send(ALL_OFF);
            PROG_TRACK_OFF;
            main_off();
            break;
        case RUN_SHORT:            // Kurzschlu
            event_send(ALL_OFF);
            PROG_TRACK_OFF;
            main_off();
            break;
        case RUN_PAUSE:            // DCC Running, all Engines Speed 0
            event_send(ALL_OFF);
            PROG_TRACK_OFF;
            main_on();
            break;

        case PROG_OKAY:
            event_send(PROGMODE);
            PROG_TRACK_ON;
            main_off();
            prog_short = UNPRESSED;
            break;
        case PROG_SHORT:           //
            PROG_TRACK_OFF;
            main_off();
            break;
        case PROG_OFF:
            PROG_TRACK_OFF;
            main_off();
            break;
        case PROG_ERROR:
            PROG_TRACK_ON;
            main_off();
            break;
      }
  }


//-----------------------------------------------------------------
// run_state is a task, called periodically to check for key and
// output shorts.
//

// Hinweis: RUN_PAUSE wird zur Zeit nicht angesprungen

unsigned char ext_stop_enabled = 0;       


void run_state(void)
  {
    register unsigned char temp;

    if (MY_CTS_STATE == 1)
      {
        if (current_led_state != STOPPED) set_led_rs232(STOPPED);
      }
   
    temp=run_key();
 
    switch(opendcc_state)
      {
        case RUN_OKAY:             // wir laufen ganz normal:
            //if (temp == STOP_STROKE) set_opendcc_state(RUN_PAUSE);
            if (temp == STOP_STROKE) 
              {
                if (KEY_GO_PRESSED) set_opendcc_state(PROG_OKAY);  // Go gedrckt halten, dann Stop -> Progmode
                else set_opendcc_state(RUN_STOP);
              }
            #if (TURN_OFF_IF_SHORT == 1)
                if (run_main_short_check() == IS_SHORT)  set_opendcc_state(RUN_SHORT);
            #endif
            if (ext_stop_enabled)
              {
                if (EXT_STOP_PRESSED) 
                  {
                     set_opendcc_state(RUN_OFF);
                  }
              }
            break;
        case RUN_STOP:             // DCC Running, all Engines Emergency Stop
            if (temp == GO_STROKE) set_opendcc_state(RUN_OKAY);
            if (temp == STOP_STROKE) set_opendcc_state(RUN_OFF);
            if (run_main_short_check() == IS_SHORT)  set_opendcc_state(RUN_SHORT);
            break;
        case RUN_OFF:              // Output disabled (2*Taste, PC)
            if (temp == GO_STROKE) set_opendcc_state(RUN_OKAY);
            break;
        case RUN_SHORT:            // Kurzschlu
            if (temp == GO_STROKE) set_opendcc_state(RUN_OKAY);
            break;
        case RUN_PAUSE:            // DCC Running, all Engines Speed 0
            if (temp == STOP_STROKE) set_opendcc_state(RUN_OFF);
            if (temp == GO_STROKE) set_opendcc_state(RUN_OKAY);
            if (run_main_short_check() == IS_SHORT)  set_opendcc_state(RUN_SHORT);
            break;

        case PROG_OKAY:
            if (temp == STOP_STROKE) 
              {
                if (KEY_GO_PRESSED) set_opendcc_state(RUN_OKAY);  // Go gedrckt halten, dann Stop -> Runmode
                else set_opendcc_state(PROG_SHORT);
              }
            if (run_prog_short_check() == IS_SHORT)  set_opendcc_state(PROG_SHORT);           
            break;
        case PROG_SHORT:           //
            if (temp == GO_STROKE) set_opendcc_state(PROG_OKAY);
            break;
        case PROG_OFF:
            break;
        case PROG_ERROR:
            break;
      }
  }
  

// returns true, if we are in prog state
unsigned char is_prog_state(void)
  {
    unsigned char retval = 0;
    switch(opendcc_state)
      {
        case RUN_OKAY:             // wir laufen ganz normal:
        case RUN_STOP:             // DCC Running, all Engines Emergency Stop
        case RUN_OFF:              // Output disabled (2*Taste, PC)
        case RUN_SHORT:            // Kurzschlu
        case RUN_PAUSE:            // DCC Running, all Engines Speed 0
            retval = 0;
            break;

        case PROG_OKAY:
        case PROG_SHORT:           //
        case PROG_OFF:
        case PROG_ERROR:
            retval = 1;
            break;
      }
    return(retval);
  }
  
